// This file is part of Hermes2D.
//
// Copyright 2005-2008 Jakub Cerveny <jakub.cerveny@gmail.com>
// Copyright 2005-2008 Lenka Dubcova <dubcova@gmail.com>
// Copyright 2005-2008 Pavel Solin <solin@utep.edu>
//
// Hermes2D is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// Hermes2D is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Hermes2D.  If not, see <http://www.gnu.org/licenses/>.

// $Id: mesh.h 1112 2008-11-06 15:50:21Z jakub $

#ifndef __HERMES2D_MESH_H
#define __HERMES2D_MESH_H

#include "common.h"
#include "curved.h"

struct Element;
class HashTable;
class Space;  
struct MItem;


/// \brief Stores one node of a mesh.
///
/// There are are two variants of this structure, depending on the value of
/// the member 'type': 
/// <ol> <li> TYPE_VERTEX -- vertex node. Has physical coordinates x, y.
///      <li> TYPE_EDGE   -- edge node. Only stores edge marker and two element pointers.
/// </ol>
///
struct Node
{
  int id;          ///< node id number
  unsigned ref:29; ///< the number of elements using the node
  unsigned type:1; ///< 0 = vertex node; 1 = edge node
  unsigned bnd:1;  ///< 1 = boundary node; 0 = inner node
  unsigned used:1; ///< array item usage flag

  union
  {
    struct // vertex node variant: 
    { 
      double x, y; ///< vertex node coordinates
    };    
    struct // edge node variant:
    {// TODO: review pointer sizes for 64-bits !!!
      int marker;       ///< edge marker
      Element* elem[2]; ///< elements sharing the edge node
      Nurbs* nurbs;     ///< temporary curved edge ptr (only for loading the mesh)
    };
  };

  int p1, p2; ///< parent id numbers
  Node* next_hash; ///< next node in hash synonym list
  
  bool is_constrained_vertex() const { assert(type == TYPE_VERTEX); return ref <= 3 && !bnd; }
  
  void ref_element(Element* e = NULL);
  void unref_element(HashTable* ht, Element* e = NULL);
};


/// \brief Stores one element of a mesh.
///
/// The element can be a triangle or a quad (nvert == 3 or nvert = 4), active or inactive.
///
/// Active elements are actual existing elements in the mesh, which take part in the
/// computation. Inactive elements are those which have once been active, but were refined,
/// ie., replaced by several other (smaller) son elements. The purpose of the union is the
/// following. Active elements store pointers to their vertex and edge nodes. Inactive
/// elements store pointers to thier son elements and to vertex nodes.
/// 
/// If an element has curved edges, the member 'cm' points to an associated CurvMap structure,
/// otherwise it is NULL.
///
struct Element
{
  int id;            ///< element id number
  unsigned nvert:30; ///< number of vertices (3 or 4)
  unsigned active:1; ///< 0 = active, no sons; 1 = inactive (refined), has sons
  unsigned used:1;   ///< array item usage flag
  int marker;        ///< element marker
  int userdata;     ///< arbitrary user-defined data
  int iro_cache;     ///< increase in integration order, see RefMap::calc_inv_ref_order()
  
  Node* vn[4];   ///< vertex node pointers
  union 
  {
    Node* en[4];      ///< edge node pointers
    Element* sons[4]; ///< son elements (up to four)
  };
  
  CurvMap* cm; ///< curved mapping, NULL if not curvilinear
  
  bool is_triangle() const { return nvert == 3; }
  bool is_quad() const { return nvert == 4; }  
  bool is_curved() const { return cm != NULL; }
  int  get_mode() const { return is_triangle() ? MODE_TRIANGLE : MODE_QUAD; }
  
  // helper functions to obtain the index of the next or previous vertex/edge
  int next_vert(int i) const { return (i < nvert-1) ? i+1 : 0; }
  int prev_vert(int i) const { return (i > 0) ? i-1 : nvert-1; }
  
  bool hsplit() const { assert(!active); return sons[0] != NULL; }
  bool vsplit() const { assert(!active); return sons[2] != NULL; }
  bool bsplit() const { assert(!active); return sons[0] != NULL && sons[2] != NULL; } 
  
  /// Returns a pointer to the neighboring element across the edge 'ie', or
  /// NULL if it does not exist or is across an irregular edge.
  Element* get_neighbor(int ie) const;
  
  /// Calculates the area of the element. For curved elements, this is only
  /// an approximation: the curvature is not accounted for.
  double get_area() const;
  
  /// Returns the length of the longest edge for triangles, and the
  /// length of the longer diagonal for quads. Ignores element curvature.
  double get_diameter() const;
  
  void ref_all_nodes();
  void unref_all_nodes(HashTable* ht);
};


#include "hash.h"


/// \brief Represents a finite element mesh.
///
/// 
///
class Mesh : public HashTable
{
public:

  Mesh();
  ~Mesh() { free(); dump_hash_stat(); }
  /// Creates a copy of another mesh.
  void copy(const Mesh* mesh);
  /// Copies the coarsest elements of another mesh.
  void copy_base(Mesh* mesh);
  /// Frees all data associated with the mesh.
  void free();
  
  /// Loads the mesh from a file. Aborts the program on error.
  /// \param filename [in] The name of the file.
  void load_old(const char* filename);
  
  void load(const char* filename);
  
  /// Saves the mesh, including all refinements, to a file.
  /// Caution: never overwrite hand-created meshes with this function --
  /// all comments in the original file will be lost.
  /// \param filename [in] The name of the file.
  void save(const char* filename);
  
  /// Creates the mesh from the given vertex, triangle, quad and marker arrays
  void create(int nv, double2* verts, int nt, int4* tris,
              int nq, int5* quads, int nm, int3* mark);
    
  /// Retrieves an element by its id number.
  Element* get_element(int id) const;
  
  /// Returns the total number of elements stored.
  int get_num_elements() const { return elements.get_num_items(); }
  /// Returns the number of coarse mesh elements.
  int get_num_base_elements() const { return nbase; }
  /// Returns the current number of active elements in the mesh. 
  int get_num_active_elements() const { return nactive; } 
  /// Returns the maximum node id number plus one. 
  int get_max_element_id() const { return elements.get_size(); }

  /// Refines an element.
  /// \param id [in] Element id number.
  /// \param refinement [in] Ignored for triangles. If the element
  /// is a quad, 0 means refine in both directions, 1 means refine
  /// horizontally (with respect to the reference domain), 2 means
  /// refine vertically.
  void refine_element(int id, int refinement = 0);
  
  /// Refines all elements.
  /// \param refinement [in] Same meaning as in refine_element().
  void refine_all_elements(int refinement = 0);
  
  /// Selects elements to refine according to a given criterion and
  /// performs 'depth' levels of refinements. The criterion function 
  /// receives a pointer to an element to be considered.
  /// It must return -1 if the element is not to be refined, 0 if it
  /// should be refined uniformly, 1 if it is a quad and should be split
  /// horizontally or 2 if it is a quad and should be split vertically.
  void refine_by_criterion(int (*criterion)(Element* e), int depth);
  
  /// Performs repeated refinements of elements containing the given vertex.
  /// A mesh graded towards the vertex is created. 
  void refine_towards_vertex(int vertex_id, int depth);
  
  /// Performs repeated refinements of elements touching a part of the
  /// boundary marked by 'marker'. Elements touching both by an edge or
  /// by a vertex are refined. 'aniso' allows or disables anisotropic
  /// splits of quads.
  void refine_towards_boundary(int marker, int depth, bool aniso = true);
  
  /// Regularizes the mesh by refining elements with hanging nodes of
  /// degree more than 'n'. As a result, n-irregular mesh is obtained.
  /// If n = 0, completely regular mesh is created. In this case, however,
  /// due to incompatible refinements, the element refinement hierarchy
  /// is removed and all elements become top-level elements. Also, total
  /// regularization does not work on curved elements.
  void regularize(int n);
  
  /// Recursively removes all son elements of the given element and
  /// makes it active.
  void unrefine_element(int id);
  
  /// Unrefines all elements with immediate active sons. In effect, this
  /// shaves off one layer of refinements from the mesh. If done immediately
  /// after refine_all_elements(), this function reverts the mesh to its
  /// original state. However, it is not exactly an inverse to
  /// refine_all_elements().
  void unrefine_all_elements();
  
  void transform(double2x2 m, double2 t);
  void transform(void (*fn)(double* x, double* y));
  
  /// Loads the entire internal state from a (binary) file.
  void load_raw(FILE* f);
  /// Saves the entire internal state to a (binary) file.
  void save_raw(FILE* f);
  
  /// For internal use.
  int get_edge_sons(Element* e, int edge, int& son1, int& son2);
  /// For internal use.
  void set_zero_remap(int id) { zero_remap = id; }
  /// For internal use.
  unsigned get_seq() const { return seq; }
  /// For internal use.
  Element* get_element_fast(int id) const { return &(elements[id]);}
  

protected:

  Array<Element> elements;
  int nbase, ntopvert;
  int nactive;
  int zero_remap;
  unsigned seq;
  
  Element* create_triangle(int marker, Node* v0, Node* v1, Node* v2, CurvMap* cm);
  Element* create_quad(int marker, Node* v0, Node* v1, Node* v2, Node* v3, CurvMap* cm);

  void refine_triangle(Element* e);
  void refine_quad(Element* e, int refinement);
  void unrefine_element_internal(Element* e);

  char*  get_line(FILE* f);
  Nurbs* load_nurbs(FILE* f, Node** en, int& p1, int& p2);
  Nurbs* load_nurbs_new(MItem* curve, int id, Node** en, int &p1, int &p2);
  Nurbs* reverse_nurbs(Nurbs* nurbs);
  Node*  get_base_edge_node(Element* base, int edge);
  void   save_refinements(FILE* f, Element* e, int id, bool& first);
  void   save_nurbs(FILE* f, int p1, int p2, Nurbs* nurbs);

  int  get_edge_degree(Node* v1, Node* v2);
  void regularize_triangle(Element* e);
  void regularize_quad(Element* e);
  void flatten();
  
};


// helper macros for easy iteration through all elements, nodes etc. in a mesh
#define for_all_elements(e, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_element_id(); _id < _max; _id++) \
          if (((e) = (mesh)->get_element_fast(_id))->used)
          
#define for_all_base_elements(e, mesh) \
        for (int _id = 0; _id < (mesh)->get_num_base_elements(); _id++) \
          if (((e) = (mesh)->get_element_fast(_id))->used)
            
#define for_all_active_elements(e, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_element_id(); _id < _max; _id++) \
          if (((e) = (mesh)->get_element_fast(_id))->used) \
            if ((e)->active)

#define for_all_inactive_elements(e, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_element_id(); _id < _max; _id++) \
          if (((e) = (mesh)->get_element_fast(_id))->used) \
            if (!(e)->active)
              
#define for_all_nodes(n, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_node_id(); _id < _max; _id++) \
          if (((n) = (mesh)->get_node(_id))->used)

#define for_all_vertex_nodes(n, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_node_id(); _id < _max; _id++) \
          if (((n) = (mesh)->get_node(_id))->used) \
            if (!(n)->type)

#define for_all_edge_nodes(n, mesh) \
        for (int _id = 0, _max = (mesh)->get_max_node_id(); _id < _max; _id++) \
          if (((n) = (mesh)->get_node(_id))->used) \
            if ((n)->type)
            

/// \brief Determines the position on an edge.
/// \details Used for the retrieval of boundary condition values.
///
struct EdgePos
{
  int v1, v2; ///< edge endpoint vertex id numbers
  int marker; ///< edge marker
  int edge;   ///< element edge number (0..3 for triangles, 0..4 for quads)
  double t;   ///< position between v1 and v2 in the range [0..1]
  
  double lo, hi; ///< for internal use
  Element* base; ///< for internal use
  Space *space, *space_u, *space_v; ///< for internal use
};


#endif
