#pragma once
#include <Events.h>
#include <DataSet.h>


namespace rofrm {


struct SystemColor
{
	static Color
		cWindow, cWindowText, cControl, cControlText, cControlShadow;
};


struct Font
{
	enum {cBold=1, cItalic=2};
	int style;
	int size;
	mutable string name;
	Font(int size=0, int style=0): size(size), style(style) {
	}
	Font(const string& name, int size, int style);
	bool operator!() const {
		return name.empty();
	}
	friend bool operator==(const Font&, const Font&);
};


class FontProperty {
public:
	int GetHeight() const;
	int StringWidth(const string& str) const;
	const Font& operator=(const Font &f);
	operator Font() const;
	FontProperty(Control *ctrl)
		:ctrl(ctrl) {}
private:
	Control *ctrl;
};


class ColorProperty {
public:
	ColorProperty(Control *ctrl)
		:ctrl(ctrl) {}
	Color operator=(const Color c);
	operator Color() const
		{ return RetriveColor(); };
	bool IsDefault() const
		{ return color==Color::cDefault; }
	int GetRGB() const
		{ return RetriveColor().GetRGB(); }
private:
	Color RetriveColor() const;
	Color color;
	Control *ctrl;
};



inline bool operator==(const Font& f1, const Font& f2) {
	return f1.size == f2.size && f1.style == f2.style &&
		f1.name == f2.name;
}

inline bool operator!=(const Font& f1, const Font& f2) {
	return !(f1 == f2);
}


class Cursor {
public:
	enum {cDefault=0, cArrow, cCross, cNo, cWait, cBeam,
		cHand, cSizeV, cSizeH, cSizeSl, cSizeBkSl, cSizeAll};
	Cursor(int type=cDefault):type(type), hcur(0){
	}
	HCursor GetHandle() const {
		if(hcur==0 && type) Create();
		return hcur;
	}
	int GetType(){ return type;}
	bool operator!=(const Cursor& cur) {
		return type != cur.type;
	}
protected:
	int type;
	mutable HCursor hcur;
	void Create() const;
};


struct Align {
	enum T {cLeft, cCenter, cRight};
};


struct Dock {
	enum T{cNone, cTop, cBootom, cLeft, cRight, cFill};
	T value;
	void operator=(T v) {
		value=v;
	}
};


enum CtrlType {
	cControl, cContainer, cForm
	};


class DataLink
{
protected:
	Field *field;
	Control *ctrl;
public:
	DataLink(Control *ctrl);
	DataLink& operator=(Field &fld) {
		field =&fld;
	}
	virtual void Load();
	virtual void Store();
};


class ActionMenu: public object
{
public:
	ActionMenu():ctrl(0){
	}
	void Add(Action *action) {
		actions.push_back(action);
	}
	bool Exec(int cmd);
	virtual void Setup(Control *ctrl) = 0;
protected:
	vector<Action*> actions;
	Control *ctrl;
};




struct BorderStyle
{
	enum {cNone, cFixed3d, cStatic, cFrame, cRised};
};


typedef std::vector<Control*> ControlList;


enum ControlState{stNone, stClose, stCreate, stLayout,  stConfig,  stShow};
class SysImpl;

class Control : public object
{
	friend class SysImpl;
	friend class Container;
public:
	SysImpl &impl;
	virtual ~Control();
	void Focus();
	Control* GetFocused();
	void Show()	   {visible=true;}
	void Hide()     {visible=false;}
	bool IsShowing() const;
	Dock dock;
	property<string, true> text;
	property<bool> enabled;
	property<bool> visible;
	property<int>  left, top;
	property<int>  width, height;
	property<Cursor> cursor;
	ColorProperty foreColor, backColor;
	FontProperty font;
	property<int> borderStyle;
	property<string> toolTip;
	event<NotifyArgs>  onCreate;
	event<KeyArgs> onKeyDown, onKeyUp, onChar;
	event<MouseArgs> onMouseDown, onMouseUp, onMouseMove;
	event<MouseArgs> onMouseWheel, onContextMenu;
	event<NotifyArgs> onMouseLeave, onDblClick;//, onMouseEnter;
	event<NotifyArgs> onMove, onResize;  //Obsuda w te Window zdublowanie ?????
	event<NotifyArgs> onGotFocus, onLostFocus;
	event<NotifyArgs> onShowHide;
	ActionMenu *contextMenu;
	const Rect& GetBounds() const          	{return bounds;}
	void SetBounds(int x, int y, int width, int height);
	void SetLocation(int x, int y)           	{SetBounds(x, y, bounds.width, bounds.height);}
	void SetSize(int width, int height)      	{SetBounds(bounds.x, bounds.y, width, height);}
	virtual Size GetMinimumSize() const  	{return minSize;}
	virtual void Layout();
	void ToFront();
	void ToBack();
	bool IsFocused();
	Control* GetParent() const               	{return parent;}
	void Insert(Control* comp, int index=-1);
	void Remove(Control* comp);
	void Remove(int index);
	int GetControlCount() const              	{return static_cast<int>(children.size());}
	Control* GetControl(int index) const  	{return children[index];}
	void ScreenToClient(Point &pt);
	void ClientToScreen(Point &pt);
	int tag;
	object *link;
	void Destroy();
	void LoadData();
	void StoreData();
protected:
	Rect bounds;
	Control();
	Control(SysImpl *impl);
	mutable Size minSize;
	bool needLayout;
	Control* parent;
	virtual void Create();
	void SpreadState(ControlState state);
	void LayoutTree();
public:
	void GetClientRect(Rect &cr);
	void InvalidateRect(const Rect *pR, bool bErase=true);
	void Invalidate() { InvalidateRect(NULL);  }
	void UpdateText();
	Control* GetParentForm();
	virtual const Color& DefaultForeColor() const;
	virtual const Color& DefaultBackColor() const;
	ControlState GetState() const {return state;};
	bool IsContainer();
protected:
	ControlState state;
	uint id;
	ControlList children;
	void SetId(uint id);
	uint GetId() const                     {return id;}
	// no copy
	Control(const Control& other);
	Control& operator=(const Control& other);
	template<class T>
	T& GetImpl() {return (T&)impl;}
	void SetDesignMode();
};


class Container : public Control{
public:
	class Impl;
	Size GetMinimumSize() const   { return minSize;  }
	void UpdateLayout() {
		needLayout =true;
		if (state) 	LayoutTree();
	}
	void Add(Control* comp)                  	{Insert(comp, -1);}
protected:
	Container(Impl *hw);
	void Layout();
private:
};




class Form : public Container
{
public:
	enum WindowType {cNormal, cToolWindow, cNoTitleBar} windowType;
	class Impl;
	bool center;
	bool resizable;
	bool undecorated;
	bool commandkey;
	int result;
	Control *focusCtrl;
	event<NotifyArgs> onClose, onDestroy, onShow;
	event<NotifyArgs> onChangeState, onActivate;
	event<NotifyArgs> onKeyPreview;
	Form(Form *parent);
	Form();
	virtual ~Form();
	void Show() {
		ShowForm(false);
	}
	void ShowDialog() {
		ShowForm(true);
	}
	void Close();
	void SetMenu(ActionMenu *am) {
		menuBar = am;
	}
	static Form* GetActiveForm() {
		return currentActive;
	}
	Control* GetFocused();
	bool FindAppWindow(const string &title, bool close=false);
	void SetForeground();
protected:
	typedef std::vector<Form*> FormList;
	bool modal;
	Form *owner;
	FormList ownedForms;
	ActionMenu* menuBar;
	void Init(Form* parent);
	void ShowForm(bool modal);
	void Destroy();
	void Activate();
	void Notify(uint type, int cmd) ;
	void AddOwnedForm(Form* w);
	void RemoveOwnedForm(Form* w);
	static Form* currentActive;
};


enum {cStockButton=1, cStockMenu=2};

struct Icons {
	enum {icoNew=1, icoOpen, icoOpenPrj, icoSave, icoSaveAll, icoCut, icoCopy, icoPaste, icoRedBall,
			icoGo, icoDesign, icoClear, icoBuild, icoNo, icoRun, icoPoint, icoIndex,
			icoCloseAll, icoStar, icoOk, icoCancel, icoColor, icoConvert, icoHome, icoEdit,
			icoFind, icoSetup, icoPrint, icoRetry, icoAdd, icoRemove, icoPicker, icoReplace,
		icoInfo, icoForward, icoBack, icoGreenBall, icoRedo, icoUndo, icoNet, icoQuit, icoUp, icoDown};
};

struct ImageResource
{
	const int resId;
	const int width, height;
	const int count;
	ImgListImpl &impl;
	ImageResource(int resId, int wd, int ht, ResRef ref);
	~ImageResource();
};

typedef vector<string> ArgumentList;

class Application {
	friend class ImageResource;
	class Impl;
public:
	Impl &impl;
	Form  *const  topForm;
	const ImageResource* GetImageResource(int resId);
	void AddImageResource(int resId, int wd, int ht, ResRef ref);
	static string appPath;
	static void ProcessMessages();
	static int MsgBox(const string &msg, const string &tit, int flags);
	static Size GetScreenSize();
	static int GetScreenResolution();
	static int Main(const rofrm::ArgumentList&);
	void (*handleIdle)();
protected:
	vector<ImageResource*> imageRes;
	Application();
	~Application();
};

inline
	bool MsgBox(string msg) {
	return Application::MsgBox(msg, "Information", cmbOK)!=0;
}

bool ConfBox(string msg,  bool *cancel=0);

void Beep(int n=-1);


extern
	Application &application;

}//namespace

extern
	void RegisterResource();
