#pragma once

#include <Core.h>

typedef void *WindowID;

#include "scin/Scintilla.h"
//#include <scin/PropSet.h>

namespace rofrm {

class SciPropSet;

enum IndentationStatus {
	isNone,		// no effect on indentation
	isBlockStart,	// indentation block begin such as "{" or VB "function"
	isBlockEnd,	// indentation end indicator such as "}" or VB "end"
	isKeyWordStart	// Keywords that cause indentation
};

enum UniMode {
	uni8Bit=0, uni16BE=1, uni16LE=2, uniUTF8=3,
	uniCookie=4
};


struct SciStyleDef{
	enum { sdBold = 0x1000, sdItalic=0x2000,  sdFilled=0x8000};
	uint styleId;
	int color;
	uint flags;
	int bgcolor;
	int Bold() const {return (flags&sdBold)?1:0;}
	int Italic() const {return (flags&sdItalic)?1:0;}
	int Filled() const {return (flags&sdFilled)?1:0;}
	int Size() const {return (flags&0xFF);}
 };


class SciLexerDef
{
public:
	string lexer;
	const SciStyleDef *styles;
	const char *keywords[6];
	const char *statementIndent;
	const char *blockStart, *blockEnd;
	bool callTipIgnoreCase;
	string calltipWordCharacters;
	string calltipParametersStart;
	string calltipParametersEnd;
	string calltipParametersSeparators;
	string calltipEndDefinition;
	bool autoCompleteIgnoreCase;
	string autoCompleteStartCharacters;
	string autoCompleteFillUpCharacters;
	int autoCompleteChooseSingle;
	int autoCompleteSetCancelAtStart;
	int autoCompleteSetDropRestOfWord;
	bool autoCompleteWordAutomatic;
	const char *fileFilter;
	SciLexerDef(const SciStyleDef styles[], const char* kwords);
};

struct MarkerDef {
	int shape;
	long foreColor;
	long backColor;
	bool single, select;
	MarkerDef(int shp, long fc, long bc, bool sng=true, bool sel=false):
		shape(shp), foreColor(fc), backColor(bc), single(sng), select(sel) {
	}
};

struct RangeToFormat;

class SciEditor : public Control {
//friend class SciPropSet;
friend class DataAccessor;
public:
	class Impl;
	string filePath;
	SciEditor();
	~SciEditor();
	virtual bool SciNotify(SCNotification *notif);
	bool MarginClick(int position, int modifiers);
	event<NotifyArgs>  onChoose, onDelete;
	int GetCurrentLineNumber();
	int GetCaretInLine();
	int GetCurrentPos()             {return SendEditor(SCI_GETCURRENTPOS);}
	int DocumentLength()            {return SendEditor(SCI_GETLENGTH);}
	int LineFromPos(int pos)        {return SendEditor(SCI_LINEFROMPOSITION, pos);}
	int PosFromLine(int line)       {return SendEditor(SCI_POSITIONFROMLINE, line);}
	int PosFromPoint(int x, int y)  {return SendEditor(SCI_POSITIONFROMPOINTCLOSE, x, y);}
	int GetLineCount()              {return SendEditor(SCI_GETLINECOUNT);}
	int GetLineLength(int line);
	int GetLineState(int line)      {return SendEditor(SCI_GETLINESTATE, line);}
	int GetFoldLevel(int line)      {return SendEditor(SCI_GETFOLDLEVEL, line)&SC_FOLDLEVELNUMBERMASK;}
	int GetFoldParent(int line)     {return SendEditor(SCI_GETFOLDPARENT, line);}
	int GetFoldLastChild(int line)  {return SendEditor(SCI_GETLASTCHILD, line);}
	int GetLineIndentation(int line){return SendEditor(SCI_GETLINEINDENTATION, line);}
	int GetLineIndentPosition(int line) {
		return SendEditor(SCI_GETLINEINDENTPOSITION, line);
	}
	void SetLineIndentation(int line, int indent);
	void GetRange(int start, int end, char *text);
	string GetWord();
	string WordFromPos(int pos);
	string GetLine(int line=-1);
	string GetRange(int start, int end);
	CharacterRange GetSelection();
	void SetSelection(int anchor, int currentPos);
	string RangeExtendAndGrab(int &selStart, int &selEnd,
	bool (SciEditor::*ischarforsel)(char ch), bool stripEol = true);
	int IsLinePreprocessorCondition(char *line);
	bool FindMatchingPreprocessorCondition(int &curLine, int direction, int condEnd1, int condEnd2);
	bool FindMatchingPreprocCondPosition(bool isForward, int &mppcAtCaret, int &mppcMatch);
	bool FindMatchingBracePosition(bool editor, int &braceAtCaret, int &braceOpposite, bool sloppy);
	void BraceMatch(bool editor);
	void GotoLineEnsureVisible(int line);
	void GoToPos(int pos)								{ SendEditor(SCI_GOTOPOS, pos);}
	void EnsureRangeVisible(int posStart, int posEnd, bool enforcePolicy=true);
	void BookmarkAdd(int lineno = -1);
	void BookmarkDelete(int lineno = -1);
	bool BookmarkPresent(int lineno = -1);
	void BookmarkToggle(int lineno = -1);
	void BookmarkNext(bool forwardScan = true, bool select = false);
	void SetLineNumberWidth();
	void ViewWhitespace(bool view);
	void RemoveFindMarks();
	void GoMatchingBrace(bool select);
	bool StartAutoComplete();
	bool StartAutoCompleteWord(bool onlyOneWord);
	void FillFunctionDefinition(int pos=-1);
	bool StartCallTip();
	void ContinueCallTip();
	Point GetCursorPos();
	int HighlightLine(int srcline, int mid=0, bool Add = true);
	int GetMarkerLineNumber(int sciH);
	int GetMarker(int line) {return SendEditor(SCI_MARKERGET, line);}
	int GetNextMarker(int line, int msk) {return SendEditor(SCI_MARKERNEXT, line, msk);}
	void SetLexerDef(SciLexerDef *lexer);
	int FindInTarget(const char *findWhat, int len, int startPos, int endPos);
	int FindNext(bool revDir);
	void ReplaceOnce();
	int ReplaceAll(bool inSelection);
	string SelectionExtend(bool (SciEditor::*ischarforsel)(char ch), bool stripEol = true);
	string SelectionWord(bool stripEol=true);
	void SelectionIntoFind(bool stripEol=true);
	bool InsertText(const char *s, int pos=-1);
	bool ReplaceSel(const char *s);
	void StripTrailingSpaces();
	void EnsureFinalNewLine();
	void MoveLeft(bool ext=false)       {SendEditor(SCI_CHARLEFT|(ext?1:0));}
	void MoveRight(bool ext=false)      {SendEditor(SCI_CHARRIGHT|(ext?1:0));}
	void MoveWordLeft(bool ext=false)   {SendEditor(SCI_WORDLEFT|(ext?1:0));}
	void MoveWordRight(bool ext=false)  {SendEditor(SCI_WORDRIGHT|(ext?1:0));}
	void MoveHome(bool ext=false)       {SendEditor(SCI_HOME|(ext?1:0));}
	void MoveEnd(bool ext=false)        {SendEditor(SCI_LINEEND|(ext?1:0));}
	void MoveUp(bool ext=false)         {SendEditor(SCI_LINEUP|(ext?1:0));}
	void MoveDown(bool ext=false)       {SendEditor(SCI_LINEDOWN|(ext?1:0));}
	void MoveDocStart(bool ext=false)   {SendEditor(SCI_DOCUMENTSTART|(ext?1:0));}
	void MoveDocEnd(bool ext=false)     {SendEditor(SCI_DOCUMENTEND|(ext?1:0));}
	void Find();
	void Replace(bool rep);
	void GoToLine();
	void ShowTooltip(const string &s, int pos);
	void HideTooltip();
	void ApplyFoldProp();
	void SetIndentSettings();
	void ApplyFontProperties();
	void ApplyProperties();
	void New();
	void OpenFile(const char *fileName);
	bool SaveBuffer(const char *fileName, bool strip=true, bool eof=true);
	int FormatRange(bool  print,  RangeToFormat *range=0) {
		return SendEditor(SCI_FORMATRANGE, print, reinterpret_cast<sptr_t>(range));
	}
	void Print();
	void ExecCmd(int cmdID);
protected:
	bool modified;
	bool tooltipVisible;
	SciPropSet &props;
	int braceCount;
	bool wrap;
	bool isReadOnly;
	bool indentationWSVisible;
	const char* warnNoOtherBookmark;
	bool macrosEnabled;
	string currentMacro;
	bool recording;
	bool lineNumbersExpand;
	bool autoCCausedByOnlyOne;
	int startCalltipWord;
	int currentCallTip;
	int lastPosCallTip;
	string currentCallTipWord;
	string functionDefinition;
	StringList curMacro;
	char fullPath[128];
	string language;
	int maxMarkers;
	MarkerDef *markers;
	int lexLanguage;
	SciLexerDef *lexerDefs;
	UniMode unicodeMode;
	int codePage;
	bool showMessages;
	bool RangeIsAllWhitespace(int start, int end);
	// do props!!!
	unsigned int GetLinePartsInStyle(int line, int style1, int style2, string sv[], int len);
	IndentationStatus GetIndentState(int line);
	int IndentOfBlock(int line);
	void MaintainIndentation(char ch);
	void AutomaticIndentation(char ch);
	void CharAdded(char ch);
	void RecordMacroCommand(SCNotification *notification);
	void ExecuteMacroCommand(const char *command);
	void PlayCurMacro();
	//SciPropSet props;
	sptr_t SendEditor(unsigned int msg, uptr_t wParam=0, sptr_t lParam=0);
	sptr_t SendEditorString(unsigned int msg, uptr_t wParam, const char *s) {
		return SendEditor(msg, wParam, reinterpret_cast<sptr_t>(s));
	}
	void DefineMarker(int marker, int markerType, long cfore, long cback);
	void ApplyFontStyle(const SciStyleDef &sd);
	MarkerDef* GetMarkerDef(int mid);
	bool iswordcharforsel(char ch);
	bool islexerwordcharforsel(char ch);
	virtual void OnCharAdded() {};
	virtual void OnChangeState() {};
	virtual void OnWordTooltip(int pos) {};
	virtual bool FillCompleteList(string &pat, string &list) {};
	virtual bool FillCallTipList(string &pat, string &list) {};
	virtual void UpdateDisplay(int type) {};
private:
	Point tooltipPt;
	void HandleReplaceCmd(NotifyArgs &e);
	Impl& pim() {return (Impl&)impl;}
};


class DataAccessor {
protected:
	enum {extremePosition=0x7FFFFFFF};
	enum {bufferSize=4000, slopSize=bufferSize/8};
	int startPos;
	int endPos;
	char buf[bufferSize+1];
	int lenDoc;
	SciEditor &editor;
	void Fill(int position);
public:
	DataAccessor(SciEditor *ed): editor(*ed), startPos(extremePosition), endPos(0), lenDoc(-1) {
	}
	char operator[](int position) {
		if (position < startPos || position >= endPos) {
			Fill(position);
		}
		return buf[position - startPos];
	}
	void SetDocEndLimit(int lim)  {lenDoc = lim;}
	char StyleAt(int position) {
		return static_cast<char>(editor.SendEditor(SCI_GETSTYLEAT, position, 0));
	}
};


 struct StyleAndWords {
	int styleNumber;
	string words;
  StyleAndWords():styleNumber(10){};
	bool IsEmpty() { return words.length() == 0; }
	bool IsSingleChar() { return words.length() == 1; }
};


class SciPropSet//: public PropSet
{
friend class DataAccessor;
public:
	SciEditor::Impl &impl;
	int codePage;
	int characterSet;
	int wrapStyle;
	int bracesStyle;
	//margin
	bool margin;
	int marginWidth;
	enum { marginWidthDefault = 20};
	//line numbers
	bool lineNumbers;
	int lineNumbersWidth;
	enum { lineNumbersWidthDefault = 3 };
	//fold
	int foldSymbol;
	bool foldMargin;
	int foldMarginWidth;
	enum { foldMarginWidthDefault = 14};
	//buffer
	//bool bufferedDraw;
	//bool twoPhaseDraw;
	//breces
	bool bracesCheck;
	bool bracesSloppy;
	//chars
	string wordCharacters;
	string whitespaceCharacters;
	//indent
	bool indentOpening;
	bool indentClosing;
	bool indentMaintain;
	bool indentAutomatic;
	int statementLookback;
	bool highlightIndentationGuides;
	int caretWidth;
	long caretLineBack;
	int caretLineBackAlpha;
	long findMark;
	int caretXFlags;
	int caretYFlags;
	int visiblePolicyFlags;
	bool findReplaceRegexpPosix;
	int monoFontSize, propFontSize;
	string monoFontName, propFontName;
	bool useMonoFont;
	StyleAndWords statementIndent;
	StyleAndWords blockStart, blockEnd;
	SciPropSet(SciEditor::Impl *pimpl);
};


inline
int colourDesired(unsigned int red, unsigned int green, unsigned int blue) {
  return  red | (green << 8) | (blue << 16);
}


enum SciEditCommands {
	cedFirstCmd = 800,
	cedUprCase, cedLwrCase, cedJoin, cedSplit, cedExpand, cedToggleFoldRecursive, cedCut, cedCopy, cedPaste,
	cedFind, cedIncSearch, cedFindNext, cedFindNextBack, cedFindNextSel, cedEnterSelection, cedFindNextBackSel,
	cedReplace, cedGoTo, cedShowCalltip, cedComplete, cedCompleteWord, cedMatchBrace, cedSelectToBrace,
	cedLineNumberMargin, cedSelMargin, cedFoldMargin, cedViewEOL, cedWrap, cedReadonly, cedEol_crlf,
	cedEol_cr, cedEol_lf, cedEolConvert, cedViewSpace, cedViewGuides, cedBookmarkToggle, cedBookmarkNext,
	cedBookmarkPrev, cedBookmarkNextSelect, cedBookmarkPrevSelect, cedBookmarkClearAll,
	cedDeleteLine, cedMacroList, cedMacroPlay, cedMacroRecord, cedSwitchFont, cedUndo, cedRedo,
	cedLastCmd
};


struct EditorActs: ActionList
{
typedef Action *PAction;
	PAction	edCut,edCopy,edPaste,edFind,edFindSel,edFindSelBk,edFindRep,edFindPrv,edReplace,edGoTo,
				edSetBook,edNextBook,edPrevBook,edSelNextBook,edSelPrevBook,edClearBooks,
				edMatchBrace,edSelBrace,edToLower,edToUpper,edLineNum,edDelLine, edUndo,edRedo,
				edRecMacro,edPlayMacro,edSwitchFont;
	EditorActs();
};

struct SciEditorConfig
{
	int useTabs;
	int tabSize;
	int indentSize;
	int useMonoFont;
	SciEditorConfig() :useTabs(1), tabSize(4), indentSize(4), useMonoFont(0) {
	}
};



extern
	SciEditorConfig editorConfig;

extern
	EditorActs editorActs;

}

