##:filebuf@filebuf
filebuf
/*-p-
@b:The actual work of input and output (for files).
This class associates both its input and output sequence with an
external disk file, and maintains a joint file position for both
sequences.  Many of its sematics are described in terms of similar
behavior in the Standard C Library's c FILE streams.
/*/

#-basic_filebuf@filebuf
basic_filebuf()
/*-p-
@b:Does not open any files.
The default constructor initializes the parent class using its
own default ctor.
/*/

#-~basic_filebuf@filebuf
virtual ~basic_filebuf()
/*-p-
@b:The destructor closes the file first.
/*/

#-is_open@filebuf
bool is_open() const
/*-p-
@b:Returns true if the external file is open.
/*/

#-open@filebuf
__filebuf_type open(const char* s, ios_base::openmode mode)
/*-p-
@b:Opens an external file.
@p:s  The name of the file.
@p:mode  The open mode flags.
@r:c this on success, NULL on failure
If a file is already open, this function immediately fails.
Otherwise it tries to open the file named  s using the flags
given in  mode.
Table 92, adapted here, gives the relation between openmode
combinations and the equivalent fopen() flags.
(NB: lines in|out|app and binary|in|out|app per DR 596)
+---------------------------------------------------------+
| ios_base Flag combination            stdio equivalent   |
|binary  in  out  trunc  app                              |
+---------------------------------------------------------+
|             +                        "w"                |
|             +           +            "a"                |
|             +     +                  "w"                |
|         +                            "r"                |
|         +   +                        "r+"               |
|         +   +     +                  "w+"               |
|         +   +           +            "a+"               |
+---------------------------------------------------------+
|   +         +                        "wb"               |
|   +         +           +            "ab"               |
|   +         +     +                  "wb"               |
|   +     +                            "rb"               |
|   +     +   +                        "r+b"              |
|   +     +   +     +                  "w+b"              |
|   +     +   +           +            "a+b"              |
+---------------------------------------------------------+
/*/

#-close@filebuf
__filebuf_type close()
/*-p-
@b:Closes the currently associated file.
@r:c this on success, NULL on failure
If no file is currently open, this function immediately fails.
If a "put buffer area" exists, c overflow(eof) is called to flush
all the characters.  The file is then closed.
If any operations fail, this function also fails.
/*/

#-showmanyc@filebuf
streamsize showmanyc()
/*-p-
/*/

#-underflow@filebuf
int_type underflow()
/*-p-
/*/

#-pbackfail@filebuf
int_type pbackfail(int_type c = Traits::eof())
/*-p-
/*/

#-overflow@filebuf
int_type overflow(int_type c = Traits::eof())
/*-p-
/*/

#-setbuf@filebuf
__streambuf_type setbuf(char_type* s, streamsize n)
/*-p-
@b:Manipulates the buffer.
@p:s  Pointer to a buffer area.
@p:n  Size of  s.
@r:c this
If no file has been opened, and both  s and  n are zero, then
the stream becomes unbuffered.  Otherwise, c s is used as a
buffer; see
http://gcc.gnu.org/onlinedocs/libstdc++/27_io/howto.html#2
for more.
/*/

#-seekoff@filebuf
pos_type seekoff(off_type off, ios_base::seekdir way, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

#-seekpos@filebuf
pos_type seekpos(pos_type pos, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

#-sync@filebuf
int sync()
/*-p-
/*/

#-imbue@filebuf
void imbue(const locale& loc)
/*-p-
/*/

#-xsgetn@filebuf
streamsize xsgetn(char_type* s, streamsize n)
/*-p-
/*/

#-xsputn@filebuf
streamsize xsputn(const char_type* s, streamsize n)
/*-p-
/*/

##:fstream@fstream
fstream
/*-p-
@b:Controlling intput and output for files.
This class supports reading from and writing to named files, using
the inherited functions from std::basic_iostream.  To control the
associated sequence, an instance of std::basic_filebuf is used, which
this page refers to as c sb.
/*/

#-basic_fstream@fstream
basic_fstream()
/*-p-
@b:Default constructor.
Initializes c sb using its default constructor, and passes
(you haven't given it a filename to open).
/*/

#-basic_fstream@fstream
explicit basic_fstream(const char* s, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
@b:Create an input/output file stream.
@p:s  Null terminated string specifying the filename.
@p:mode  Open file in specified mode (see std::ios_base).
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-~basic_fstream@fstream
~basic_fstream()
/*-p-
@b:The destructor does nothing.
The file is closed by the filebuf object, not the formatting
stream.
/*/

#-rdbuf@fstream
__filebuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_filebuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-is_open@fstream
bool is_open()
/*-p-
@b:Wrapper to test for an open file.
@r:c rdbuf()->is_open()
/*/

#-is_open@fstream
bool is_open() const
/*-p-
/*/

#-open@fstream
void open(const char* s, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
@b:Opens an external file.
@p:s  The name of the file.
@p:mode  The open mode flags.
Calls c std::basic_filebuf::open(s,mode).  If that
function fails, c failbit is set in the stream's error state.
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-close@fstream
void close()
/*-p-
@b:Close the file.
Calls c std::basic_filebuf::close().  If that function
fails, c failbit is set in the stream's error state.
/*/

##:ifstream@ifstream
ifstream
/*-p-
@b:Controlling input for files.
This class supports reading from named files, using the inherited
functions from std::basic_istream.  To control the associated
sequence, an instance of std::basic_filebuf is used, which this page
refers to as c sb.
/*/

#-basic_ifstream@ifstream
basic_ifstream()
/*-p-
@b:Default constructor.
Initializes c sb using its default constructor, and passes
(you haven't given it a filename to open).
/*/

#-basic_ifstream@ifstream
explicit basic_ifstream(const char* s, ios_base::openmode mode = ios_base::in)
/*-p-
@b:Create an input file stream.
@p:s  Null terminated string specifying the filename.
@p:mode  Open file in specified mode (see std::ios_base).
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-~basic_ifstream@ifstream
~basic_ifstream()
/*-p-
@b:The destructor does nothing.
The file is closed by the filebuf object, not the formatting
stream.
/*/

#-rdbuf@ifstream
__filebuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_filebuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-is_open@ifstream
bool is_open()
/*-p-
@b:Wrapper to test for an open file.
@r:c rdbuf()->is_open()
/*/

#-is_open@ifstream
bool is_open() const
/*-p-
/*/

#-open@ifstream
void open(const char* s, ios_base::openmode mode = ios_base::in)
/*-p-
@b:Opens an external file.
@p:s  The name of the file.
@p:mode  The open mode flags.
Calls c std::basic_filebuf::open(s,mode|in).  If that function
fails, c failbit is set in the stream's error state.
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-close@ifstream
void close()
/*-p-
@b:Close the file.
Calls c std::basic_filebuf::close().  If that function
fails, c failbit is set in the stream's error state.
/*/

##:iostream@iostream
iostream
/*-p-
@b:Merging istream and ostream capabilities.
This class multiply inherits from the input and output stream classes
simply to provide a single interface.
/*/

#-char_type@iostream
_CharT char_type
/*-p-
/*/

#-int_type@iostream
int_type int_type
/*-p-
/*/

#-pos_type@iostream
pos_type pos_type
/*-p-
/*/

#-off_type@iostream
off_type off_type
/*-p-
/*/

#-traits_type@iostream
_Traits traits_type
/*-p-
/*/

#-basic_iostream@iostream
explicit basic_iostream(basic_streambuf<_CharT, Traits>* sb)
/*-p-
@b:Constructor does nothing.
Both of the parent classes are initialized with the same
streambuf pointer passed to this constructor.
/*/

#-~basic_iostream@iostream
virtual ~basic_iostream()
/*-p-
@b:Destructor does nothing.
/*/

#-basic_iostream@iostream
explicit basic_iostream()
/*-p-
/*/

##:istream@istream
istream
/*-p-
@b:Controlling input.
This is the base class for all input streams.  It provides text
formatting of all builtin types, and communicates with any class
derived from basic_streambuf to do the actual input.
/*/

#-operator >>@istream
basic_istream operator >>(basic_istream<_CharT2, Traits2>&, CharT2*)
/*-p-
/*/

#-basic_istream@istream
explicit basic_istream(__streambuf_type* sb)
/*-p-
@b:Base constructor.
This ctor is almost never called by the user directly, rather from
derived classes' initialization lists, which pass a pointer to
their own stream buffer.
/*/

#-~basic_istream@istream
virtual ~basic_istream()
/*-p-
@b:Base destructor.
This does very little apart from providing a virtual base dtor.
/*/

#-gcount@istream
streamsize gcount() const
/*-p-
@b:Character counting
@r:The number of characters extracted by the previous
unformatted input function dispatched for this stream.
/*/

#-get@istream
int_type get()
/*-p-
@b:Simple extraction.
@r:A character, or eof().
Tries to extract a character.  If none are available, sets failbit
and returns traits::eof().
/*/

#-get@istream
__istream_type get(char_type& c)
/*-p-
@b:Simple extraction.
@p:c  The character in which to store data.
@r:*this
Tries to extract a character and store it in  c.  If none are
available, sets failbit and returns traits::eof().
unsigned char.
/*/

#-get@istream
__istream_type get(char_type* s, streamsize n, char_type delim)
/*-p-
@b:Simple multiple-character extraction.
@p:s  Pointer to an array.
@p:n  Maximum number of characters to store in  s.
@p:delim  A "stop" character.
@r:*this
Characters are extracted and stored into  s until one of the
following happens:
- c n-1 characters are stored
- the input sequence reaches EOF
- the next character equals  delim, in which case the character
is not extracted
If no characters are stored, failbit is set in the stream's error
state.
In any case, a null character is stored into the next location in
the array.
unsigned char.
/*/

#-get@istream
__istream_type get(char_type* s, streamsize n)
/*-p-
@b:Simple multiple-character extraction.
@p:s  Pointer to an array.
@p:n  Maximum number of characters to store in  s.
@r:*this
Returns c get(s,n,widen('\n')).
/*/

#-get@istream
__istream_type get(__streambuf_type& sb, char_type delim)
/*-p-
@b:Extraction into another streambuf.
@p:sb  A streambuf in which to store data.
@p:delim  A "stop" character.
@r:*this
Characters are extracted and inserted into  sb until one of the
following happens:
- the input sequence reaches EOF
- insertion into the output buffer fails (in this case, the
character that would have been inserted is not extracted)
- the next character equals  delim (in this case, the character
is not extracted)
- an exception occurs (and in this case is caught)
If no characters are stored, failbit is set in the stream's error
state.
/*/

#-get@istream
__istream_type get(__streambuf_type& sb)
/*-p-
@b:Extraction into another streambuf.
@p:sb  A streambuf in which to store data.
@r:*this
Returns c get(sb,widen('\n')).
/*/

#-getline@istream
__istream_type getline(char_type* s, streamsize n, char_type delim)
/*-p-
@b:String extraction.
@p:s  A character array in which to store the data.
@p:n  Maximum number of characters to extract.
@p:delim  A "stop" character.
@r:*this
Extracts and stores characters into  s until one of the
following happens.  Note that these criteria are required to be
tested in the order listed here, to allow an input line to exactly
fill the  s array without setting failbit.
-# the input sequence reaches end-of-file, in which case eofbit
is set in the stream error state
-# the next character equals c delim, in which case the character
is extracted (and therefore counted in c gcount()) but not stored
-# c n-1 characters are stored, in which case failbit is set
in the stream error state
If no characters are extracted, failbit is set.  (An empty line of
input should therefore not cause failbit to be set.)
In any case, a null character is stored in the next location in
the array.
/*/

#-getline@istream
__istream_type getline(char_type* s, streamsize n)
/*-p-
@b:String extraction.
@p:s  A character array in which to store the data.
@p:n  Maximum number of characters to extract.
@r:*this
Returns c getline(s,n,widen('\n')).
/*/

#-ignore@istream
__istream_type ignore()
/*-p-
@b:Discarding characters
@p:n  Number of characters to discard.
@p:delim  A "stop" character.
@r:*this
Extracts characters and throws them away until one of the
following happens:
- if  n c != c std::numeric_limits<int>::max(),  n
characters are extracted
- the input sequence reaches end-of-file
- the next character equals  delim (in this case, the character
is extracted); note that this condition will never occur if
NB: Provide three overloads, instead of the single function
(with defaults) mandated by the Standard: this leads to a
better performing implementation, while still conforming to
the Standard.
/*/

#-ignore@istream
__istream_type ignore(streamsize n)
/*-p-
/*/

#-ignore@istream
__istream_type ignore(streamsize n, int_type delim)
/*-p-
/*/

#-peek@istream
int_type peek()
/*-p-
@b:Looking ahead in the stream
@r:The next character, or eof().
If, after constructing the sentry object, c good() is false,
returns c traits::eof().  Otherwise reads but does not extract
the next input character.
/*/

#-read@istream
__istream_type read(char_type* s, streamsize n)
/*-p-
@b:Extraction without delimiters.
@p:s  A character array.
@p:n  Maximum number of characters to store.
@r:*this
If the stream state is c good(), extracts characters and stores
them into  s until one of the following happens:
-  n characters are stored
- the input sequence reaches end-of-file, in which case the error
state is set to c failbit|eofbit.
unsigned char.
/*/

#-readsome@istream
streamsize readsome(char_type* s, streamsize n)
/*-p-
@b:Extraction until the buffer is exhausted, but no more.
@p:s  A character array.
@p:n  Maximum number of characters to store.
@r:The number of characters extracted.
Extracts characters and stores them into  s depending on the
number of characters remaining in the streambuf's buffer,
- if c A c == c -1, sets eofbit and extracts no characters
- if c A c == c 0, extracts no characters
- if c A c > c 0, extracts c min(A,n)
The goal is to empty the current buffer, and to not request any
more from the external input sequence controlled by the streambuf.
/*/

#-putback@istream
__istream_type putback(char_type c)
/*-p-
@b:Unextracting a single character.
@p:c  The character to push back into the input stream.
@r:*this
If c rdbuf() is not null, calls c rdbuf()->sputbackc(c).
If c rdbuf() is null or if c sputbackc() fails, sets badbit in
the error state.
/*/

#-unget@istream
__istream_type unget()
/*-p-
@b:Unextracting the previous character.
@r:*this
If c rdbuf() is not null, calls c rdbuf()->sungetc(c).
If c rdbuf() is null or if c sungetc() fails, sets badbit in
the error state.
/*/

#-sync@istream
int sync()
/*-p-
@b:Synchronizing the stream buffer.
@r:0 on success, -1 on failure
If c rdbuf() is a null pointer, returns -1.
Otherwise, calls c rdbuf()->pubsync(), and if that returns -1,
sets badbit and returns -1.
Otherwise, returns 0.
extracted, if any, and therefore does not affect the next
call to c gcount().
/*/

#-tellg@istream
pos_type tellg()
/*-p-
@b:Getting the current read position.
@r:A file position object.
If c fail() is not false, returns c pos_type(-1) to indicate
failure.  Otherwise returns c rdbuf()->pubseekoff(0,cur,in).
extracted, if any, and therefore does not affect the next
call to c gcount().
/*/

#-seekg@istream
__istream_type seekg(pos_type)
/*-p-
@b:Changing the current read position.
@p:pos  A file position object.
@r:*this
If c fail() is not true, calls c rdbuf()->pubseekpos(pos).  If
that function fails, sets failbit.
extracted, if any, and therefore does not affect the next
call to c gcount().
/*/

#-seekg@istream
__istream_type seekg(off_type, ios_base::seekdir)
/*-p-
@b:Changing the current read position.
@p:off  A file offset object.
@p:dir  The direction in which to seek.
@r:*this
If c fail() is not true, calls c rdbuf()->pubseekoff(off,dir).
If that function fails, sets failbit.
extracted, if any, and therefore does not affect the next
call to c gcount().
/*/

#-basic_istream@istream
explicit basic_istream()
/*-p-
/*/

#-getline@istream
basic_istream getline(char_type* s, streamsize n, char_type delim)
/*-p-
/*/

#-ignore@istream
basic_istream ignore(streamsize n)
/*-p-
/*/

#-ignore@istream
basic_istream ignore(streamsize n, int_type delim)
/*-p-
/*/

#-getline@istream
basic_istream getline(char_type* s, streamsize n, char_type delim)
/*-p-
/*/

#-ignore@istream
basic_istream ignore(streamsize n)
/*-p-
/*/

#-ignore@istream
basic_istream ignore(streamsize n, int_type delim)
/*-p-
/*/

##:istringstream@istringstream
istringstream
/*-p-
@b:Controlling input for std::string.
This class supports reading from objects of type std::basic_string,
using the inherited functions from std::basic_istream.  To control
the associated sequence, an instance of std::basic_stringbuf is used,
which this page refers to as c sb.
/*/

#-char_type@istringstream
_CharT char_type
/*-p-
/*/

#-traits_type@istringstream
_Traits traits_type
/*-p-
/*/

#-allocator_type@istringstream
_Alloc allocator_type
/*-p-
/*/

#-int_type@istringstream
int_type int_type
/*-p-
/*/

#-pos_type@istringstream
pos_type pos_type
/*-p-
/*/

#-off_type@istringstream
off_type off_type
/*-p-
/*/

#-basic_istringstream@istringstream
explicit basic_istringstream(ios_base::openmode mode = ios_base::in)
/*-p-
@b:Default constructor starts with an empty string buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using c mode|in, and passes c &sb to the base
class initializer.  Does not allocate any buffer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-basic_istringstream@istringstream
explicit basic_istringstream(const string_type& str, ios_base::openmode mode = ios_base::in)
/*-p-
@b:Starts with an existing string buffer.
@p:str  A string to copy as a starting buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using  str and c mode|in, and passes c &sb
to the base class initializer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-~basic_istringstream@istringstream
~basic_istringstream()
/*-p-
@b:The destructor does nothing.
The buffer is deallocated by the stringbuf object, not the
formatting stream.
/*/

#-rdbuf@istringstream
__stringbuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_stringbuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-str@istringstream
__string_type str() const
/*-p-
@b:Copying out the string buffer.
@r:c rdbuf()->str()
/*/

#-str@istringstream
void str(const string_type& s)
/*-p-
@b:Setting a new buffer.
@p:s  The string to use as a new sequence.
Calls c rdbuf()->str(s).
/*/

##:ofstream@ofstream
ofstream
/*-p-
@b:Controlling output for files.
This class supports reading from named files, using the inherited
functions from std::basic_ostream.  To control the associated
sequence, an instance of std::basic_filebuf is used, which this page
refers to as c sb.
/*/

#-char_type@ofstream
_CharT char_type
/*-p-
/*/

#-traits_type@ofstream
_Traits traits_type
/*-p-
/*/

#-int_type@ofstream
int_type int_type
/*-p-
/*/

#-pos_type@ofstream
pos_type pos_type
/*-p-
/*/

#-off_type@ofstream
off_type off_type
/*-p-
/*/

#-basic_ofstream@ofstream
basic_ofstream()
/*-p-
@b:Default constructor.
Initializes c sb using its default constructor, and passes
(you haven't given it a filename to open).
/*/

#-basic_ofstream@ofstream
explicit basic_ofstream(const char* s, ios_base::openmode mode = ios_base::out|ios_base::trunc)
/*-p-
@b:Create an output file stream.
@p:s  Null terminated string specifying the filename.
@p:mode  Open file in specified mode (see std::ios_base).
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-~basic_ofstream@ofstream
~basic_ofstream()
/*-p-
@b:The destructor does nothing.
The file is closed by the filebuf object, not the formatting
stream.
/*/

#-rdbuf@ofstream
__filebuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_filebuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-is_open@ofstream
bool is_open()
/*-p-
@b:Wrapper to test for an open file.
@r:c rdbuf()->is_open()
/*/

#-is_open@ofstream
bool is_open() const
/*-p-
/*/

#-open@ofstream
void open(const char* s, ios_base::openmode mode = ios_base::out | ios_base::trunc)
/*-p-
@b:Opens an external file.
@p:s  The name of the file.
@p:mode  The open mode flags.
Calls c std::basic_filebuf::open(s,mode|out|trunc).  If that
function fails, c failbit is set in the stream's error state.
Tip:  When using std::string to hold the filename, you must use
.c_str() before passing it to this constructor.
/*/

#-close@ofstream
void close()
/*-p-
@b:Close the file.
Calls c std::basic_filebuf::close().  If that function
fails, c failbit is set in the stream's error state.
/*/

##:ostream@ostream
ostream
/*-p-
@b:Controlling output.
This is the base class for all output streams.  It provides text
formatting of all builtin types, and communicates with any class
derived from basic_streambuf to do the actual output.
/*/

#-basic_ostream@ostream
explicit basic_ostream(__streambuf_type* sb)
/*-p-
@b:Base constructor.
This ctor is almost never called by the user directly, rather from
derived classes' initialization lists, which pass a pointer to
their own stream buffer.
/*/

#-~basic_ostream@ostream
virtual ~basic_ostream()
/*-p-
@b:Base destructor.
This does very little apart from providing a virtual base dtor.
/*/

#-put@ostream
__ostream_type put(char_type c)
/*-p-
@b:Simple insertion.
@p:c  The character to insert.
@r:*this
Tries to insert  c.
unsigned char.
/*/

#-write@ostream
__ostream_type write(const char_type* s, streamsize n)
/*-p-
@b:Character string insertion.
@p:s  The array to insert.
@p:n  Maximum number of characters to insert.
@r:*this
Characters are copied from  s and inserted into the stream until
one of the following happens:
-  n characters are inserted
- inserting into the output sequence fails (in this case, badbit
will be set in the stream's error state)
unsigned char.
/*/

#-flush@ostream
__ostream_type flush()
/*-p-
@b:Synchronizing the stream buffer.
@r:*this
If c rdbuf() is a null pointer, changes nothing.
Otherwise, calls c rdbuf()->pubsync(), and if that returns -1,
sets badbit.
/*/

#-tellp@ostream
pos_type tellp()
/*-p-
@b:Getting the current write position.
@r:A file position object.
If c fail() is not false, returns c pos_type(-1) to indicate
failure.  Otherwise returns c rdbuf()->pubseekoff(0,cur,out).
/*/

#-seekp@ostream
__ostream_type seekp(pos_type)
/*-p-
@b:Changing the current write position.
@p:pos  A file position object.
@r:*this
If c fail() is not true, calls c rdbuf()->pubseekpos(pos).  If
that function fails, sets failbit.
/*/

#-seekp@ostream
__ostream_type seekp(off_type, ios_base::seekdir)
/*-p-
@b:Changing the current write position.
@p:off  A file offset object.
@p:dir  The direction in which to seek.
@r:*this
If c fail() is not true, calls c rdbuf()->pubseekoff(off,dir).
If that function fails, sets failbit.
/*/

#-basic_ostream@ostream
explicit basic_ostream()
/*-p-
/*/

##:ostringstream@ostringstream
ostringstream
/*-p-
@b:Controlling output for std::string.
This class supports writing to objects of type std::basic_string,
using the inherited functions from std::basic_ostream.  To control
the associated sequence, an instance of std::basic_stringbuf is used,
which this page refers to as c sb.
/*/

#-char_type@ostringstream
_CharT char_type
/*-p-
/*/

#-traits_type@ostringstream
_Traits traits_type
/*-p-
/*/

#-allocator_type@ostringstream
_Alloc allocator_type
/*-p-
/*/

#-int_type@ostringstream
int_type int_type
/*-p-
/*/

#-pos_type@ostringstream
pos_type pos_type
/*-p-
/*/

#-off_type@ostringstream
off_type off_type
/*-p-
/*/

#-basic_ostringstream@ostringstream
explicit basic_ostringstream(ios_base::openmode mode = ios_base::out)
/*-p-
@b:Default constructor starts with an empty string buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using c mode|out, and passes c &sb to the base
class initializer.  Does not allocate any buffer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-basic_ostringstream@ostringstream
explicit basic_ostringstream(const string_type& str, ios_base::openmode mode = ios_base::out)
/*-p-
@b:Starts with an existing string buffer.
@p:str  A string to copy as a starting buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using  str and c mode|out, and passes c &sb
to the base class initializer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-~basic_ostringstream@ostringstream
~basic_ostringstream()
/*-p-
@b:The destructor does nothing.
The buffer is deallocated by the stringbuf object, not the
formatting stream.
/*/

#-rdbuf@ostringstream
__stringbuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_stringbuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-str@ostringstream
__string_type str() const
/*-p-
@b:Copying out the string buffer.
@r:c rdbuf()->str()
/*/

#-str@ostringstream
void str(const string_type& s)
/*-p-
@b:Setting a new buffer.
@p:s  The string to use as a new sequence.
Calls c rdbuf()->str(s).
/*/

##:streambuf@streambuf
streambuf
/*-p-
/*/

#-char_type@streambuf
_CharT char_type
/*-p-
These are standard types.  They permit a standardized way of
referring to names of (or names dependant on) the template
parameters, which are specific to the implementation.
/*/

#-find@streambuf
__type find(istreambuf_iterator<_CharT2>, istreambuf_iterator<_CharT2>, const CharT2&)
/*-p-
This is a non-standard type.
/*/

#-operator >>@streambuf
basic_istream operator >>(basic_istream<_CharT2, Traits2>&, CharT2*)
/*-p-
/*/

#-operator >>@streambuf
basic_istream operator >>(basic_istream<_CharT2, Traits2>&, basic_string<_CharT2, Traits2, Alloc>&)
/*-p-
/*/

#-getline@streambuf
basic_istream getline(basic_istream<_CharT2, Traits2>&, basic_string<_CharT2, Traits2, Alloc>&, CharT2)
/*-p-
/*/

#-~basic_streambuf@streambuf
virtual ~basic_streambuf()
/*-p-
Current locale setting.
/*/

#-pubimbue@streambuf
locale pubimbue(const locale &__loc)
/*-p-
@b:Entry point for imbue().
@p:loc  The new locale.
@r:The previous locale.
Calls the derived imbue(loc).
/*/

#-getloc@streambuf
locale getloc() const
/*-p-
@b:Locale access.
@r:The current locale in effect.
If pubimbue(loc) has been called, then the most recent c loc
is returned.  Otherwise the global locale in effect at the time
of construction is returned.
/*/

#-pubsetbuf@streambuf
__streambuf_type pubsetbuf(char_type* s, streamsize n)
/*-p-
@b:Entry points for derived buffer functions.
The public versions of c pubfoo dispatch to the protected
derived c foo member functions, passing the arguments (if any)
and returning the result unchanged.
/*/

#-pubseekoff@streambuf
pos_type pubseekoff(off_type off, ios_base::seekdir way, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

#-pubseekpos@streambuf
pos_type pubseekpos(pos_type sp, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

#-pubsync@streambuf
int pubsync()
/*-p-
/*/

#-in_avail@streambuf
streamsize in_avail()
/*-p-
@b:Looking ahead into the stream.
@r:The number of characters available.
If a read position is available, returns the number of characters
available for reading before the buffer must be refilled.
Otherwise returns the derived c showmanyc().
/*/

#-snextc@streambuf
int_type snextc()
/*-p-
@b:Getting the next character.
@r:The next character, or eof.
Calls c sbumpc(), and if that function returns
/*/

#-sbumpc@streambuf
int_type sbumpc()
/*-p-
@b:Getting the next character.
@r:The next character, or eof.
If the input read position is available, returns that character
and increments the read pointer, otherwise calls and returns
/*/

#-sgetc@streambuf
int_type sgetc()
/*-p-
@b:Getting the next character.
@r:The next character, or eof.
If the input read position is available, returns that character,
otherwise calls and returns c underflow().  Does not move the 
read position after fetching the character.
/*/

#-sgetn@streambuf
streamsize sgetn(char_type* s, streamsize n)
/*-p-
@b:Entry point for xsgetn.
@p:s  A buffer area.
@p:n  A count.
Returns xsgetn(s,n).  The effect is to fill  s[0] through
/*/

#-sputbackc@streambuf
int_type sputbackc(char_type c)
/*-p-
@b:Pushing characters back into the input stream.
@p:c  The character to push back.
@r:The previous character, if possible.
Similar to sungetc(), but  c is pushed onto the stream instead
of "the previous character".  If successful, the next character
fetched from the input stream will be  c.
/*/

#-sungetc@streambuf
int_type sungetc()
/*-p-
@b:Moving backwards in the input stream.
@r:The previous character, if possible.
If a putback position is available, this function decrements the
input pointer and returns that character.  Otherwise, calls and
returns pbackfail().  The effect is to "unget" the last character
"gotten".
/*/

#-sputc@streambuf
int_type sputc(char_type c)
/*-p-
@b:Entry point for all single-character output functions.
@p:c  A character to output.
@r:c, if possible.
One of two public output functions.
If a write position is available for the output sequence (i.e.,
the buffer is not full), stores  c in that position, increments
the position, and returns c traits::to_int_type(c).  If a write
position is not available, returns c overflow(c).
/*/

#-sputn@streambuf
streamsize sputn(const char_type* s, streamsize n)
/*-p-
@b:Entry point for all single-character output functions.
@p:s  A buffer read area.
@p:n  A count.
One of two public output functions.
Returns xsputn(s,n).  The effect is to write  s[0] through
/*/

#-basic_streambuf@streambuf
basic_streambuf()
/*-p-
@b:Base constructor.
Only called from derived constructors, and sets up all the
buffer data to zero, including the pointers described in the
basic_streambuf class description.  Note that, as a result,
- the class starts with no read nor write positions available,
- this is not an error
/*/

#-eback@streambuf
char_type eback() const
/*-p-
@b:Access to the get area.
These functions are only available to other protected functions,
including derived classes.
- eback() returns the beginning pointer for the input sequence
- gptr() returns the next pointer for the input sequence
- egptr() returns the end pointer for the input sequence
/*/

#-gptr@streambuf
char_type gptr() const
/*-p-
/*/

#-egptr@streambuf
char_type egptr() const
/*-p-
/*/

#-gbump@streambuf
void gbump(int n)
/*-p-
@b:Moving the read position.
@p:n  The delta by which to move.
This just advances the read position without returning any data.
/*/

#-setg@streambuf
void setg(char_type* gbeg, char_type* gnext, char_type* gend)
/*-p-
@b:Setting the three read area pointers.
@p:gbeg  A pointer.
@p:gnext  A pointer.
@p:gend  A pointer.
@p:gbeg == c eback(),  gnext == c gptr(), and
/*/

#-pbase@streambuf
char_type pbase() const
/*-p-
@b:Access to the put area.
These functions are only available to other protected functions,
including derived classes.
- pbase() returns the beginning pointer for the output sequence
- pptr() returns the next pointer for the output sequence
- epptr() returns the end pointer for the output sequence
/*/

#-pptr@streambuf
char_type pptr() const
/*-p-
/*/

#-epptr@streambuf
char_type epptr() const
/*-p-
/*/

#-pbump@streambuf
void pbump(int n)
/*-p-
@b:Moving the write position.
@p:n  The delta by which to move.
This just advances the write position without returning any data.
/*/

#-setp@streambuf
void setp(char_type* pbeg, char_type* pend)
/*-p-
@b:Setting the three write area pointers.
@p:pbeg  A pointer.
@p:pend  A pointer.
@p:pbeg == c pbase(),  pbeg == c pptr(), and
/*/

#-imbue@streambuf
void imbue(const locale&)
/*-p-
@b:Changes translations.
@p:loc  A new locale.
Translations done during I/O which depend on the current locale
are changed by this call.  The standard adds, "Between invocations
of this function a class derived from streambuf can safely cache
results of calls to locale functions and to members of facets
so obtained."
/*/

#-setbuf@streambuf
basic_streambuf setbuf(char_type*, streamsize)
/*-p-
@b:Maniuplates the buffer.
Each derived class provides its own appropriate behavior.  See
the next-to-last paragraph of 
http://gcc.gnu.org/onlinedocs/libstdc++/27_io/howto.html#2 for
more on this function.
/*/

#-seekoff@streambuf
pos_type seekoff(off_type, ios_base::seekdir, ios_base::openmode = ios_base::in | ios_base::out)
/*-p-
@b:Alters the stream positions.
Each derived class provides its own appropriate behavior.
that represents an invalid stream position.
/*/

#-seekpos@streambuf
pos_type seekpos(pos_type, ios_base::openmode = ios_base::in | ios_base::out)
/*-p-
@b:Alters the stream positions.
Each derived class provides its own appropriate behavior.
that represents an invalid stream position.
/*/

#-sync@streambuf
int sync()
/*-p-
@b:Synchronizes the buffer arrays with the controlled sequences.
@r:-1 on failure.
Each derived class provides its own appropriate behavior,
including the definition of "failure".
/*/

#-showmanyc@streambuf
streamsize showmanyc()
/*-p-
@b:Investigating the data available.
@r:An estimate of the number of characters available in the
input sequence, or -1.
"If it returns a positive value, then successive calls to
number of characters have been supplied.  If c showmanyc()
returns -1, then calls to c underflow() or c uflow() will fail."
[27.5.2.4.3]/1
calls [to underflow or uflow] will not return c eof() but
that they will return "immediately".
"es-how-many-see", not "show-manic".
/*/

#-xsgetn@streambuf
streamsize xsgetn(char_type* s, streamsize n)
/*-p-
@b:Multiple character extraction.
@p:s  A buffer area.
@p:n  Maximum number of characters to assign.
@r:The number of characters assigned.
Fills  s[0] through  s[n-1] with characters from the input
sequence, as if by c sbumpc().  Stops when either  n characters
have been copied, or when c traits::eof() would be copied.
It is expected that derived classes provide a more efficient
implementation by overriding this definition.
/*/

#-underflow@streambuf
int_type underflow()
/*-p-
@b:Fetches more data from the controlled sequence.
@r:The first character from the <em>pending sequence</em>.
Informally, this function is called when the input buffer is
exhausted (or does not exist, as buffering need not actually be
done).  If a buffer exists, it is "refilled".  In either case, the
next available character is returned, or c traits::eof() to
indicate a null pending sequence.
For a formal definiton of the pending sequence, see a good text
such as Langer & Kreft, or [27.5.2.4.3]/7-14.
A functioning input streambuf can be created by overriding only
this function (no buffer area will be used).  For an example, see
http://gcc.gnu.org/onlinedocs/libstdc++/27_io/howto.html#6
/*/

#-uflow@streambuf
int_type uflow()
/*-p-
@b:Fetches more data from the controlled sequence.
@r:The first character from the <em>pending sequence</em>.
Informally, this function does the same thing as c underflow(),
and in fact is required to call that function.  It also returns
the new character, like c underflow() does.  However, this
function also moves the read position forward by one.
/*/

#-pbackfail@streambuf
int_type pbackfail(int_type = traits_type::eof())
/*-p-
@b:Tries to back up the input sequence.
@p:c  The character to be inserted back into the sequence.
@p:The constraints of c gptr(), c eback(), and c pptr()
@r:eof() on failure, "some other value" on success
are the same as for c underflow().
/*/

#-xsputn@streambuf
streamsize xsputn(const char_type* s, streamsize n)
/*-p-
@b:Multiple character insertion.
@p:s  A buffer area.
@p:n  Maximum number of characters to write.
@r:The number of characters written.
Writes  s[0] through  s[n-1] to the output sequence, as if
by c sputc().  Stops when either  n characters have been
copied, or when c sputc() would return c traits::eof().
It is expected that derived classes provide a more efficient
implementation by overriding this definition.
/*/

#-overflow@streambuf
int_type overflow(int_type = traits_type::eof())
/*-p-
@b:Consumes data from the buffer; writes to the
@p:c  An additional character to consume.
@r:eof() to indicate failure, something else (usually
controlled sequence.
Informally, this function is called when the output buffer is full
(or does not exist, as buffering need not actually be done).  If a
buffer exists, it is "consumed", with "some effect" on the
controlled sequence.  (Typically, the buffer is written out to the
sequence verbatim.)  In either case, the character  c is also
written out, if  c is not c eof().
For a formal definiton of this function, see a good text
such as Langer & Kreft, or [27.5.2.4.5]/3-7.
A functioning output streambuf can be created by overriding only
this function (no buffer area will be used).
/*/

#-stossc@streambuf
void stossc()
/*-p-
@b:Tosses a character.
Advances the read pointer, ignoring the character that would have
been read.
See http://gcc.gnu.org/ml/libstdc++/2002-05/msg00168.html
must define c _GLIBCXX_DEPRECATED to make this visible; see
c++config.h.
/*/

#-basic_streambuf@streambuf
basic_streambuf(const streambuf_type& sb)
/*-p-
/*/

#-operator =@streambuf
__streambuf_type operator =(const streambuf_type&)
/*-p-
/*/

##:string@string
string
/*-p-
/*/


#-iterator@string
__normal_iterator iterator
/*-p-
/*/

#-const_iterator@string
__normal_iterator const_iterator
/*-p-
/*/

#-const_reverse_iterator@string
reverse_iterator const_reverse_iterator
/*-p-
/*/

#-reverse_iterator@string
reverse_iterator reverse_iterator
/*-p-
/*/

#-npos@string
size_type npos
/*-p-
/*/

#-basic_string@string
inline basic_string()
/*-p-
@b:Default constructor creates an empty string.
/*/

#-basic_string@string
explicit basic_string(const Alloc& a)
/*-p-
@b:Construct an empty string using allocator  a.
/*/

#-basic_string@string
basic_string(const basic_string& str)
/*-p-
@b:Construct string with copy of value of  str.
@p:str  Source string.
/*/

#-basic_string@string
basic_string(const basic_string& str, size_type pos, size_type n = npos)
/*-p-
@b:Construct string as copy of a substring.
@p:str  Source string.
@p:pos  Index of first character to copy from.
@p:n  Number of characters to copy (default remainder).
/*/

#-basic_string@string
basic_string(const basic_string& str, size_type pos, size_type n, const Alloc& a)
/*-p-
@b:Construct string as copy of a substring.
@p:str  Source string.
@p:pos  Index of first character to copy from.
@p:n  Number of characters to copy.
@p:a  Allocator to use.
/*/

#-basic_string@string
basic_string(const CharT* s, size_type n, const Alloc& a = Alloc())
/*-p-
@b:Construct string initialized by a character array.
@p:s  Source character array.
@p:n  Number of characters to copy.
@p:a  Allocator to use (default is default allocator).
NB:  s must have at least  n characters, '\0' has no special
meaning.
/*/

#-basic_string@string
basic_string(const CharT* s, const Alloc& a = Alloc())
/*-p-
@b:Construct string as copy of a C string.
@p:s  Source C string.
@p:a  Allocator to use (default is default allocator).
/*/

#-basic_string@string
basic_string(size_type n, CharT c, const Alloc& a = Alloc())
/*-p-
@b:Construct string as multiple characters.
@p:n  Number of characters.
@p:c  Character to use.
@p:a  Allocator to use (default is default allocator).
/*/

#-basic_string@string
template basic_string(_InputIterator beg, InputIterator end, const Alloc& a = Alloc())
/*-p-
@b:Construct string as copy of a range.
@p:beg  Start of range.
@p:end  End of range.
@p:a  Allocator to use (default is default allocator).
/*/

#-~basic_string@string
~basic_string()
/*-p-
@b:Destroy the string instance.
/*/

#-operator =@string
basic_string operator =(const basic_string& str)
/*-p-
@b:Assign the value of  str to this string.
@p:str  Source string.
/*/

#-operator =@string
basic_string operator =(const CharT* s)
/*-p-
@b:Copy contents of  s into this string.
@p:s  Source null-terminated string.
/*/

#-operator =@string
basic_string operator =(_CharT c)
/*-p-
@b:Set value to string of length 1.
@p:c  Source character.
Assigning to a character makes this string length 1 and
(*this)[0] ==  c.
/*/

#-begin@string
iterator begin()
/*-p-
Returns a read/write iterator that points to the first character in
the string.  Unshares the string.
/*/

#-begin@string
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the first
character in the string.
/*/

#-end@string
iterator end()
/*-p-
Returns a read/write iterator that points one past the last
character in the string.  Unshares the string.
/*/

#-end@string
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past the
last character in the string.
/*/

#-rbegin@string
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the last
character in the string.  Iteration is done in reverse element
order.  Unshares the string.
/*/

#-rbegin@string
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to the last character in the string.  Iteration is done in
reverse element order.
/*/

#-rend@string
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one before the
first character in the string.  Iteration is done in reverse
element order.  Unshares the string.
/*/

#-rend@string
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to one before the first character in the string.  Iteration
is done in reverse element order.
/*/

#-size@string
size_type size() const
/*-p-
/*/

#-length@string
size_type length() const
/*-p-
/*/

#-max_size@string
size_type max_size() const
/*-p-
/*/

#-resize@string
void resize(size_type n, CharT c)
/*-p-
@b:Resizes the string to the specified number of characters.
@p:n  Number of characters the string should contain.
@p:c  Character to fill any new elements.
This function will resize the string to the specified
number of characters.  If the number is smaller than the
string's current size the string is truncated, otherwise
the string is extended and new elements are set to  c.
/*/

#-resize@string
void resize(size_type n)
/*-p-
@b:Resizes the string to the specified number of characters.
@p:n  Number of characters the string should contain.
This function will resize the string to the specified length.  If
the new size is smaller than the string's current size the string
is truncated, otherwise the string is extended and new characters
are default-constructed.  For basic types such as char, this means
setting them to 0.
/*/

#-capacity@string
size_type capacity() const
/*-p-
Returns the total number of characters that the string can hold
before needing to allocate more memory.
/*/

#-reserve@string
void reserve(size_type res_arg = 0)
/*-p-
@b:Attempt to preallocate enough memory for specified number of
@p:res_arg  Number of characters required.
@t:std::length_error  If  res_arg exceeds c max_size().
characters.
This function attempts to reserve enough memory for the
string to hold the specified number of characters.  If the
number requested is more than max_size(), length_error is
thrown.
The advantage of this function is that if optimal code is a
necessity and the user can determine the string length that will be
required, the user can reserve the memory in dvance, and thus
prevent a possible reallocation of memory and copying of string
data.
/*/

#-clear@string
void clear()
/*-p-
Erases the string, making it empty.
/*/

#-empty@string
bool empty() const
/*-p-
Returns true if the string is empty.  Equivalent to *this == "".
/*/

#-operator []@string
const_reference operator [](size_type pos) const
/*-p-
@b:Subscript access to the data contained in the string.
@p:pos  The index of the character to access.
@r:Read-only (constant) reference to the character.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)
/*/

#-operator []@string
reference operator [](size_type pos)
/*-p-
@b:Subscript access to the data contained in the string.
@p:pos  The index of the character to access.
@r:Read/write reference to the character.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)  Unshares the string.
/*/

#-at@string
const_reference at(size_type n) const
/*-p-
@b:Provides access to the data contained in the string.
@p:n The index of the character to access.
@r:Read-only (const) reference to the character.
@t:std::out_of_range  If  n is an invalid index.
This function provides for safer data access.  The parameter is
first checked that it is in the range of the string.  The function
throws out_of_range if the check fails.
/*/

#-at@string
reference at(size_type n)
/*-p-
@b:Provides access to the data contained in the string.
@p:n The index of the character to access.
@r:Read/write reference to the character.
@t:std::out_of_range  If  n is an invalid index.
This function provides for safer data access.  The parameter is
first checked that it is in the range of the string.  The function
throws out_of_range if the check fails.  Success results in
unsharing the string.
/*/

#-operator +=@string
basic_string operator +=(const basic_string& str)
/*-p-
@b:Append a string to this string.
@p:str  The string to append.
@r:Reference to this string.
/*/

#-operator +=@string
basic_string operator +=(const CharT* s)
/*-p-
@b:Append a C string.
@p:s  The C string to append.
@r:Reference to this string.
/*/

#-operator +=@string
basic_string operator +=(_CharT c)
/*-p-
@b:Append a character.
@p:c  The character to append.
@r:Reference to this string.
/*/

#-append@string
basic_string append(const basic_string& str)
/*-p-
@b:Append a string to this string.
@p:str  The string to append.
@r:Reference to this string.
/*/

#-append@string
basic_string append(const basic_string& str, size_type pos, size_type n)
/*-p-
@b:Append a substring.
@p:str  The string to append.
@p:pos  Index of the first character of str to append.
@p:n  The number of characters to append.
@r:Reference to this string.
@t:std::out_of_range if  pos is not a valid index.
This function appends  n characters from  str starting at  pos
to this string.  If  n is is larger than the number of available
characters in  str, the remainder of  str is appended.
/*/

#-append@string
basic_string append(const CharT* s, size_type n)
/*-p-
@b:Append a C substring.
@p:s  The C string to append.
@p:n  The number of characters to append.
@r:Reference to this string.
/*/

#-append@string
basic_string append(const CharT* s)
/*-p-
@b:Append a C string.
@p:s  The C string to append.
@r:Reference to this string.
/*/

#-append@string
basic_string append(size_type n, CharT c)
/*-p-
@b:Append multiple characters.
@p:n  The number of characters to append.
@p:c  The character to use.
@r:Reference to this string.
Appends n copies of c to this string.
/*/

#-append@string
basic_string append(_InputIterator first, InputIterator last)
/*-p-
@b:Append a range of characters.
@p:first  Iterator referencing the first character to append.
@p:last  Iterator marking the end of the range.
@r:Reference to this string.
Appends characters in the range [first,last) to this string.
/*/

#-push_back@string
void push_back(_CharT c)
/*-p-
@b:Append a single character.
@p:c  Character to append.
/*/

#-assign@string
basic_string assign(const basic_string& str)
/*-p-
@b:Set value to contents of another string.
@p:str  Source string to use.
@r:Reference to this string.
/*/

#-assign@string
basic_string assign(const basic_string& str, size_type pos, size_type n)
/*-p-
@b:Set value to a substring of a string.
@p:str  The string to use.
@p:pos  Index of the first character of str.
@p:n  Number of characters to use.
@r:Reference to this string.
@t:std::out_of_range if  pos is not a valid index.
This function sets this string to the substring of  str consisting
of  n characters at  pos.  If  n is is larger than the number
of available characters in  str, the remainder of  str is used.
/*/

#-assign@string
basic_string assign(const CharT* s, size_type n)
/*-p-
@b:Set value to a C substring.
@p:s  The C string to use.
@p:n  Number of characters to use.
@r:Reference to this string.
This function sets the value of this string to the first  n
characters of  s.  If  n is is larger than the number of
available characters in  s, the remainder of  s is used.
/*/

#-assign@string
basic_string assign(const CharT* s)
/*-p-
@b:Set value to contents of a C string.
@p:s  The C string to use.
@r:Reference to this string.
This function sets the value of this string to the value of  s.
The data is copied, so there is no dependence on  s once the
function returns.
/*/

#-assign@string
basic_string assign(size_type n, CharT c)
/*-p-
@b:Set value to multiple characters.
@p:n  Length of the resulting string.
@p:c  The character to use.
@r:Reference to this string.
This function sets the value of this string to  n copies of
character  c.
/*/

#-assign@string
basic_string assign(_InputIterator first, InputIterator last)
/*-p-
@b:Set value to a range of characters.
@p:first  Iterator referencing the first character to append.
@p:last  Iterator marking the end of the range.
@r:Reference to this string.
Sets value of string to characters in the range [first,last).
/*/

#-insert@string
void insert(iterator p, size_type n, CharT c)
/*-p-
@b:Insert multiple characters.
@p:p  Iterator referencing location in string to insert at.
@p:n  Number of characters to insert
@p:c  The character to insert.
@t:std::length_error  If new length exceeds c max_size().
Inserts  n copies of character  c starting at the position
referenced by iterator  p.  If adding characters causes the length
to exceed max_size(), length_error is thrown.  The value of the
string doesn't change if an error is thrown.
/*/

#-insert@string
void insert(iterator p, InputIterator beg, InputIterator end)
/*-p-
@b:Insert a range of characters.
@p:p  Iterator referencing location in string to insert at.
@p:beg  Start of range.
@p:end  End of range.
@t:std::length_error  If new length exceeds c max_size().
Inserts characters in range [beg,end).  If adding characters causes
the length to exceed max_size(), length_error is thrown.  The value
of the string doesn't change if an error is thrown.
/*/

#-insert@string
basic_string insert(size_type pos1, const basic_string& str)
/*-p-
@b:Insert value of a string.
@p:pos1  Iterator referencing location in string to insert at.
@p:str  The string to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Inserts value of  str starting at  pos1.  If adding characters
causes the length to exceed max_size(), length_error is thrown.  The
value of the string doesn't change if an error is thrown.
/*/

#-insert@string
basic_string insert(size_type pos1, const basic_string& str, size_type pos2, size_type n)
/*-p-
@b:Insert a substring.
@p:pos1  Iterator referencing location in string to insert at.
@p:str  The string to insert.
@p:pos2  Start of characters in str to insert.
@p:n  Number of characters to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
@t:std::out_of_range  If  pos1 > size() or
Starting at  pos1, insert  n character of  str beginning with
max_size(), length_error is thrown.  If  pos1 is beyond the end of
this string or  pos2 is beyond the end of  str, out_of_range is
thrown.  The value of the string doesn't change if an error is
thrown.
/*/

#-insert@string
basic_string insert(size_type pos, const CharT* s, size_type n)
/*-p-
@b:Insert a C substring.
@p:pos  Iterator referencing location in string to insert at.
@p:s  The C string to insert.
@p:n  The number of characters to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
@t:std::out_of_range  If  pos is beyond the end of this
string.
Inserts the first  n characters of  s starting at  pos.  If
adding characters causes the length to exceed max_size(),
length_error is thrown.  If  pos is beyond end(), out_of_range is
thrown.  The value of the string doesn't change if an error is
thrown.
/*/

#-insert@string
basic_string insert(size_type pos, const CharT* s)
/*-p-
@b:Insert a C string.
@p:pos  Iterator referencing location in string to insert at.
@p:s  The C string to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
@t:std::out_of_range  If  pos is beyond the end of this
string.
Inserts the first  n characters of  s starting at  pos.  If
adding characters causes the length to exceed max_size(),
length_error is thrown.  If  pos is beyond end(), out_of_range is
thrown.  The value of the string doesn't change if an error is
thrown.
/*/

#-insert@string
basic_string insert(size_type pos, size_type n, CharT c)
/*-p-
@b:Insert multiple characters.
@p:pos  Index in string to insert at.
@p:n  Number of characters to insert
@p:c  The character to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
@t:std::out_of_range  If  pos is beyond the end of this
string.
Inserts  n copies of character  c starting at index  pos.  If
adding characters causes the length to exceed max_size(),
length_error is thrown.  If  pos > length(), out_of_range is
thrown.  The value of the string doesn't change if an error is
thrown.
/*/

#-insert@string
iterator insert(iterator p, CharT c)
/*-p-
@b:Insert one character.
@p:p  Iterator referencing position in string to insert at.
@p:c  The character to insert.
@r:Iterator referencing newly inserted char.
@t:std::length_error  If new length exceeds c max_size().
Inserts character  c at position referenced by  p.  If adding
character causes the length to exceed max_size(), length_error is
thrown.  If  p is beyond end of string, out_of_range is thrown.
The value of the string doesn't change if an error is thrown.
/*/

#-erase@string
basic_string erase(size_type pos = 0, size_type n = npos)
/*-p-
@b:Remove characters.
@p:pos  Index of first character to remove (default 0).
@p:n  Number of characters to remove (default remainder).
@r:Reference to this string.
@t:std::out_of_range  If  pos is beyond the end of this
string.
Removes  n characters from this string starting at  pos.  The
length of the string is reduced by  n.  If there are <  n
characters to remove, the remainder of the string is truncated.  If
the string doesn't change if an error is thrown.
/*/

#-erase@string
iterator erase(iterator position)
/*-p-
@b:Remove one character.
@p:position  Iterator referencing the character to remove.
@r:iterator referencing same location after removal.
Removes the character at  position from this string. The value
of the string doesn't change if an error is thrown.
/*/

#-erase@string
iterator erase(iterator first, iterator last)
/*-p-
@b:Remove a range of characters.
@p:first  Iterator referencing the first character to remove.
@p:last  Iterator referencing the end of the range.
@r:Iterator referencing location of first after removal.
Removes the characters in the range [first,last) from this string.
The value of the string doesn't change if an error is thrown.
/*/

#-replace@string
basic_string replace(size_type pos, size_type n, const basic_string& str)
/*-p-
@b:Replace characters with value from another string.
@p:pos  Index of first character to replace.
@p:n  Number of characters to be replaced.
@p:str  String to insert.
@r:Reference to this string.
@t:std::out_of_range  If  pos is beyond the end of this
@t:std::length_error  If new length exceeds c max_size().
string.
Removes the characters in the range [pos,pos+n) from this string.
In place, the value of  str is inserted.  If  pos is beyond end
of string, out_of_range is thrown.  If the length of the result
exceeds max_size(), length_error is thrown.  The value of the string
doesn't change if an error is thrown.
/*/

#-replace@string
basic_string replace(size_type pos1, size_type n1, const basic_string& str, size_type pos2, size_type n2)
/*-p-
@b:Replace characters with value from another string.
@p:pos1  Index of first character to replace.
@p:n1  Number of characters to be replaced.
@p:str  String to insert.
@p:pos2  Index of first character of str to use.
@p:n2  Number of characters from str to use.
@r:Reference to this string.
@t:std::out_of_range  If  pos1 > size() or  pos2 >
@t:std::length_error  If new length exceeds c max_size().
str.size().
Removes the characters in the range [pos1,pos1 + n) from this
string.  In place, the value of  str is inserted.  If  pos is
beyond end of string, out_of_range is thrown.  If the length of the
result exceeds max_size(), length_error is thrown.  The value of the
string doesn't change if an error is thrown.
/*/

#-replace@string
basic_string replace(size_type pos, size_type n1, const CharT* s, size_type n2)
/*-p-
@b:Replace characters with value of a C substring.
@p:pos  Index of first character to replace.
@p:n1  Number of characters to be replaced.
@p:s  C string to insert.
@p:n2  Number of characters from  s to use.
@r:Reference to this string.
@t:std::out_of_range  If  pos1 > size().
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [pos,pos + n1) from this string.
In place, the first  n2 characters of  s are inserted, or all
of  s if  n2 is too large.  If  pos is beyond end of string,
out_of_range is thrown.  If the length of result exceeds max_size(),
length_error is thrown.  The value of the string doesn't change if
an error is thrown.
/*/

#-replace@string
basic_string replace(size_type pos, size_type n1, const CharT* s)
/*-p-
@b:Replace characters with value of a C string.
@p:pos  Index of first character to replace.
@p:n1  Number of characters to be replaced.
@p:s  C string to insert.
@r:Reference to this string.
@t:std::out_of_range  If  pos > size().
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [pos,pos + n1) from this string.
In place, the first  n characters of  s are inserted.  If 
pos is beyond end of string, out_of_range is thrown.  If the length
of result exceeds max_size(), length_error is thrown.  The value of
the string doesn't change if an error is thrown.
/*/

#-replace@string
basic_string replace(size_type pos, size_type n1, size_type n2, CharT c)
/*-p-
@b:Replace characters with multiple characters.
@p:pos  Index of first character to replace.
@p:n1  Number of characters to be replaced.
@p:n2  Number of characters to insert.
@p:c  Character to insert.
@r:Reference to this string.
@t:std::out_of_range  If  pos > size().
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [pos,pos + n1) from this string.
In place,  n2 copies of  c are inserted.  If  pos is beyond
end of string, out_of_range is thrown.  If the length of result
exceeds max_size(), length_error is thrown.  The value of the string
doesn't change if an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, const basic_string& str)
/*-p-
@b:Replace range of characters with string.
@p:i1  Iterator referencing start of range to replace.
@p:i2  Iterator referencing end of range to replace.
@p:str  String value to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [i1,i2).  In place, the value of
length_error is thrown.  The value of the string doesn't change if
an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, const CharT* s, size_type n)
/*-p-
@b:Replace range of characters with C substring.
@p:i1  Iterator referencing start of range to replace.
@p:i2  Iterator referencing end of range to replace.
@p:s  C string value to insert.
@p:n  Number of characters from s to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [i1,i2).  In place, the first 
n characters of  s are inserted.  If the length of result exceeds
max_size(), length_error is thrown.  The value of the string doesn't
change if an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, const CharT* s)
/*-p-
@b:Replace range of characters with C string.
@p:i1  Iterator referencing start of range to replace.
@p:i2  Iterator referencing end of range to replace.
@p:s  C string value to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [i1,i2).  In place, the
characters of  s are inserted.  If the length of result exceeds
max_size(), length_error is thrown.  The value of the string doesn't
change if an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, size_type n, CharT c)
/*-p-
@b:Replace range of characters with multiple characters
@p:i1  Iterator referencing start of range to replace.
@p:i2  Iterator referencing end of range to replace.
@p:n  Number of characters to insert.
@p:c  Character to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [i1,i2).  In place,  n copies
of  c are inserted.  If the length of result exceeds max_size(),
length_error is thrown.  The value of the string doesn't change if
an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, InputIterator k1, InputIterator k2)
/*-p-
@b:Replace range of characters with range.
@p:i1  Iterator referencing start of range to replace.
@p:i2  Iterator referencing end of range to replace.
@p:k1  Iterator referencing start of range to insert.
@p:k2  Iterator referencing end of range to insert.
@r:Reference to this string.
@t:std::length_error  If new length exceeds c max_size().
Removes the characters in the range [i1,i2).  In place, characters
in the range [k1,k2) are inserted.  If the length of result exceeds
max_size(), length_error is thrown.  The value of the string doesn't
change if an error is thrown.
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, CharT* k1, CharT* k2)
/*-p-
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, const CharT* k1, const CharT* k2)
/*-p-
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, iterator k1, iterator k2)
/*-p-
/*/

#-replace@string
basic_string replace(iterator i1, iterator i2, const_iterator k1, const_iterator k2)
/*-p-
/*/

#-copy@string
size_type copy(_CharT* s, size_type n, size_type pos = 0) const
/*-p-
@b:Copy substring into C string.
@p:s  C string to copy value into.
@p:n  Number of characters to copy.
@p:pos  Index of first character to copy.
@r:Number of characters actually copied
@t:std::out_of_range  If pos > size().
Copies up to  n characters starting at  pos into the C string 
s.  If  pos is greater than size(), out_of_range is thrown.
/*/

#-swap@string
void swap(basic_string& s)
/*-p-
@b:Swap contents with another string.
@p:s  String to swap with.
Exchanges the contents of this string with that of  s in constant
time.
/*/

#-c_str@string
_CharT c_str() const
/*-p-
@b:Return const pointer to null-terminated contents.
This is a handle to internal data.  Do not modify or dire things may
happen.
/*/

#-data@string
_CharT data() const
/*-p-
@b:Return const pointer to contents.
This is a handle to internal data.  Do not modify or dire things may
happen.
/*/

#-get_allocator@string
allocator_type get_allocator() const
/*-p-
@b:Return copy of allocator used to construct this string.
/*/

#-find@string
size_type find(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find position of a C substring.
@p:s  C string to locate.
@p:pos  Index of character to search from.
@p:n  Number of characters from  s to search for.
@r:Index of start of first occurrence.
Starting from  pos, searches forward for the first  n characters
in  s within this string.  If found, returns the index where it
begins.  If not found, returns npos.
/*/

#-find@string
size_type find(const basic_string& str, size_type pos = 0) const
/*-p-
@b:Find position of a string.
@p:str  String to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of start of first occurrence.
Starting from  pos, searches forward for value of  str within
this string.  If found, returns the index where it begins.  If not
found, returns npos.
/*/

#-find@string
size_type find(const CharT* s, size_type pos = 0) const
/*-p-
@b:Find position of a C string.
@p:s  C string to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of start of first occurrence.
Starting from  pos, searches forward for the value of  s within
this string.  If found, returns the index where it begins.  If not
found, returns npos.
/*/

#-find@string
size_type find(_CharT c, size_type pos = 0) const
/*-p-
@b:Find position of a character.
@p:c  Character to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for  c within this string.
If found, returns the index where it was found.  If not found,
returns npos.
/*/

#-rfind@string
size_type rfind(const basic_string& str, size_type pos = npos) const
/*-p-
@b:Find last position of a string.
@p:str  String to locate.
@p:pos  Index of character to search back from (default end).
@r:Index of start of last occurrence.
Starting from  pos, searches backward for value of  str within
this string.  If found, returns the index where it begins.  If not
found, returns npos.
/*/

#-rfind@string
size_type rfind(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find last position of a C substring.
@p:s  C string to locate.
@p:pos  Index of character to search back from.
@p:n  Number of characters from s to search for.
@r:Index of start of last occurrence.
Starting from  pos, searches backward for the first  n
characters in  s within this string.  If found, returns the index
where it begins.  If not found, returns npos.
/*/

#-rfind@string
size_type rfind(const CharT* s, size_type pos = npos) const
/*-p-
@b:Find last position of a C string.
@p:s  C string to locate.
@p:pos  Index of character to start search at (default end).
@r:Index of start of  last occurrence.
Starting from  pos, searches backward for the value of  s within
this string.  If found, returns the index where it begins.  If not
found, returns npos.
/*/

#-rfind@string
size_type rfind(_CharT c, size_type pos = npos) const
/*-p-
@b:Find last position of a character.
@p:c  Character to locate.
@p:pos  Index of character to search back from (default end).
@r:Index of last occurrence.
Starting from  pos, searches backward for  c within this string.
If found, returns the index where it was found.  If not found,
returns npos.
/*/

#-find_first_of@string
size_type find_first_of(const basic_string& str, size_type pos = 0) const
/*-p-
@b:Find position of a character of string.
@p:str  String containing characters to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for one of the characters of
found.  If not found, returns npos.
/*/

#-find_first_of@string
size_type find_first_of(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find position of a character of C substring.
@p:s  String containing characters to locate.
@p:pos  Index of character to search from (default 0).
@p:n  Number of characters from s to search for.
@r:Index of first occurrence.
Starting from  pos, searches forward for one of the first  n
characters of  s within this string.  If found, returns the index
where it was found.  If not found, returns npos.
/*/

#-find_first_of@string
size_type find_first_of(const CharT* s, size_type pos = 0) const
/*-p-
@b:Find position of a character of C string.
@p:s  String containing characters to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for one of the characters of
found.  If not found, returns npos.
/*/

#-find_first_of@string
size_type find_first_of(_CharT c, size_type pos = 0) const
/*-p-
@b:Find position of a character.
@p:c  Character to locate.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for the character  c within
this string.  If found, returns the index where it was found.  If
not found, returns npos.
Note: equivalent to find(c, pos).
/*/

#-find_last_of@string
size_type find_last_of(const basic_string& str, size_type pos = npos) const
/*-p-
@b:Find last position of a character of string.
@p:str  String containing characters to locate.
@p:pos  Index of character to search back from (default end).
@r:Index of last occurrence.
Starting from  pos, searches backward for one of the characters of
found.  If not found, returns npos.
/*/

#-find_last_of@string
size_type find_last_of(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find last position of a character of C substring.
@p:s  C string containing characters to locate.
@p:pos  Index of character to search back from (default end).
@p:n  Number of characters from s to search for.
@r:Index of last occurrence.
Starting from  pos, searches backward for one of the first  n
characters of  s within this string.  If found, returns the index
where it was found.  If not found, returns npos.
/*/

#-find_last_of@string
size_type find_last_of(const CharT* s, size_type pos = npos) const
/*-p-
@b:Find last position of a character of C string.
@p:s  C string containing characters to locate.
@p:pos  Index of character to search back from (default end).
@r:Index of last occurrence.
Starting from  pos, searches backward for one of the characters of
found.  If not found, returns npos.
/*/

#-find_last_of@string
size_type find_last_of(_CharT c, size_type pos = npos) const
/*-p-
@b:Find last position of a character.
@p:c  Character to locate.
@p:pos  Index of character to search back from (default 0).
@r:Index of last occurrence.
Starting from  pos, searches backward for  c within this string.
If found, returns the index where it was found.  If not found,
returns npos.
Note: equivalent to rfind(c, pos).
/*/

#-find_first_not_of@string
size_type find_first_not_of(const basic_string& str, size_type pos = 0) const
/*-p-
@b:Find position of a character not in string.
@p:str  String containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for a character not contained
in  str within this string.  If found, returns the index where it
was found.  If not found, returns npos.
/*/

#-find_first_not_of@string
size_type find_first_not_of(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find position of a character not in C substring.
@p:s  C string containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@p:n  Number of characters from s to consider.
@r:Index of first occurrence.
Starting from  pos, searches forward for a character not contained
in the first  n characters of  s within this string.  If found,
returns the index where it was found.  If not found, returns npos.
/*/

#-find_first_not_of@string
size_type find_first_not_of(const CharT* s, size_type pos = 0) const
/*-p-
@b:Find position of a character not in C string.
@p:s  C string containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for a character not contained
in  s within this string.  If found, returns the index where it
was found.  If not found, returns npos.
/*/

#-find_first_not_of@string
size_type find_first_not_of(_CharT c, size_type pos = 0) const
/*-p-
@b:Find position of a different character.
@p:c  Character to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches forward for a character other than  c
within this string.  If found, returns the index where it was found.
If not found, returns npos.
/*/

#-find_last_not_of@string
size_type find_last_not_of(const basic_string& str, size_type pos = npos) const
/*-p-
@b:Find last position of a character not in string.
@p:str  String containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches backward for a character not
contained in  str within this string.  If found, returns the index
where it was found.  If not found, returns npos.
/*/

#-find_last_not_of@string
size_type find_last_not_of(const CharT* s, size_type pos, size_type n) const
/*-p-
@b:Find last position of a character not in C substring.
@p:s  C string containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@p:n  Number of characters from s to consider.
@r:Index of first occurrence.
Starting from  pos, searches backward for a character not
contained in the first  n characters of  s within this string.
If found, returns the index where it was found.  If not found,
returns npos.
/*/

#-find_last_not_of@string
size_type find_last_not_of(const CharT* s, size_type pos = npos) const
/*-p-
@b:Find position of a character not in C string.
@p:s  C string containing characters to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches backward for a character not
contained in  s within this string.  If found, returns the index
where it was found.  If not found, returns npos.
/*/

#-find_last_not_of@string
size_type find_last_not_of(_CharT c, size_type pos = npos) const
/*-p-
@b:Find last position of a different character.
@p:c  Character to avoid.
@p:pos  Index of character to search from (default 0).
@r:Index of first occurrence.
Starting from  pos, searches backward for a character other than
found.  If not found, returns npos.
/*/

#-substr@string
basic_string substr(size_type pos = 0, size_type n = npos) const
/*-p-
@b:Get a substring.
@p:pos  Index of first character (default 0).
@p:n  Number of characters in substring (default remainder).
@r:The new string.
@t:std::out_of_range  If pos > size().
Construct and return a new string using the  n characters starting
at  pos.  If the string is too short, use the remainder of the
characters.  If  pos is beyond the end of the string, out_of_range
is thrown.
/*/

#-compare@string
int compare(const basic_string& str) const
/*-p-
@b:Compare to a string.
@p:str  String to compare against.
@r:Integer < 0, 0, or > 0.
Returns an integer < 0 if this string is ordered before  str, 0 if
their values are equivalent, or > 0 if this string is ordered after
compare as the smallest of size() and str.size().  The function
then compares the two strings by calling traits::compare(data(),
str.data(),rlen).  If the result of the comparison is nonzero returns
it, otherwise the shorter one is ordered first.
/*/

#-compare@string
int compare(size_type pos, size_type n, const basic_string& str) const
/*-p-
@b:Compare substring to a string.
@p:pos  Index of first character of substring.
@p:n  Number of characters in substring.
@p:str  String to compare against.
@r:Integer < 0, 0, or > 0.
Form the substring of this string from the  n characters starting
at  pos.  Returns an integer < 0 if the substring is ordered
before  str, 0 if their values are equivalent, or > 0 if the
substring is ordered after  str.  Determines the effective length
rlen of the strings to compare as the smallest of the length of the
substring and  str.size().  The function then compares the two
strings by calling traits::compare(substring.data(),str.data(),rlen).
If the result of the comparison is nonzero returns it, otherwise the
shorter one is ordered first.
/*/

#-compare@string
int compare(size_type pos1, size_type n1, const basic_string& str, size_type pos2, size_type n2) const
/*-p-
@b:Compare substring to a substring.
@p:pos1  Index of first character of substring.
@p:n1  Number of characters in substring.
@p:str  String to compare against.
@p:pos2  Index of first character of substring of str.
@p:n2  Number of characters in substring of str.
@r:Integer < 0, 0, or > 0.
Form the substring of this string from the  n1 characters starting
at  pos1.  Form the substring of  str from the  n2 characters
starting at  pos2.  Returns an integer < 0 if this substring is
ordered before the substring of  str, 0 if their values are
equivalent, or > 0 if this substring is ordered after the substring
of  str.  Determines the effective length rlen of the strings
to compare as the smallest of the lengths of the substrings.  The
function then compares the two strings by calling
traits::compare(substring.data(),str.substr(pos2,n2).data(),rlen).
If the result of the comparison is nonzero returns it, otherwise the
shorter one is ordered first.
/*/

#-compare@string
int compare(const CharT* s) const
/*-p-
@b:Compare to a C string.
@p:s  C string to compare against.
@r:Integer < 0, 0, or > 0.
Returns an integer < 0 if this string is ordered before  s, 0 if
their values are equivalent, or > 0 if this string is ordered after
compare as the smallest of size() and the length of a string
constructed from  s.  The function then compares the two strings
by calling traits::compare(data(),s,rlen).  If the result of the
comparison is nonzero returns it, otherwise the shorter one is
ordered first.
/*/

#-compare@string
int compare(size_type pos, size_type n1, const CharT* s) const
/*-p-
@b:Compare substring to a C string.
@p:pos  Index of first character of substring.
@p:n1  Number of characters in substring.
@p:s  C string to compare against.
@r:Integer < 0, 0, or > 0.
Form the substring of this string from the  n1 characters starting
at  pos.  Returns an integer < 0 if the substring is ordered
before  s, 0 if their values are equivalent, or > 0 if the
substring is ordered after  s.  Determines the effective length
rlen of the strings to compare as the smallest of the length of the 
substring and the length of a string constructed from  s.  The
function then compares the two string by calling
traits::compare(substring.data(),s,rlen).  If the result of the
comparison is nonzero returns it, otherwise the shorter one is
ordered first.
/*/

#-compare@string
int compare(size_type pos, size_type n1, const CharT* s, size_type n2) const
/*-p-
@b:Compare substring against a character array.
@p:pos1  Index of first character of substring.
@p:n1  Number of characters in substring.
@p:s  character array to compare against.
@p:n2  Number of characters of s.
@r:Integer < 0, 0, or > 0.
Form the substring of this string from the  n1 characters starting
at  pos1.  Form a string from the first  n2 characters of  s.
Returns an integer < 0 if this substring is ordered before the string
from  s, 0 if their values are equivalent, or > 0 if this substring
is ordered after the string from  s.   Determines the effective
length rlen of the strings to compare as the smallest of the length
of the substring and  n2.  The function then compares the two
strings by calling traits::compare(substring.data(),s,rlen).  If the
result of the comparison is nonzero returns it, otherwise the shorter
one is ordered first.
NB: s must have at least n2 characters, '\0' has no special
meaning.
/*/

#-basic_string@string
inline basic_string()
/*-p-
/*/

##:stringbuf@stringbuf
stringbuf
/*-p-
@b:The actual work of input and output (for std::string).
This class associates either or both of its input and output sequences
with a sequence of characters, which can be initialized from, or made
available as, a c std::basic_string.  (Paraphrased from [27.7.1]/1.)
For this class, open modes (of type c ios_base::openmode) have
output sequence can be written.
/*/

#-char_type@stringbuf
_CharT char_type
/*-p-
/*/

#-traits_type@stringbuf
_Traits traits_type
/*-p-
/*/

#-allocator_type@stringbuf
_Alloc allocator_type
/*-p-
/*/

#-int_type@stringbuf
int_type int_type
/*-p-
/*/

#-pos_type@stringbuf
pos_type pos_type
/*-p-
/*/

#-off_type@stringbuf
off_type off_type
/*-p-
/*/

#-basic_stringbuf@stringbuf
explicit basic_stringbuf(ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
@b:Starts with an empty string buffer.
@p:mode  Whether the buffer can read, or write, or both.
The default constructor initializes the parent class using its
own default ctor.
/*/

#-basic_stringbuf@stringbuf
explicit basic_stringbuf(const string_type& str, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
@b:Starts with an existing string buffer.
@p:str  A string to copy as a starting buffer.
@p:mode  Whether the buffer can read, or write, or both.
This constructor initializes the parent class using its
own default ctor.
/*/

#-str@stringbuf
__string_type str() const
/*-p-
@b:Copying out the string buffer.
@r:A copy of one of the underlying sequences.
"If the buffer is only created in input mode, the underlying
character sequence is equal to the input sequence; otherwise, it
is equal to the output sequence." [27.7.1.2]/1
/*/

#-str@stringbuf
void str(const string_type& s)
/*-p-
@b:Setting a new buffer.
@p:s  The string to use as a new sequence.
Deallocates any previous stored sequence, then copies  s to
use as a new one.
/*/

#-showmanyc@stringbuf
streamsize showmanyc()
/*-p-
/*/

#-underflow@stringbuf
int_type underflow()
/*-p-
/*/

#-pbackfail@stringbuf
int_type pbackfail(int_type c = traits_type::eof())
/*-p-
/*/

#-overflow@stringbuf
int_type overflow(int_type c = traits_type::eof())
/*-p-
/*/

#-setbuf@stringbuf
__streambuf_type setbuf(char_type* s, streamsize n)
/*-p-
@b:Manipulates the buffer.
@p:s  Pointer to a buffer area.
@p:n  Size of  s.
@r:c this
If no buffer has already been created, and both  s and  n are
non-zero, then c s is used as a buffer; see
http://gcc.gnu.org/onlinedocs/libstdc++/27_io/howto.html#2
for more.
/*/

#-seekoff@stringbuf
pos_type seekoff(off_type off, ios_base::seekdir way, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

#-seekpos@stringbuf
pos_type seekpos(pos_type sp, ios_base::openmode mode = ios_base::in | ios_base::out)
/*-p-
/*/

##:stringstream@stringstream
stringstream
/*-p-
@b:Controlling input and output for std::string.
This class supports reading from and writing to objects of type
std::basic_string, using the inherited functions from
std::basic_iostream.  To control the associated sequence, an instance
of std::basic_stringbuf is used, which this page refers to as c sb.
/*/

#-char_type@stringstream
_CharT char_type
/*-p-
/*/

#-traits_type@stringstream
_Traits traits_type
/*-p-
/*/

#-allocator_type@stringstream
_Alloc allocator_type
/*-p-
/*/

#-int_type@stringstream
int_type int_type
/*-p-
/*/

#-pos_type@stringstream
pos_type pos_type
/*-p-
/*/

#-off_type@stringstream
off_type off_type
/*-p-
/*/

#-basic_stringstream@stringstream
explicit basic_stringstream(ios_base::openmode m = ios_base::out | ios_base::in)
/*-p-
@b:Default constructor starts with an empty string buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using c mode, and passes c &sb to the base
class initializer.  Does not allocate any buffer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-basic_stringstream@stringstream
explicit basic_stringstream(const string_type& str, ios_base::openmode m = ios_base::out | ios_base::in)
/*-p-
@b:Starts with an existing string buffer.
@p:str  A string to copy as a starting buffer.
@p:mode  Whether the buffer can read, or write, or both.
Initializes c sb using  str and c mode, and passes c &sb
to the base class initializer.
That's a lie.  We initialize the base class with NULL, because the
string class does its own memory management.
/*/

#-~basic_stringstream@stringstream
~basic_stringstream()
/*-p-
@b:The destructor does nothing.
The buffer is deallocated by the stringbuf object, not the
formatting stream.
/*/

#-rdbuf@stringstream
__stringbuf_type rdbuf() const
/*-p-
@b:Accessing the underlying buffer.
@r:The current basic_stringbuf buffer.
This hides both signatures of std::basic_ios::rdbuf().
/*/

#-str@stringstream
__string_type str() const
/*-p-
@b:Copying out the string buffer.
@r:c rdbuf()->str()
/*/

#-str@stringstream
void str(const string_type& s)
/*-p-
@b:Setting a new buffer.
@p:s  The string to use as a new sequence.
Calls c rdbuf()->str(s).
/*/

##:bitset@bitset
bitset
/*-p-
/*/

#-bitset@bitset
bitset()
/*-p-
This encapsulates the concept of a single bit.  An instance of this
class is a proxy for an actual bit; this way the individual bit
operations are done as faster word-size bitwise instructions.
Most users will never need to use this class directly; conversions
to and from bool are automatic and should be transparent.  Overloaded
operators help to preserve the illusion.
(On a typical system, this "bit reference" is 64 times the size of
an actual bit.  Ha.)
/*/

#-bitset@bitset
bitset(unsigned long val)
/*-p-
/*/

#-bitset@bitset
explicit bitset(const std::basic_string<_CharT, Traits, Alloc>& s, size_t position = 0)
/*-p-
@b:Use a subset of a string.
@p:s  A string of '0' and '1' characters.
@p:position  Index of the first character in  s to use;
@t:std::out_of_range  If  pos is bigger the size of  s.
@t:std::invalid_argument  If a character appears in the string
defaults to zero.
which is neither '0' nor '1'.
/*/

#-bitset@bitset
template bitset(const std::basic_string<_CharT, Traits, Alloc>& s, size_t position, size_t n)
/*-p-
@b:Use a subset of a string.
@p:s  A string of '0' and '1' characters.
@p:position  Index of the first character in  s to use.
@p:n    The number of characters to copy.
@t:std::out_of_range  If  pos is bigger the size of  s.
@t:std::invalid_argument  If a character appears in the string
which is neither '0' nor '1'.
/*/

#-operator &=@bitset
bitset operator &=(const bitset<_Nb>& rhs)
/*-p-
@b:Operations on bitsets.
@p:rhs  A same-sized bitset.
These should be self-explanatory.
/*/

#-operator |=@bitset
bitset operator |=(const bitset<_Nb>& rhs)
/*-p-
/*/

#-operator ^=@bitset
bitset operator ^=(const bitset<_Nb>& rhs)
/*-p-
/*/

#-operator <<=@bitset
bitset operator <<=(size_t position)
/*-p-
@b:Operations on bitsets.
@p:position  The number of places to shift.
These should be self-explanatory.
/*/

#-operator >>=@bitset
bitset operator >>=(size_t position)
/*-p-
/*/

#-set@bitset
bitset set()
/*-p-
@b:Sets every bit to true.
/*/

#-set@bitset
bitset set(size_t position, bool val = true)
/*-p-
@b:Sets a given bit to a particular value.
@p:position  The index of the bit.
@p:val  Either true or false, defaults to true.
@t:std::out_of_range  If  pos is bigger the size of the set.
/*/

#-reset@bitset
bitset reset()
/*-p-
@b:Sets every bit to false.
/*/

#-reset@bitset
bitset reset(size_t position)
/*-p-
@b:Sets a given bit to false.
@p:position  The index of the bit.
@t:std::out_of_range  If  pos is bigger the size of the set.
Same as writing c set(pos,false).
/*/

#-flip@bitset
bitset flip()
/*-p-
@b:Toggles every bit to its opposite value.
/*/

#-flip@bitset
bitset flip(size_t position)
/*-p-
@b:Toggles a given bit to its opposite value.
@p:position  The index of the bit.
@t:std::out_of_range  If  pos is bigger the size of the set.
/*/

#-operator ~@bitset
bitset operator ~() const
/*-p-
/*/

#-operator []@bitset
reference operator [](size_t position)
/*-p-
@b:Array-indexing support.
@p:position  Index into the bitset.
@r:A bool for a 'const bitset'.  For non-const bitsets, an
instance of the reference proxy class.
as required by DR 11 to the standard.
_GLIBCXX_RESOLVE_LIB_DEFECTS Note that this implementation already
resolves DR 11 (items 1 and 2), but does not do the range-checking
required by that DR's resolution.  -pme
The DR has since been changed:  range-checking is a precondition
(users' responsibility), and these functions must not throw.  -pme
/*/

#-operator []@bitset
bool operator [](size_t position) const
/*-p-
/*/

#-to_ulong@bitset
long to_ulong() const
/*-p-
@b:Retuns a numerical interpretation of the bitset.
@r:The integral equivalent of the bits.
@t:std::overflow_error  If there are too many bits to be
represented in an c unsigned c long.
/*/

#-to_string@bitset
basic_string to_string() const
/*-p-
@b:Retuns a character interpretation of the bitset.
@r:The string equivalent of the bits.
Note the ordering of the bits:  decreasing character positions
correspond to increasing bit positions (see the main class notes for
an example).
/*/

#-to_string@bitset
basic_string to_string() const
/*-p-
/*/

#-to_string@bitset
basic_string to_string() const
/*-p-
/*/

#-to_string@bitset
basic_string to_string() const
/*-p-
/*/

#-count@bitset
size_t count() const
/*-p-
/*/

#-size@bitset
size_t size() const
/*-p-
/*/

#-operator ==@bitset
bool operator ==(const bitset<_Nb>& rhs) const
/*-p-
/*/

#-operator !=@bitset
bool operator !=(const bitset<_Nb>& rhs) const
/*-p-
/*/

#-test@bitset
bool test(size_t position) const
/*-p-
@b:Tests the value of a bit.
@p:position  The index of a bit.
@r:The value at  pos.
@t:std::out_of_range  If  pos is bigger the size of the set.
/*/

#-any@bitset
bool any() const
/*-p-
@b:Tests whether any of the bits are on.
@r:True if at least one bit is set.
/*/

#-none@bitset
bool none() const
/*-p-
@b:Tests whether any of the bits are on.
@r:True if none of the bits are set.
/*/

#-operator <<@bitset
bitset operator <<(size_t position) const
/*-p-
/*/

#-operator >>@bitset
bitset operator >>(size_t position) const
/*-p-
/*/

#-set@bitset
set(size_t, bool)
/*-p-
@b:Global I/O operators for bitsets.
Direct I/O between streams and bitsets is supported.  Output is
straightforward.  Input will skip whitespace, only accept '0' and '1'
characters, and will only extract as many digits as the bitset will
hold.
/*/

#-reset@bitset
reset(size_t)
/*-p-
/*/

#-flip@bitset
flip(size_t)
/*-p-
/*/

#-test@bitset
test(size_t) const
/*-p-
/*/

##:complex@complex
complex
/*-p-
/*/

#-value_type@complex
_Tp value_type
/*-p-
/*/

#-complex@complex
complex(const Tp& = Tp(), const Tp & = Tp())
/*-p-
/*/

#-complex@complex
template complex(const complex<_Up>&)
/*-p-
/*/

#-real@complex
_Tp real()
/*-p-
/*/

#-real@complex
_Tp real() const
/*-p-
/*/

#-imag@complex
_Tp imag()
/*-p-
/*/

#-imag@complex
_Tp imag() const
/*-p-
/*/

#-operator =@complex
complex operator =(const Tp&)
/*-p-
/*/

#-operator +=@complex
complex operator +=(const Tp&)
/*-p-
/*/

#-operator -=@complex
complex operator -=(const Tp&)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const Tp&)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const Tp&)
/*-p-
/*/

#-operator =@complex
complex operator =(const complex<_Up>&)
/*-p-
/*/

#-operator +=@complex
complex operator +=(const complex<_Up>&)
/*-p-
/*/

#-operator -=@complex
complex operator -=(const complex<_Up>&)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const complex<_Up>&)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const complex<_Up>&)
/*-p-
/*/

#-real@complex
real()
/*-p-
/*/

#-real@complex
real() const
/*-p-
/*/

#-imag@complex
imag()
/*-p-
/*/

#-imag@complex
imag() const
/*-p-
/*/

#-complex@complex
inline complex(const Tp& r, const Tp& i)
/*-p-
/*/

#-complex@complex
complex(const complex<_Up>& z)
/*-p-
/*/

#-operator =@complex
complex operator =(const Tp& t)
/*-p-
/*/

#-operator +=@complex
operator +=(const Tp& t)
/*-p-
/*/

#-operator -=@complex
operator -=(const Tp& t)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const Tp& t)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const Tp& t)
/*-p-
/*/

#-operator =@complex
operator =(const complex<_Up>& z)
/*-p-
/*/

#-operator +=@complex
operator +=(const complex<_Up>& z)
/*-p-
/*/

#-operator -=@complex
operator -=(const complex<_Up>& z)
/*-p-
/*/

#-operator *=@complex
operator *=(const complex<_Up>& z)
/*-p-
/*/

#-operator /=@complex
operator /=(const complex<_Up>& z)
/*-p-
/*/

#-value_type@complex
float value_type
/*-p-
/*/

#-complex@complex
complex(_ComplexT z)
/*-p-
/*/

#-complex@complex
complex(float = 0.0f, float = 0.0f)
/*-p-
/*/

#-complex@complex
explicit complex(const complex<double>&)
/*-p-
/*/

#-complex@complex
explicit complex(const complex<long double>&)
/*-p-
/*/

#-real@complex
float real()
/*-p-
/*/

#-real@complex
float real() const
/*-p-
/*/

#-imag@complex
float imag()
/*-p-
/*/

#-imag@complex
float imag() const
/*-p-
/*/

#-operator =@complex
complex operator =(float)
/*-p-
/*/

#-operator +=@complex
complex operator +=(float)
/*-p-
/*/

#-operator -=@complex
complex operator -=(float)
/*-p-
/*/

#-operator *=@complex
complex operator *=(float)
/*-p-
/*/

#-operator /=@complex
complex operator /=(float)
/*-p-
/*/

#-operator =@complex
complex operator =(const complex<_Tp>&)
/*-p-
/*/

#-operator +=@complex
complex operator +=(const complex<_Tp>&)
/*-p-
/*/

#-operator -=@complex
complex operator -=(const complex<_Tp>&)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const complex<_Tp>&)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const complex<_Tp>&)
/*-p-
/*/

#-real@complex
float real()
/*-p-
/*/

#-real@complex
real() const
/*-p-
/*/

#-imag@complex
float imag()
/*-p-
/*/

#-imag@complex
imag() const
/*-p-
/*/

#-complex@complex
inline complex(float r, float i)
/*-p-
/*/

#-operator =@complex
complex operator =(float f)
/*-p-
/*/

#-operator +=@complex
complex operator +=(float f)
/*-p-
/*/

#-operator -=@complex
complex operator -=(float f)
/*-p-
/*/

#-operator *=@complex
complex operator *=(float f)
/*-p-
/*/

#-operator /=@complex
complex operator /=(float f)
/*-p-
/*/

#-operator =@complex
operator =(const complex<_Tp>& z)
/*-p-
/*/

#-operator +=@complex
operator +=(const complex<_Tp>& z)
/*-p-
/*/

#-operator -=@complex
operator -=(const complex<_Tp>& z)
/*-p-
/*/

#-operator *=@complex
operator *=(const complex<_Tp>& z)
/*-p-
/*/

#-operator /=@complex
operator /=(const complex<_Tp>& z)
/*-p-
/*/

#-value_type@complex
double value_type
/*-p-
/*/

#-complex@complex
complex(_ComplexT z)
/*-p-
/*/

#-complex@complex
complex(double = 0.0, double = 0.0)
/*-p-
/*/

#-complex@complex
complex(const complex<float>&)
/*-p-
/*/

#-complex@complex
explicit complex(const complex<long double>&)
/*-p-
/*/

#-real@complex
double real()
/*-p-
/*/

#-real@complex
double real() const
/*-p-
/*/

#-imag@complex
double imag()
/*-p-
/*/

#-imag@complex
double imag() const
/*-p-
/*/

#-operator =@complex
complex operator =(double)
/*-p-
/*/

#-operator +=@complex
complex operator +=(double)
/*-p-
/*/

#-operator -=@complex
complex operator -=(double)
/*-p-
/*/

#-operator *=@complex
complex operator *=(double)
/*-p-
/*/

#-operator /=@complex
complex operator /=(double)
/*-p-
/*/

#-operator =@complex
complex operator =(const complex<_Tp>&)
/*-p-
/*/

#-operator +=@complex
complex operator +=(const complex<_Tp>&)
/*-p-
/*/

#-operator -=@complex
complex operator -=(const complex<_Tp>&)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const complex<_Tp>&)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const complex<_Tp>&)
/*-p-
/*/

#-real@complex
double real()
/*-p-
/*/

#-real@complex
real() const
/*-p-
/*/

#-imag@complex
double imag()
/*-p-
/*/

#-imag@complex
imag() const
/*-p-
/*/

#-complex@complex
inline complex(double r, double i)
/*-p-
/*/

#-operator =@complex
complex operator =(double d)
/*-p-
/*/

#-operator +=@complex
complex operator +=(double d)
/*-p-
/*/

#-operator -=@complex
complex operator -=(double d)
/*-p-
/*/

#-operator *=@complex
complex operator *=(double d)
/*-p-
/*/

#-operator /=@complex
complex operator /=(double d)
/*-p-
/*/

#-operator =@complex
operator =(const complex<_Tp>& z)
/*-p-
/*/

#-operator +=@complex
operator +=(const complex<_Tp>& z)
/*-p-
/*/

#-operator -=@complex
operator -=(const complex<_Tp>& z)
/*-p-
/*/

#-operator *=@complex
operator *=(const complex<_Tp>& z)
/*-p-
/*/

#-operator /=@complex
operator /=(const complex<_Tp>& z)
/*-p-
/*/

#-value_type@complex
double value_type
/*-p-
/*/

#-complex@complex
complex(_ComplexT z)
/*-p-
/*/

#-complex@complex
complex(long double = 0.0L, long double = 0.0L)
/*-p-
/*/

#-complex@complex
complex(const complex<float>&)
/*-p-
/*/

#-complex@complex
complex(const complex<double>&)
/*-p-
/*/

#-real@complex
double real()
/*-p-
/*/

#-real@complex
double real() const
/*-p-
/*/

#-imag@complex
double imag()
/*-p-
/*/

#-imag@complex
double imag() const
/*-p-
/*/

#-operator =@complex
complex operator =(long double)
/*-p-
/*/

#-operator +=@complex
complex operator +=(long double)
/*-p-
/*/

#-operator -=@complex
complex operator -=(long double)
/*-p-
/*/

#-operator *=@complex
complex operator *=(long double)
/*-p-
/*/

#-operator /=@complex
complex operator /=(long double)
/*-p-
/*/

#-operator =@complex
complex operator =(const complex<_Tp>&)
/*-p-
/*/

#-operator +=@complex
complex operator +=(const complex<_Tp>&)
/*-p-
/*/

#-operator -=@complex
complex operator -=(const complex<_Tp>&)
/*-p-
/*/

#-operator *=@complex
complex operator *=(const complex<_Tp>&)
/*-p-
/*/

#-operator /=@complex
complex operator /=(const complex<_Tp>&)
/*-p-
/*/

#-complex@complex
inline complex(long double r, long double i)
/*-p-
/*/

#-real@complex
real()
/*-p-
/*/

#-real@complex
real() const
/*-p-
/*/

#-imag@complex
imag()
/*-p-
/*/

#-imag@complex
imag() const
/*-p-
/*/

#-operator =@complex
complex operator =(long double r)
/*-p-
/*/

#-operator +=@complex
complex operator +=(long double r)
/*-p-
/*/

#-operator -=@complex
complex operator -=(long double r)
/*-p-
/*/

#-operator *=@complex
complex operator *=(long double r)
/*-p-
/*/

#-operator /=@complex
complex operator /=(long double r)
/*-p-
/*/

#-operator =@complex
operator =(const complex<_Tp>& z)
/*-p-
/*/

#-operator +=@complex
operator +=(const complex<_Tp>& z)
/*-p-
/*/

#-operator -=@complex
operator -=(const complex<_Tp>& z)
/*-p-
/*/

#-operator *=@complex
operator *=(const complex<_Tp>& z)
/*-p-
/*/

#-operator /=@complex
operator /=(const complex<_Tp>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<double>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<long double>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<float>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<long double>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<float>& z)
/*-p-
/*/

#-complex@complex
inline complex(const complex<double>& z)
/*-p-
/*/

##:deque@deque
deque
/*-p-
/*/

#-iterator@deque
iterator iterator
/*-p-
/*/

#-const_iterator@deque
const_iterator const_iterator
/*-p-
/*/

#-const_reverse_iterator@deque
reverse_iterator const_reverse_iterator
/*-p-
/*/

#-reverse_iterator@deque
reverse_iterator reverse_iterator
/*-p-
/*/

#-size_type@deque
size_t size_type
/*-p-
/*/

#-difference_type@deque
ptrdiff_t difference_type
/*-p-
/*/

#-allocator_type@deque
_Alloc allocator_type
/*-p-
/*/

#-deque@deque
explicit deque(const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
/*/

#-deque@deque
explicit deque(size_type n, const value_type& value = value_type(), const allocator_type& a = allocator_type())
/*-p-
@b:Create a deque with copies of an exemplar element.
@p:n  The number of elements to initially create.
@p:value  An element to copy.
This constructor fills the deque with  n copies of  value.
/*/

#-deque@deque
deque(const deque& x)
/*-p-
@b:Deque copy constructor.
@p:x  A deque of identical element and allocator types.
The newly-created deque uses a copy of the allocation object used
by  x.
/*/

#-deque@deque
template deque(_InputIterator first, InputIterator last, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a deque from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a deque consisting of copies of the elements from [first,
last).
If the iterators are forward, bidirectional, or random-access, then
this will call the elements' copy constructor N times (where N is
distance(first,last)) and do no memory reallocation.  But if only
input iterators are used, then this will do at most 2N calls to the
copy constructor, and logN memory reallocations.
/*/

#-~deque@deque
~deque()
/*-p-
The dtor only erases the elements, and note that if the elements
themselves are pointers, the pointed-to memory is not touched in any
way.  Managing the pointer is the user's responsibilty.
/*/

#-operator =@deque
deque operator =(const deque& x)
/*-p-
@b:Deque assignment operator.
@p:x  A deque of identical element and allocator types.
All the elements of  x are copied, but unlike the copy constructor,
the allocator object is not copied.
/*/

#-assign@deque
void assign(size_type n, const value_type& val)
/*-p-
@b:Assigns a given value to a deque.
@p:n  Number of elements to be assigned.
@p:val  Value to be assigned.
This function fills a deque with  n copies of the given
value.  Note that the assignment completely changes the
deque and that the resulting deque's size is the same as
the number of elements assigned.  Old data may be lost.
/*/

#-assign@deque
void assign(_InputIterator first, InputIterator last)
/*-p-
@b:Assigns a range to a deque.
@p:first  An input iterator.
@p:last   An input iterator.
This function fills a deque with copies of the elements in the
range [first,last).
Note that the assignment completely changes the deque and that the
resulting deque's size is the same as the number of elements
assigned.  Old data may be lost.
/*/

#-get_allocator@deque
allocator_type get_allocator() const
/*-p-
/*/

#-begin@deque
iterator begin()
/*-p-
Returns a read/write iterator that points to the first element in the
deque.  Iteration is done in ordinary element order.
/*/

#-begin@deque
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the first
element in the deque.  Iteration is done in ordinary element order.
/*/

#-end@deque
iterator end()
/*-p-
Returns a read/write iterator that points one past the last
element in the deque.  Iteration is done in ordinary
element order.
/*/

#-end@deque
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past
the last element in the deque.  Iteration is done in
ordinary element order.
/*/

#-rbegin@deque
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the
last element in the deque.  Iteration is done in reverse
element order.
/*/

#-rbegin@deque
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to the last element in the deque.  Iteration is done in
reverse element order.
/*/

#-rend@deque
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one
before the first element in the deque.  Iteration is done
in reverse element order.
/*/

#-rend@deque
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to one before the first element in the deque.  Iteration is
done in reverse element order.
/*/

#-size@deque
size_type size() const
/*-p-
/*/

#-max_size@deque
size_type max_size() const
/*-p-
/*/

#-resize@deque
void resize(size_type new_size, value_type x = value_type())
/*-p-
@b:Resizes the deque to the specified number of elements.
@p:new_size  Number of elements the deque should contain.
@p:x  Data with which new elements should be populated.
This function will resize the deque to the specified
number of elements.  If the number is smaller than the
deque's current size the deque is truncated, otherwise the
deque is extended and new elements are populated with given
data.
/*/

#-empty@deque
bool empty() const
/*-p-
Returns true if the deque is empty.  (Thus begin() would
equal end().)
/*/

#-operator []@deque
reference operator [](size_type n)
/*-p-
@b:Subscript access to the data contained in the deque.
@p:n The index of the element for which data should be
@r:Read/write reference to data.
accessed.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)
/*/

#-operator []@deque
const_reference operator [](size_type n) const
/*-p-
@b:Subscript access to the data contained in the deque.
@p:n The index of the element for which data should be
@r:Read-only (constant) reference to data.
accessed.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)
/*/

#-at@deque
reference at(size_type n)
/*-p-
@b:Provides access to the data contained in the deque.
@p:n The index of the element for which data should be
@r:Read/write reference to data.
@t:std::out_of_range  If  n is an invalid index.
accessed.
This function provides for safer data access.  The parameter
is first checked that it is in the range of the deque.  The
function throws out_of_range if the check fails.
/*/

#-at@deque
const_reference at(size_type n) const
/*-p-
@b:Provides access to the data contained in the deque.
@p:n The index of the element for which data should be
@r:Read-only (constant) reference to data.
@t:std::out_of_range  If  n is an invalid index.
accessed.
This function provides for safer data access.  The parameter is first
checked that it is in the range of the deque.  The function throws
out_of_range if the check fails.
/*/

#-front@deque
reference front()
/*-p-
Returns a read/write reference to the data at the first
element of the deque.
/*/

#-front@deque
const_reference front() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the deque.
/*/

#-back@deque
reference back()
/*-p-
Returns a read/write reference to the data at the last element of the
deque.
/*/

#-back@deque
const_reference back() const
/*-p-
Returns a read-only (constant) reference to the data at the last
element of the deque.
/*/

#-push_front@deque
void push_front(const value_type& x)
/*-p-
@b:Add data to the front of the deque.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the front of the deque and assigns the given
data to it.  Due to the nature of a deque this operation
can be done in constant time.
/*/

#-push_back@deque
void push_back(const value_type& x)
/*-p-
@b:Add data to the end of the deque.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the end of the deque and assigns the given data
to it.  Due to the nature of a deque this operation can be
done in constant time.
/*/

#-pop_front@deque
void pop_front()
/*-p-
@b:Removes first element.
This is a typical stack operation.  It shrinks the deque by one.
Note that no data is returned, and if the first element's data is
needed, it should be retrieved before pop_front() is called.
/*/

#-pop_back@deque
void pop_back()
/*-p-
@b:Removes last element.
This is a typical stack operation.  It shrinks the deque by one.
Note that no data is returned, and if the last element's data is
needed, it should be retrieved before pop_back() is called.
/*/

#-insert@deque
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Inserts given value into deque before specified iterator.
@p:position  An iterator into the deque.
@p:x  Data to be inserted.
@r:An iterator that points to the inserted data.
This function will insert a copy of the given value before the
specified location.
/*/

#-insert@deque
void insert(iterator position, size_type n, const value_type& x)
/*-p-
@b:Inserts a number of copies of given data into the deque.
@p:position  An iterator into the deque.
@p:n  Number of elements to be inserted.
@p:x  Data to be inserted.
This function will insert a specified number of copies of the given
data before the location specified by  position.
/*/

#-insert@deque
void insert(iterator position, InputIterator first, InputIterator last)
/*-p-
@b:Inserts a range into the deque.
@p:position  An iterator into the deque.
@p:first  An input iterator.
@p:last   An input iterator.
This function will insert copies of the data in the range
[first,last) into the deque before the location specified
by  pos.  This is known as "range insert."
/*/

#-erase@deque
iterator erase(iterator position)
/*-p-
@b:Remove element at given position.
@p:position  Iterator pointing to element to be erased.
@r:An iterator pointing to the next element (or end()).
This function will erase the element at the given position and thus
shorten the deque by one.
The user is cautioned that
this function only erases the element, and that if the element is
itself a pointer, the pointed-to memory is not touched in any way.
Managing the pointer is the user's responsibilty.
/*/

#-erase@deque
iterator erase(iterator first, iterator last)
/*-p-
@b:Remove a range of elements.
@p:first  Iterator pointing to the first element to be erased.
@p:last  Iterator pointing to one past the last element to be
@r:An iterator pointing to the element pointed to by  last
erased.
prior to erasing (or end()).
This function will erase the elements in the range [first,last) and
shorten the deque accordingly.
The user is cautioned that
this function only erases the elements, and that if the elements
themselves are pointers, the pointed-to memory is not touched in any
way.  Managing the pointer is the user's responsibilty.
/*/

#-swap@deque
void swap(deque& x)
/*-p-
@b:Swaps data with another deque.
@p:x  A deque of the same element and allocator types.
This exchanges the elements between two deques in constant time.
(Four pointers, so it should be quite fast.)
Note that the global std::swap() function is specialized such that
std::swap(d1,d2) will feed to this function.
/*/

#-clear@deque
void clear()
/*-p-
Erases all the elements.  Note that this function only erases the
elements, and that if the elements themselves are pointers, the
pointed-to memory is not touched in any way.  Managing the pointer is
the user's responsibilty.
/*/

##:list@list
list
/*-p-
@b:A standard container with linear time access to elements,
and fixed time insertion/deletion at any point in the sequence.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and a
<a href="tables.html#67">sequence</a>, including the
<a href="tables.html#68">optional sequence requirements</a> with the
exception of c at and c operator[].
This is a e doubly e linked list.  Traversal up and down the
list requires linear time, but adding and removing elements (or
change takes place.  Unlike std::vector and std::deque,
random-access iterators are not provided, so subscripting ( c
[] ) access is not allowed.  For algorithms which only need
sequential access, this lack makes no difference.
Also unlike the other standard containers, std::list provides
specialized algorithms unique to linked lists, such as
splicing, sorting, and in-place reversal.
A couple points on memory allocation for list<Tp>:
First, we never actually allocate a Tp, we allocate
List_node<Tp>'s and trust [20.1.5]/4 to DTRT.  This is to ensure
that after elements from list<X,Alloc1> are spliced into
list<X,Alloc2>, destroying the memory of the second list is a
valid operation, i.e., Alloc1 giveth and Alloc2 taketh away.
Second, a list conceptually represented as
A <---> B <---> C <---> D
is actually circular; a link exists between A and D.  The list
class holds (as its only data member) a private list::iterator
pointing to e D, not to e A!  To get to the head of the list,
we start at the tail and move forward by one.  When this member
iterator's next/previous pointers refer to itself, the list is
empty.  endif
/*/

#-iterator@list
_List_iterator iterator
/*-p-
/*/

#-const_iterator@list
_List_const_iterator const_iterator
/*-p-
/*/

#-const_reverse_iterator@list
reverse_iterator const_reverse_iterator
/*-p-
/*/

#-reverse_iterator@list
reverse_iterator reverse_iterator
/*-p-
/*/

#-list@list
explicit list(const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
/*/

#-list@list
explicit list(size_type n, const value_type& value = value_type(), const allocator_type& a = allocator_type())
/*-p-
@b:Create a list with copies of an exemplar element.
@p:n  The number of elements to initially create.
@p:value  An element to copy.
This constructor fills the list with  n copies of  value.
/*/

#-list@list
list(const list& x)
/*-p-
@b:List copy constructor.
@p:x  A list of identical element and allocator types.
The newly-created list uses a copy of the allocation object used
by  x.
/*/

#-list@list
template list(_InputIterator first, InputIterator last, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a list from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a list consisting of copies of the elements from
[ first, last).  This is linear in N (where N is
distance( first, last)).
/*/

#-operator =@list
list operator =(const list& x)
/*-p-
@b:List assignment operator.
@p:x  A list of identical element and allocator types.
All the elements of  x are copied, but unlike the copy
constructor, the allocator object is not copied.
/*/

#-assign@list
void assign(size_type n, const value_type& val)
/*-p-
@b:Assigns a given value to a list.
@p:n  Number of elements to be assigned.
@p:val  Value to be assigned.
This function fills a list with  n copies of the given
value.  Note that the assignment completely changes the list
and that the resulting list's size is the same as the number
of elements assigned.  Old data may be lost.
/*/

#-assign@list
void assign(_InputIterator first, InputIterator last)
/*-p-
@b:Assigns a range to a list.
@p:first  An input iterator.
@p:last   An input iterator.
This function fills a list with copies of the elements in the
range [ first, last).
Note that the assignment completely changes the list and
that the resulting list's size is the same as the number of
elements assigned.  Old data may be lost.
/*/

#-get_allocator@list
allocator_type get_allocator() const
/*-p-
/*/

#-begin@list
iterator begin()
/*-p-
Returns a read/write iterator that points to the first element in the
list.  Iteration is done in ordinary element order.
/*/

#-begin@list
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the
first element in the list.  Iteration is done in ordinary
element order.
/*/

#-end@list
iterator end()
/*-p-
Returns a read/write iterator that points one past the last
element in the list.  Iteration is done in ordinary element
order.
/*/

#-end@list
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past
the last element in the list.  Iteration is done in ordinary
element order.
/*/

#-rbegin@list
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the last
element in the list.  Iteration is done in reverse element
order.
/*/

#-rbegin@list
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points to
the last element in the list.  Iteration is done in reverse
element order.
/*/

#-rend@list
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one
before the first element in the list.  Iteration is done in
reverse element order.
/*/

#-rend@list
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points to one
before the first element in the list.  Iteration is done in reverse
element order.
/*/

#-empty@list
bool empty() const
/*-p-
Returns true if the list is empty.  (Thus begin() would equal
end().)
/*/

#-size@list
size_type size() const
/*-p-
/*/

#-max_size@list
size_type max_size() const
/*-p-
/*/

#-resize@list
void resize(size_type new_size, value_type x = value_type())
/*-p-
@b:Resizes the list to the specified number of elements.
@p:new_size Number of elements the list should contain.
@p:x Data with which new elements should be populated.
This function will resize the list to the specified number
of elements.  If the number is smaller than the list's
current size the list is truncated, otherwise the list is
extended and new elements are populated with given data.
/*/

#-front@list
reference front()
/*-p-
Returns a read/write reference to the data at the first
element of the list.
/*/

#-front@list
const_reference front() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the list.
/*/

#-back@list
reference back()
/*-p-
Returns a read/write reference to the data at the last element
of the list.
/*/

#-back@list
const_reference back() const
/*-p-
Returns a read-only (constant) reference to the data at the last
element of the list.
/*/

#-push_front@list
void push_front(const value_type& x)
/*-p-
@b:Add data to the front of the list.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the front of the list and assigns the given data
to it.  Due to the nature of a list this operation can be
done in constant time, and does not invalidate iterators and
references.
/*/

#-pop_front@list
void pop_front()
/*-p-
@b:Removes first element.
This is a typical stack operation.  It shrinks the list by
one.  Due to the nature of a list this operation can be done
in constant time, and only invalidates iterators/references to
the element being removed.
Note that no data is returned, and if the first element's data
is needed, it should be retrieved before pop_front() is
called.
/*/

#-push_back@list
void push_back(const value_type& x)
/*-p-
@b:Add data to the end of the list.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the end of the list and assigns the given data to
it.  Due to the nature of a list this operation can be done
in constant time, and does not invalidate iterators and
references.
/*/

#-pop_back@list
void pop_back()
/*-p-
@b:Removes last element.
This is a typical stack operation.  It shrinks the list by
one.  Due to the nature of a list this operation can be done
in constant time, and only invalidates iterators/references to
the element being removed.
Note that no data is returned, and if the last element's data
is needed, it should be retrieved before pop_back() is called.
/*/

#-insert@list
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Inserts given value into list before specified iterator.
@p:position  An iterator into the list.
@p:x  Data to be inserted.
@r:An iterator that points to the inserted data.
This function will insert a copy of the given value before
the specified location.  Due to the nature of a list this
operation can be done in constant time, and does not
invalidate iterators and references.
/*/

#-insert@list
void insert(iterator position, size_type n, const value_type& x)
/*-p-
@b:Inserts a number of copies of given data into the list.
@p:position  An iterator into the list.
@p:n  Number of elements to be inserted.
@p:x  Data to be inserted.
This function will insert a specified number of copies of the
given data before the location specified by  position.
This operation is linear in the number of elements inserted and
does not invalidate iterators and references.
/*/

#-insert@list
void insert(iterator position, InputIterator first, InputIterator last)
/*-p-
@b:Inserts a range into the list.
@p:position  An iterator into the list.
@p:first  An input iterator.
@p:last   An input iterator.
This function will insert copies of the data in the range [
first, last) into the list before the location specified by
This operation is linear in the number of elements inserted and
does not invalidate iterators and references.
/*/

#-erase@list
iterator erase(iterator position)
/*-p-
@b:Remove element at given position.
@p:position  Iterator pointing to element to be erased.
@r:An iterator pointing to the next element (or end()).
This function will erase the element at the given position and thus
shorten the list by one.
Due to the nature of a list this operation can be done in
constant time, and only invalidates iterators/references to
the element being removed.  The user is also cautioned that
this function only erases the element, and that if the element
is itself a pointer, the pointed-to memory is not touched in
any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@list
iterator erase(iterator first, iterator last)
/*-p-
@b:Remove a range of elements.
@p:first  Iterator pointing to the first element to be erased.
@p:last  Iterator pointing to one past the last element to be
@r:An iterator pointing to the element pointed to by  last
erased.
prior to erasing (or end()).
This function will erase the elements in the range 
[first,last) and shorten the list accordingly.
This operation is linear time in the size of the range and only
invalidates iterators/references to the element being removed.
The user is also cautioned that this function only erases the
elements, and that if the elements themselves are pointers, the
pointed-to memory is not touched in any way.  Managing the pointer
is the user's responsibilty.
/*/

#-swap@list
void swap(list& x)
/*-p-
@b:Swaps data with another list.
@p:x  A list of the same element and allocator types.
This exchanges the elements between two lists in constant
time.  Note that the global std::swap() function is
specialized such that std::swap(l1,l2) will feed to this
function.
/*/

#-clear@list
void clear()
/*-p-
Erases all the elements.  Note that this function only erases
the elements, and that if the elements themselves are
pointers, the pointed-to memory is not touched in any way.
Managing the pointer is the user's responsibilty.
/*/

#-splice@list
void splice(iterator position, list& x)
/*-p-
@b:Insert contents of another list.
@p:position  Iterator referencing the element to insert before.
@p:x  Source list.
The elements of  x are inserted in constant time in front of
the element referenced by  position.   x becomes an empty
list.
Requires this !=  x.
/*/

#-splice@list
void splice(iterator position, list& x, iterator i)
/*-p-
@b:Insert element from another list.
@p:position  Iterator referencing the element to insert before.
@p:x  Source list.
@p:i  Iterator referencing the element to move.
Removes the element in list  x referenced by  i and
inserts it into the current list before  position.
/*/

#-splice@list
void splice(iterator position, list& x, iterator first, iterator last)
/*-p-
@b:Insert range from another list.
@p:position  Iterator referencing the element to insert before.
@p:x  Source list.
@p:first  Iterator referencing the start of range in x.
@p:last  Iterator referencing the end of range in x.
Removes elements in the range [first,last) and inserts them
before  position in constant time.
Undefined if  position is in [first,last).
/*/

#-remove@list
void remove(const Tp& value)
/*-p-
@b:Remove all elements equal to value.
@p:value  The value to remove.
Removes every element in the list equal to  value.
Remaining elements stay in list order.  Note that this
function only erases the elements, and that if the elements
themselves are pointers, the pointed-to memory is not
touched in any way.  Managing the pointer is the user's
responsibilty.
/*/

#-remove_if@list
void remove_if(_Predicate)
/*-p-
@b:Remove all elements satisfying a predicate.
@p:Predicate  Unary predicate function or object.
Removes every element in the list for which the predicate
returns true.  Remaining elements stay in list order.  Note
that this function only erases the elements, and that if the
elements themselves are pointers, the pointed-to memory is
not touched in any way.  Managing the pointer is the user's
responsibilty.
/*/

#-unique@list
void unique()
/*-p-
@b:Remove consecutive duplicate elements.
For each consecutive set of elements with the same value,
remove all but the first one.  Remaining elements stay in
list order.  Note that this function only erases the
elements, and that if the elements themselves are pointers,
the pointed-to memory is not touched in any way.  Managing
the pointer is the user's responsibilty.
/*/

#-unique@list
void unique(_BinaryPredicate)
/*-p-
@b:Remove consecutive elements satisfying a predicate.
@p:BinaryPredicate  Binary predicate function or object.
For each consecutive set of elements [first,last) that
satisfy predicate(first,i) where i is an iterator in
[first,last), remove all but the first one.  Remaining
elements stay in list order.  Note that this function only
erases the elements, and that if the elements themselves are
pointers, the pointed-to memory is not touched in any way.
Managing the pointer is the user's responsibilty.
/*/

#-merge@list
void merge(list& x)
/*-p-
@b:Merge sorted lists.
@p:x  Sorted list to merge.
Assumes that both  x and this list are sorted according to
operator<().  Merges elements of  x into this list in
sorted order, leaving  x empty when complete.  Elements in
this list precede elements in  x that are equal.
/*/

#-merge@list
void merge(list&, StrictWeakOrdering)
/*-p-
@b:Merge sorted lists according to comparison function.
@p:x  Sorted list to merge.
@p:StrictWeakOrdering Comparison function definining
sort order.
Assumes that both  x and this list are sorted according to
StrictWeakOrdering.  Merges elements of  x into this list
in sorted order, leaving  x empty when complete.  Elements
in this list precede elements in  x that are equivalent
according to StrictWeakOrdering().
/*/

#-reverse@list
void reverse()
/*-p-
@b:Reverse the elements in list.
Reverse the order of elements in the list in linear time.
/*/

#-sort@list
void sort()
/*-p-
@b:Sort the elements.
Sorts the elements of this list in NlogN time.  Equivalent
elements remain in list order.
/*/

#-sort@list
void sort(_StrictWeakOrdering)
/*-p-
@b:Sort the elements according to comparison function.
Sorts the elements of this list in NlogN time.  Equivalent
elements remain in list order.
/*/

##:map@map
map
/*-p-
@b:A standard container made up of (key,value) pairs, which can be
retrieved based on a key, in logarithmic time.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and an
<a href="tables.html#69">associative container</a> (using unique keys).
For a c map<Key,T> the key_type is Key, the mapped_type is T, and the
value_type is std::pair<const Key,T>.
Maps support bidirectional iterators.
The private tree data is declared exactly the same way for map and
multimap; the distinction is made entirely in how the tree functions are
called (*_unique versus *_equal, same as the standard).
/*/

#-iterator@map
iterator iterator
/*-p-
/*/

#-const_iterator@map
const_iterator const_iterator
/*-p-
/*/

#-size_type@map
size_type size_type
/*-p-
/*/

#-difference_type@map
difference_type difference_type
/*-p-
/*/

#-reverse_iterator@map
reverse_iterator reverse_iterator
/*-p-
/*/

#-const_reverse_iterator@map
const_reverse_iterator const_reverse_iterator
/*-p-
/*/

#-map@map
map()
/*-p-
@b:Default constructor creates no elements.
/*/

#-map@map
explicit map(const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
/*/

#-map@map
map(const map& x)
/*-p-
@b:Map copy constructor.
@p:x  A map of identical element and allocator types.
The newly-created map uses a copy of the allocation object used
by  x.
/*/

#-map@map
template map(_InputIterator first, InputIterator last)
/*-p-
@b:Builds a map from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a map consisting of copies of the elements from [first,last).
This is linear in N if the range is already sorted, and NlogN
otherwise (where N is distance(first,last)).
/*/

#-map@map
template map(_InputIterator first, InputIterator last, const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a map from a range.
@p:first  An input iterator.
@p:last  An input iterator.
@p:comp  A comparison functor.
@p:a  An allocator object.
Create a map consisting of copies of the elements from [first,last).
This is linear in N if the range is already sorted, and NlogN
otherwise (where N is distance(first,last)).
/*/

#-operator =@map
map operator =(const map& x)
/*-p-
@b:Map assignment operator.
@p:x  A map of identical element and allocator types.
All the elements of  x are copied, but unlike the copy constructor,
the allocator object is not copied.
/*/

#-get_allocator@map
allocator_type get_allocator() const
/*-p-
/*/

#-begin@map
iterator begin()
/*-p-
Returns a read/write iterator that points to the first pair in the
map.
Iteration is done in ascending order according to the keys.
/*/

#-begin@map
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the first pair
in the map.  Iteration is done in ascending order according to the
keys.
/*/

#-end@map
iterator end()
/*-p-
Returns a read/write iterator that points one past the last
pair in the map.  Iteration is done in ascending order
according to the keys.
/*/

#-end@map
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past the last
pair in the map.  Iteration is done in ascending order according to
the keys.
/*/

#-rbegin@map
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the last pair in
the map.  Iteration is done in descending order according to the
keys.
/*/

#-rbegin@map
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points to the
last pair in the map.  Iteration is done in descending order
according to the keys.
/*/

#-rend@map
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one before the
first pair in the map.  Iteration is done in descending order
according to the keys.
/*/

#-rend@map
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points to one
before the first pair in the map.  Iteration is done in descending
order according to the keys.
/*/

#-empty@map
bool empty() const
/*-p-
end().)
/*/

#-size@map
size_type size() const
/*-p-
/*/

#-max_size@map
size_type max_size() const
/*-p-
/*/

#-operator []@map
mapped_type operator [](const key_type& k)
/*-p-
@b:Subscript ( c [] ) access to map data.
@p:k  The key for which data should be retrieved.
@r:A reference to the data of the (key,data) pair.
Allows for easy lookup with the subscript ( c [] )
operator.  Returns data associated with the key specified in
subscript.  If the key does not exist, a pair with that key
is created using default values, which is then returned.
Lookup requires logarithmic time.
/*/

#-at@map
mapped_type at(const key_type& k)
/*-p-
@b:Access to map data.
@p:k  The key for which data should be retrieved.
@r:A reference to the data whose key is equivalent to  k, if
@t:std::out_of_range  If no such data is present.
such a data is present in the map.
/*/

#-at@map
mapped_type at(const key_type& k) const
/*-p-
/*/

#-insert@map
pair insert(const value_type& x)
/*-p-
@b:Attempts to insert a std::pair into the map.
@p:x  Pair to be inserted (see std::make_pair for easy creation 
@r:A pair, of which the first element is an iterator that 
of pairs).
points to the possibly inserted pair, and the second is 
a bool that is true if the pair was actually inserted.
This function attempts to insert a (key, value) pair into the map.
A map relies on unique keys and thus a pair is only inserted if its
first element (the key) is not already present in the map.
Insertion requires logarithmic time.
/*/

#-insert@map
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Attempts to insert a std::pair into the map.
@p:position  An iterator that serves as a hint as to where the
@p:x  Pair to be inserted (see std::make_pair for easy creation
@r:An iterator that points to the element with key of  x (may
pair should be inserted.
of pairs).
or may not be the pair passed in).
This function is not concerned about whether the insertion
took place, and thus does not return a boolean like the
single-argument insert() does.  Note that the first
parameter is only a hint and can potentially improve the
performance of the insertion process.  A bad hint would
cause no gains in efficiency.
See
http://gcc.gnu.org/onlinedocs/libstdc++/23_containers/howto.html#4
for more on "hinting".
Insertion requires logarithmic time (if the hint is not taken).
/*/

#-insert@map
void insert(_InputIterator first, InputIterator last)
/*-p-
@b:Template function that attemps to insert a range of elements.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range.
inserted.
Complexity similar to that of the range constructor.
/*/

#-erase@map
void erase(iterator position)
/*-p-
@b:Erases an element from a map.
@p:position  An iterator pointing to the element to be erased.
This function erases an element, pointed to by the given
iterator, from a map.  Note that this function only erases
the element, and that if the element is itself a pointer,
the pointed-to memory is not touched in any way.  Managing
the pointer is the user's responsibilty.
/*/

#-erase@map
size_type erase(const key_type& x)
/*-p-
@b:Erases elements according to the provided key.
@p:x  Key of element to be erased.
@r:The number of elements erased.
This function erases all the elements located by the given key from
a map.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@map
void erase(iterator first, iterator last)
/*-p-
@b:Erases a [first,last) range of elements from a map.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range to be erased.
erased.
This function erases a sequence of elements from a map.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-swap@map
void swap(map& x)
/*-p-
@b:Swaps data with another map.
@p:x  A map of the same element and allocator types.
This exchanges the elements between two maps in constant
time.  (It is only swapping a pointer, an integer, and an
instance of the c Compare type (which itself is often
stateless and empty), so it should be quite fast.)  Note
that the global std::swap() function is specialized such
that std::swap(m1,m2) will feed to this function.
/*/

#-clear@map
void clear()
/*-p-
Erases all elements in a map.  Note that this function only
erases the elements, and that if the elements themselves are
pointers, the pointed-to memory is not touched in any way.
Managing the pointer is the user's responsibilty.
/*/

#-key_comp@map
key_compare key_comp() const
/*-p-
Returns the key comparison object out of which the map was
constructed.
/*/

#-value_comp@map
value_compare value_comp() const
/*-p-
Returns a value comparison object, built from the key comparison
object out of which the map was constructed.
/*/

#-find@map
iterator find(const key_type& x)
/*-p-
@b:Tries to locate an element in a map.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to sought-after element, or end() if not
found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns an iterator
pointing to the sought after pair.  If unsuccessful it returns the
past-the-end ( c end() ) iterator.
/*/

#-find@map
const_iterator find(const key_type& x) const
/*-p-
@b:Tries to locate an element in a map.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to sought-after
element, or end() if not found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns a constant
iterator pointing to the sought after pair. If unsuccessful it
returns the past-the-end ( c end() ) iterator.
/*/

#-count@map
size_type count(const key_type& x) const
/*-p-
@b:Finds the number of elements with given key.
@p:x  Key of (key, value) pairs to be located.
@r:Number of elements with specified key.
This function only makes sense for multimaps; for map the result will
either be 0 (not present) or 1 (present).
/*/

#-lower_bound@map
iterator lower_bound(const key_type& x)
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to first element equal to or greater
than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful it returns an iterator
pointing to the first element that has a greater value than given key
or end() if no such element exists.
/*/

#-lower_bound@map
const_iterator lower_bound(const key_type& x) const
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to first element
equal to or greater than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful it returns an iterator
pointing to the first element that has a greater value than given key
or end() if no such element exists.
/*/

#-upper_bound@map
iterator upper_bound(const key_type& x)
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to the first element
greater than key, or end().
/*/

#-upper_bound@map
const_iterator upper_bound(const key_type& x) const
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to first iterator
greater than key, or end().
/*/

#-equal_range@map
pair equal_range(const key_type& x)
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key of (key, value) pairs to be located.
@r:Pair of iterators that possibly points to the subsequence
matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
This function probably only makes sense for multimaps.
/*/

#-equal_range@map
pair equal_range(const key_type& x) const
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key of (key, value) pairs to be located.
@r:Pair of read-only (constant) iterators that possibly points
to the subsequence matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
This function probably only makes sense for multimaps.
/*/

#-operator ==@map
bool operator ==(const map<_K1, T1, C1, A1>&, const map<_K1, T1, C1, A1>&)
/*-p-
/*/

#-operator <@map
bool operator <(const map<_K1, T1, C1, A1>&, const map<_K1, T1, C1, A1>&)
/*-p-
/*/

##:multimap@multimap
multimap
/*-p-
@b:A standard container made up of (key,value) pairs, which can be
retrieved based on a key, in logarithmic time.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and an
<a href="tables.html#69">associative container</a> (using equivalent
keys).  For a c multimap<Key,T> the key_type is Key, the mapped_type
is T, and the value_type is std::pair<const Key,T>.
Multimaps support bidirectional iterators.
The private tree data is declared exactly the same way for map and
multimap; the distinction is made entirely in how the tree functions are
called (*_unique versus *_equal, same as the standard).
/*/


#-iterator@multimap
iterator iterator
/*-p-
/*/

#-const_iterator@multimap
const_iterator const_iterator
/*-p-
/*/

#-size_type@multimap
size_type size_type
/*-p-
/*/

#-difference_type@multimap
difference_type difference_type
/*-p-
/*/

#-reverse_iterator@multimap
reverse_iterator reverse_iterator
/*-p-
/*/

#-const_reverse_iterator@multimap
const_reverse_iterator const_reverse_iterator
/*-p-
/*/

#-multimap@multimap
multimap()
/*-p-
@b:Default constructor creates no elements.
/*/

#-multimap@multimap
explicit multimap(const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
/*/

#-multimap@multimap
multimap(const multimap& x)
/*-p-
@b:Multimap copy constructor.
@p:x  A multimap of identical element and allocator types.
The newly-created multimap uses a copy of the allocation object used
by  x.
/*/

#-multimap@multimap
template multimap(_InputIterator first, InputIterator last)
/*-p-
@b:Builds a multimap from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a multimap consisting of copies of the elements from
[first,last).  This is linear in N if the range is already sorted,
and NlogN otherwise (where N is distance(first,last)).
/*/

#-multimap@multimap
template multimap(_InputIterator first, InputIterator last, const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a multimap from a range.
@p:first  An input iterator.
@p:last  An input iterator.
@p:comp  A comparison functor.
@p:a  An allocator object.
Create a multimap consisting of copies of the elements from
[first,last).  This is linear in N if the range is already sorted,
and NlogN otherwise (where N is distance(first,last)).
/*/

#-operator =@multimap
multimap operator =(const multimap& x)
/*-p-
@b:Multimap assignment operator.
@p:x  A multimap of identical element and allocator types.
All the elements of  x are copied, but unlike the copy constructor,
the allocator object is not copied.
/*/

#-get_allocator@multimap
allocator_type get_allocator() const
/*-p-
/*/

#-begin@multimap
iterator begin()
/*-p-
Returns a read/write iterator that points to the first pair in the
multimap.  Iteration is done in ascending order according to the
keys.
/*/

#-begin@multimap
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the first pair
in the multimap.  Iteration is done in ascending order according to
the keys.
/*/

#-end@multimap
iterator end()
/*-p-
Returns a read/write iterator that points one past the last pair in
the multimap.  Iteration is done in ascending order according to the
keys.
/*/

#-end@multimap
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past the last
pair in the multimap.  Iteration is done in ascending order according
to the keys.
/*/

#-rbegin@multimap
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the last pair in
the multimap.  Iteration is done in descending order according to the
keys.
/*/

#-rbegin@multimap
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points to the
last pair in the multimap.  Iteration is done in descending order
according to the keys.
/*/

#-rend@multimap
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one before the
first pair in the multimap.  Iteration is done in descending order
according to the keys.
/*/

#-rend@multimap
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points to one
before the first pair in the multimap.  Iteration is done in
descending order according to the keys.
/*/

#-empty@multimap
bool empty() const
/*-p-
/*/

#-size@multimap
size_type size() const
/*-p-
/*/

#-max_size@multimap
size_type max_size() const
/*-p-
/*/

#-insert@multimap
iterator insert(const value_type& x)
/*-p-
@b:Inserts a std::pair into the multimap.
@p:x  Pair to be inserted (see std::make_pair for easy creation
@r:An iterator that points to the inserted (key,value) pair.
of pairs).
This function inserts a (key, value) pair into the multimap.
Contrary to a std::map the multimap does not rely on unique keys and
thus multiple pairs with the same key can be inserted.
Insertion requires logarithmic time.
/*/

#-insert@multimap
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Inserts a std::pair into the multimap.
@p:position  An iterator that serves as a hint as to where the
@p:x  Pair to be inserted (see std::make_pair for easy creation
@r:An iterator that points to the inserted (key,value) pair.
pair should be inserted.
of pairs).
This function inserts a (key, value) pair into the multimap.
Contrary to a std::map the multimap does not rely on unique keys and
thus multiple pairs with the same key can be inserted.
Note that the first parameter is only a hint and can potentially
improve the performance of the insertion process.  A bad hint would
cause no gains in efficiency.
See http://gcc.gnu.org/onlinedocs/libstdc++/23_containers/howto.html#4
for more on "hinting".
Insertion requires logarithmic time (if the hint is not taken).
/*/

#-insert@multimap
void insert(_InputIterator first, InputIterator last)
/*-p-
@b:A template function that attemps to insert a range of elements.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range.
inserted.
Complexity similar to that of the range constructor.
/*/

#-erase@multimap
void erase(iterator position)
/*-p-
@b:Erases an element from a multimap.
@p:position  An iterator pointing to the element to be erased.
This function erases an element, pointed to by the given iterator,
from a multimap.  Note that this function only erases the element,
and that if the element is itself a pointer, the pointed-to memory is
not touched in any way.  Managing the pointer is the user's
responsibilty.
/*/

#-erase@multimap
size_type erase(const key_type& x)
/*-p-
@b:Erases elements according to the provided key.
@p:x  Key of element to be erased.
@r:The number of elements erased.
This function erases all elements located by the given key from a
multimap.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@multimap
void erase(iterator first, iterator last)
/*-p-
@b:Erases a [first,last) range of elements from a multimap.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range to be erased.
erased.
This function erases a sequence of elements from a multimap.
Note that this function only erases the elements, and that if
the elements themselves are pointers, the pointed-to memory is not
touched in any way.  Managing the pointer is the user's responsibilty.
/*/

#-swap@multimap
void swap(multimap& x)
/*-p-
@b:Swaps data with another multimap.
@p:x  A multimap of the same element and allocator types.
This exchanges the elements between two multimaps in constant time.
(It is only swapping a pointer, an integer, and an instance of
the c Compare type (which itself is often stateless and empty), so it
should be quite fast.)
Note that the global std::swap() function is specialized such that
std::swap(m1,m2) will feed to this function.
/*/

#-clear@multimap
void clear()
/*-p-
Erases all elements in a multimap.  Note that this function only
erases the elements, and that if the elements themselves are pointers,
the pointed-to memory is not touched in any way.  Managing the pointer
is the user's responsibilty.
/*/

#-key_comp@multimap
key_compare key_comp() const
/*-p-
Returns the key comparison object out of which the multimap
was constructed.
/*/

#-value_comp@multimap
value_compare value_comp() const
/*-p-
Returns a value comparison object, built from the key comparison
object out of which the multimap was constructed.
/*/

#-find@multimap
iterator find(const key_type& x)
/*-p-
@b:Tries to locate an element in a multimap.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to sought-after element,
or end() if not found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns an iterator
pointing to the sought after pair.  If unsuccessful it returns the
past-the-end ( c end() ) iterator.
/*/

#-find@multimap
const_iterator find(const key_type& x) const
/*-p-
@b:Tries to locate an element in a multimap.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to sought-after
element, or end() if not found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns a constant
iterator pointing to the sought after pair.  If unsuccessful it
returns the past-the-end ( c end() ) iterator.
/*/

#-count@multimap
size_type count(const key_type& x) const
/*-p-
@b:Finds the number of elements with given key.
@p:x  Key of (key, value) pairs to be located.
@r:Number of elements with specified key.
/*/

#-lower_bound@multimap
iterator lower_bound(const key_type& x)
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to first element equal to or greater
than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful it returns an iterator
pointing to the first element that has a greater value than given key
or end() if no such element exists.
/*/

#-lower_bound@multimap
const_iterator lower_bound(const key_type& x) const
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to first element
equal to or greater than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful the iterator will point
to the next greatest element or, if no such greater element exists, to
end().
/*/

#-upper_bound@multimap
iterator upper_bound(const key_type& x)
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Iterator pointing to the first element
greater than key, or end().
/*/

#-upper_bound@multimap
const_iterator upper_bound(const key_type& x) const
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key of (key, value) pair to be located.
@r:Read-only (constant) iterator pointing to first iterator
greater than key, or end().
/*/

#-equal_range@multimap
pair equal_range(const key_type& x)
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key of (key, value) pairs to be located.
@r:Pair of iterators that possibly points to the subsequence
matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
/*/

#-equal_range@multimap
pair equal_range(const key_type& x) const
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key of (key, value) pairs to be located.
@r:Pair of read-only (constant) iterators that possibly points
to the subsequence matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
/*/

#-operator ==@multimap
bool operator ==(const multimap<_K1, T1, C1, A1>&, const multimap<_K1, T1, C1, A1>&)
/*-p-
/*/

#-operator <@multimap
bool operator <(const multimap<_K1, T1, C1, A1>&, const multimap<_K1, T1, C1, A1>&)
/*-p-
/*/

##:multiset@multiset
multiset
/*-p-
@b:A standard container made up of elements, which can be retrieved
in logarithmic time.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and an
<a href="tables.html#69">associative container</a> (using equivalent
keys).  For a c multiset<Key> the key_type and value_type are Key.
Multisets support bidirectional iterators.
The private tree data is declared exactly the same way for set and
multiset; the distinction is made entirely in how the tree functions are
called (*_unique versus *_equal, same as the standard).
/*/

#-iterator@multiset
const_iterator iterator
/*-p-
/*/

#-const_iterator@multiset
const_iterator const_iterator
/*-p-
/*/

#-reverse_iterator@multiset
const_reverse_iterator reverse_iterator
/*-p-
/*/

#-const_reverse_iterator@multiset
const_reverse_iterator const_reverse_iterator
/*-p-
/*/

#-size_type@multiset
size_type size_type
/*-p-
/*/

#-difference_type@multiset
difference_type difference_type
/*-p-
/*/

#-multiset@multiset
multiset()
/*-p-
@b:Default constructor creates no elements.
/*/

#-multiset@multiset
explicit multiset(const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
/*/

#-multiset@multiset
template multiset(_InputIterator first, InputIterator last)
/*-p-
@b:Builds a multiset from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a multiset consisting of copies of the elements from
[first,last).  This is linear in N if the range is already sorted,
and NlogN otherwise (where N is distance(first,last)).
/*/

#-multiset@multiset
template multiset(_InputIterator first, InputIterator last, const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a multiset from a range.
@p:first  An input iterator.
@p:last  An input iterator.
@p:comp  A comparison functor.
@p:a  An allocator object.
Create a multiset consisting of copies of the elements from
[first,last).  This is linear in N if the range is already sorted,
and NlogN otherwise (where N is distance(first,last)).
/*/

#-multiset@multiset
multiset(const multiset<_Key,_Compare,_Alloc>& x)
/*-p-
@b:Multiset copy constructor.
@p:x  A multiset of identical element and allocator types.
The newly-created multiset uses a copy of the allocation object used
by  x.
/*/

#-operator =@multiset
multiset operator =(const multiset<_Key,_Compare,_Alloc>& x)
/*-p-
@b:Multiset assignment operator.
@p:x  A multiset of identical element and allocator types.
All the elements of  x are copied, but unlike the copy constructor,
the allocator object is not copied.
/*/

#-key_comp@multiset
key_compare key_comp() const
/*-p-
/*/

#-value_comp@multiset
value_compare value_comp() const
/*-p-
/*/

#-get_allocator@multiset
allocator_type get_allocator() const
/*-p-
/*/

#-begin@multiset
iterator begin() const
/*-p-
Returns a read/write iterator that points to the first element in the
multiset.  Iteration is done in ascending order according to the
keys.
/*/

#-end@multiset
iterator end() const
/*-p-
Returns a read/write iterator that points one past the last element in
the multiset.  Iteration is done in ascending order according to the
keys.
/*/

#-rbegin@multiset
reverse_iterator rbegin() const
/*-p-
Returns a read/write reverse iterator that points to the last element
in the multiset.  Iteration is done in descending order according to
the keys.
/*/

#-rend@multiset
reverse_iterator rend() const
/*-p-
Returns a read/write reverse iterator that points to the last element
in the multiset.  Iteration is done in descending order according to
the keys.
/*/

#-empty@multiset
bool empty() const
/*-p-
/*/

#-size@multiset
size_type size() const
/*-p-
/*/

#-max_size@multiset
size_type max_size() const
/*-p-
/*/

#-swap@multiset
void swap(multiset<_Key, Compare, Alloc>& x)
/*-p-
@b:Swaps data with another multiset.
@p:x  A multiset of the same element and allocator types.
This exchanges the elements between two multisets in constant time.
(It is only swapping a pointer, an integer, and an instance of the c
Compare type (which itself is often stateless and empty), so it should
be quite fast.)
Note that the global std::swap() function is specialized such that
std::swap(s1,s2) will feed to this function.
/*/

#-insert@multiset
iterator insert(const value_type& x)
/*-p-
@b:Inserts an element into the multiset.
@p:x  Element to be inserted.
@r:An iterator that points to the inserted element.
This function inserts an element into the multiset.  Contrary
to a std::set the multiset does not rely on unique keys and thus
multiple copies of the same element can be inserted.
Insertion requires logarithmic time.
/*/

#-insert@multiset
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Inserts an element into the multiset.
@p:position  An iterator that serves as a hint as to where the
@p:x  Element to be inserted.
@r:An iterator that points to the inserted element.
element should be inserted.
This function inserts an element into the multiset.  Contrary
to a std::set the multiset does not rely on unique keys and thus
multiple copies of the same element can be inserted.
Note that the first parameter is only a hint and can potentially
improve the performance of the insertion process.  A bad hint would
cause no gains in efficiency.
See http://gcc.gnu.org/onlinedocs/libstdc++/23_containers/howto.html#4
for more on "hinting".
Insertion requires logarithmic time (if the hint is not taken).
/*/

#-insert@multiset
void insert(_InputIterator first, InputIterator last)
/*-p-
@b:A template function that attemps to insert a range of elements.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range.
inserted.
Complexity similar to that of the range constructor.
/*/

#-erase@multiset
void erase(iterator position)
/*-p-
@b:Erases an element from a multiset.
@p:position  An iterator pointing to the element to be erased.
This function erases an element, pointed to by the given iterator,
from a multiset.  Note that this function only erases the element,
and that if the element is itself a pointer, the pointed-to memory is
not touched in any way.  Managing the pointer is the user's
responsibilty.
/*/

#-erase@multiset
size_type erase(const key_type& x)
/*-p-
@b:Erases elements according to the provided key.
@p:x  Key of element to be erased.
@r:The number of elements erased.
This function erases all elements located by the given key from a
multiset.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@multiset
void erase(iterator first, iterator last)
/*-p-
@b:Erases a [first,last) range of elements from a multiset.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range to be erased.
erased.
This function erases a sequence of elements from a multiset.
Note that this function only erases the elements, and that if
the elements themselves are pointers, the pointed-to memory is not
touched in any way.  Managing the pointer is the user's responsibilty.
/*/

#-clear@multiset
void clear()
/*-p-
Erases all elements in a multiset.  Note that this function only
erases the elements, and that if the elements themselves are pointers,
the pointed-to memory is not touched in any way.  Managing the pointer
is the user's responsibilty.
/*/

#-count@multiset
size_type count(const key_type& x) const
/*-p-
@b:Finds the number of elements with given key.
@p:x  Key of elements to be located.
@r:Number of elements with specified key.
/*/

#-find@multiset
iterator find(const key_type& x)
/*-p-
@b:Tries to locate an element in a set.
@p:x  Element to be located.
@r:Iterator pointing to sought-after element, or end() if not
found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns an iterator
pointing to the sought after element.  If unsuccessful it returns the
past-the-end ( c end() ) iterator.
/*/

#-find@multiset
const_iterator find(const key_type& x) const
/*-p-
/*/

#-lower_bound@multiset
iterator lower_bound(const key_type& x)
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key to be located.
@r:Iterator pointing to first element equal to or greater
than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful it returns an iterator
pointing to the first element that has a greater value than given key
or end() if no such element exists.
/*/

#-lower_bound@multiset
const_iterator lower_bound(const key_type& x) const
/*-p-
/*/

#-upper_bound@multiset
iterator upper_bound(const key_type& x)
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key to be located.
@r:Iterator pointing to the first element
greater than key, or end().
/*/

#-upper_bound@multiset
const_iterator upper_bound(const key_type& x) const
/*-p-
/*/

#-equal_range@multiset
pair equal_range(const key_type& x)
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key to be located.
@r:Pair of iterators that possibly points to the subsequence
matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
This function probably only makes sense for multisets.
/*/

#-equal_range@multiset
pair equal_range(const key_type& x) const
/*-p-
/*/

#-operator ==@multiset
bool operator ==(const multiset<_K1, C1, A1>&, const multiset<_K1, C1, A1>&)
/*-p-
/*/

#-operator <@multiset
bool operator <(const multiset<_K1, C1, A1>&, const multiset<_K1, C1, A1>&)
/*-p-
/*/

##:priority_queue@priority_queue
priority_queue
/*-p-
@b:A standard container automatically sorting its contents.
This is not a true container, but an e adaptor.  It holds
another container, and provides a wrapper interface to that
container.  The wrapper is what enforces priority-based sorting 
and queue behavior.  Very few of the standard container/sequence
interface requirements are met (e.g., iterators).
The second template parameter defines the type of the underlying
sequence/container.  It defaults to std::vector, but it can be
any type that supports c front(), c push_back, c pop_back,
and random-access iterators, such as std::deque or an
appropriate user-defined type.
The third template parameter supplies the means of making
priority comparisons.  It defaults to c less<value_type> but
can be anything defining a strict weak ordering.
Members not found in "normal" containers are c container_type,
which is a typedef for the second Sequence parameter, and c
push, c pop, and c top, which are standard queue operations.
priority_queue.
and removed from, the priority_queue using the
priority_queue's member functions.  If you access the elements
by other means, and change their data such that the sorting
order would be different, the priority_queue will not re-sort
the elements for you.  (How could it know to do so?)
/*/

#-value_type@priority_queue
value_type value_type
/*-p-
/*/

#-reference@priority_queue
reference reference
/*-p-
/*/

#-const_reference@priority_queue
const_reference const_reference
/*-p-
/*/

#-size_type@priority_queue
size_type size_type
/*-p-
/*/

#-container_type@priority_queue
_Sequence container_type
/*-p-
/*/

#-c@priority_queue
_Sequence c
/*-p-
/*/

#-comp@priority_queue
_Compare comp
/*-p-
/*/

#-priority_queue@priority_queue
explicit priority_queue(const Compare& x = Compare(), const Sequence& s = Sequence())
/*-p-
@b:Default constructor creates no elements.
/*/

#-priority_queue@priority_queue
template priority_queue(_InputIterator first, InputIterator last, const Compare& x = Compare(), const Sequence& s = Sequence())
/*-p-
@b:Builds a queue from a range.
@p:first  An input iterator.
@p:last  An input iterator.
@p:x  A comparison functor describing a strict weak ordering.
@p:s  An initial sequence with which to start.
Begins by copying  s, inserting a copy of the elements
from  [first,last) into the copy of  s, then ordering
the copy according to  x.
For more information on function objects, see the
documentation on link s20_3_1_base functor base
classesendlink.
/*/

#-empty@priority_queue
bool empty() const
/*-p-
Returns true if the queue is empty.
/*/

#-size@priority_queue
size_type size() const
/*-p-
/*/

#-top@priority_queue
const_reference top() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the queue.
/*/

#-push@priority_queue
void push(const value_type& x)
/*-p-
@b:Add data to the queue.
@p:x  Data to be added.
This is a typical queue operation.
The time complexity of the operation depends on the underlying
sequence.
/*/

#-pop@priority_queue
void pop()
/*-p-
@b:Removes first element.
This is a typical queue operation.  It shrinks the queue
by one.  The time complexity of the operation depends on the
underlying sequence.
Note that no data is returned, and if the first element's
data is needed, it should be retrieved before pop() is
called.
/*/

##:queue@queue
queue
/*-p-
@b:A standard container giving FIFO behavior.
Meets many of the requirements of a
container,
but does not define anything to do with iterators.  Very few of the
other standard container interfaces are defined.
This is not a true container, but an e adaptor.  It holds another
container, and provides a wrapper interface to that container.  The
wrapper is what enforces strict first-in-first-out queue behavior.
The second template parameter defines the type of the underlying
sequence/container.  It defaults to std::deque, but it can be any type
that supports c front, c back, c push_back, and c pop_front,
such as std::list or an appropriate user-defined type.
Members not found in "normal" containers are c container_type,
which is a typedef for the second Sequence parameter, and c push and
/*/

#-operator ==@queue
bool operator ==(const queue<_Tp1, Seq1>&, const queue<_Tp1, Seq1>&)
/*-p-
/*/

#-operator <@queue
bool operator <(const queue<_Tp1, Seq1>&, const queue<_Tp1, Seq1>&)
/*-p-
/*/

#-value_type@queue
value_type value_type
/*-p-
/*/

#-reference@queue
reference reference
/*-p-
/*/

#-const_reference@queue
const_reference const_reference
/*-p-
/*/

#-size_type@queue
size_type size_type
/*-p-
/*/

#-container_type@queue
_Sequence container_type
/*-p-
/*/

#-c@queue
_Sequence c
/*-p-
'c' is the underlying container.  Maintainers wondering why
this isn't uglified as per style guidelines should note that
this name is specified in the standard, [23.2.3.1].  (Why?
Presumably for the same reason that it's protected instead
of private: to allow derivation.  But none of the other
containers allow for derivation.  Odd.)
/*/

#-queue@queue
explicit queue(const Sequence& c = Sequence())
/*-p-
@b:Default constructor creates no elements.
/*/

#-empty@queue
bool empty() const
/*-p-
Returns true if the queue is empty.
/*/

#-size@queue
size_type size() const
/*-p-
/*/

#-front@queue
reference front()
/*-p-
Returns a read/write reference to the data at the first
element of the queue.
/*/

#-front@queue
const_reference front() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the queue.
/*/

#-back@queue
reference back()
/*-p-
Returns a read/write reference to the data at the last
element of the queue.
/*/

#-back@queue
const_reference back() const
/*-p-
Returns a read-only (constant) reference to the data at the last
element of the queue.
/*/

#-push@queue
void push(const value_type& x)
/*-p-
@b:Add data to the end of the queue.
@p:x  Data to be added.
This is a typical queue operation.  The function creates an
element at the end of the queue and assigns the given data
to it.  The time complexity of the operation depends on the
underlying sequence.
/*/

#-pop@queue
void pop()
/*-p-
@b:Removes first element.
This is a typical queue operation.  It shrinks the queue by one.
The time complexity of the operation depends on the underlying
sequence.
Note that no data is returned, and if the first element's
data is needed, it should be retrieved before pop() is
called.
/*/

##:set@set
set
/*-p-
@b:A standard container made up of unique keys, which can be
@p:Key  Type of key objects.
@p:Compare  Comparison function object type, defaults to less<Key>.
@p:Alloc  Allocator type, defaults to allocator<Key>.
retrieved in logarithmic time.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and an
<a href="tables.html#69">associative container</a> (using unique keys).
Sets support bidirectional iterators.
The private tree data is declared exactly the same way for set and
multiset; the distinction is made entirely in how the tree functions are
called (*_unique versus *_equal, same as the standard).
/*/

#-iterator@set
const_iterator iterator
/*-p-
/*/

#-const_iterator@set
const_iterator const_iterator
/*-p-
/*/

#-reverse_iterator@set
const_reverse_iterator reverse_iterator
/*-p-
/*/

#-const_reverse_iterator@set
const_reverse_iterator const_reverse_iterator
/*-p-
/*/

#-size_type@set
size_type size_type
/*-p-
/*/

#-difference_type@set
difference_type difference_type
/*-p-
/*/

#-set@set
set()
/*-p-
/*/

#-set@set
explicit set(const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
@p:comp  Comparator to use.
@p:a  Allocator to use.
/*/

#-set@set
template set(_InputIterator first, InputIterator last)
/*-p-
@b:Builds a set from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a set consisting of copies of the elements from [first,last).
This is linear in N if the range is already sorted, and NlogN
otherwise (where N is distance(first,last)).
/*/

#-set@set
template set(_InputIterator first, InputIterator last, const Compare& comp, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a set from a range.
@p:first  An input iterator.
@p:last  An input iterator.
@p:comp  A comparison functor.
@p:a  An allocator object.
Create a set consisting of copies of the elements from [first,last).
This is linear in N if the range is already sorted, and NlogN
otherwise (where N is distance(first,last)).
/*/

#-set@set
set(const set<_Key,_Compare,_Alloc>& x)
/*-p-
@b:Set copy constructor.
@p:x  A set of identical element and allocator types.
The newly-created set uses a copy of the allocation object used
by  x.
/*/

#-operator =@set
set operator =(const set<_Key, Compare, Alloc>& x)
/*-p-
@b:Set assignment operator.
@p:x  A set of identical element and allocator types.
All the elements of  x are copied, but unlike the copy constructor,
the allocator object is not copied.
/*/

#-key_comp@set
key_compare key_comp() const
/*-p-
/*/

#-value_comp@set
value_compare value_comp() const
/*-p-
/*/

#-get_allocator@set
allocator_type get_allocator() const
/*-p-
/*/

#-begin@set
iterator begin() const
/*-p-
Returns a read/write iterator that points to the first element in the
set.  Iteration is done in ascending order according to the keys.
/*/

#-end@set
iterator end() const
/*-p-
Returns a read/write iterator that points one past the last element in
the set.  Iteration is done in ascending order according to the keys.
/*/

#-rbegin@set
reverse_iterator rbegin() const
/*-p-
Returns a read/write reverse iterator that points to the last element
in the set.  Iteration is done in descending order according to the
keys.
/*/

#-rend@set
reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points to the
last pair in the map.  Iteration is done in descending order
according to the keys.
/*/

#-empty@set
bool empty() const
/*-p-
/*/

#-size@set
size_type size() const
/*-p-
/*/

#-max_size@set
size_type max_size() const
/*-p-
/*/

#-swap@set
void swap(set<_Key,_Compare,_Alloc>& x)
/*-p-
@b:Swaps data with another set.
@p:x  A set of the same element and allocator types.
This exchanges the elements between two sets in constant time.
(It is only swapping a pointer, an integer, and an instance of
the c Compare type (which itself is often stateless and empty), so it
should be quite fast.)
Note that the global std::swap() function is specialized such that
std::swap(s1,s2) will feed to this function.
/*/

#-insert@set
pair insert(const value_type& x)
/*-p-
@b:Attempts to insert an element into the set.
@p:x  Element to be inserted.
@r:A pair, of which the first element is an iterator that points
to the possibly inserted element, and the second is a bool
that is true if the element was actually inserted.
This function attempts to insert an element into the set.  A set
relies on unique keys and thus an element is only inserted if it is
not already present in the set.
Insertion requires logarithmic time.
/*/

#-insert@set
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Attempts to insert an element into the set.
@p:position  An iterator that serves as a hint as to where the
@p:x  Element to be inserted.
@r:An iterator that points to the element with key of  x (may
element should be inserted.
or may not be the element passed in).
This function is not concerned about whether the insertion took place,
and thus does not return a boolean like the single-argument insert()
does.  Note that the first parameter is only a hint and can
potentially improve the performance of the insertion process.  A bad
hint would cause no gains in efficiency.
See http://gcc.gnu.org/onlinedocs/libstdc++/23_containers/howto.html#4
for more on "hinting".
Insertion requires logarithmic time (if the hint is not taken).
/*/

#-insert@set
void insert(_InputIterator first, InputIterator last)
/*-p-
@b:A template function that attemps to insert a range of elements.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range.
inserted.
Complexity similar to that of the range constructor.
/*/

#-erase@set
void erase(iterator position)
/*-p-
@b:Erases an element from a set.
@p:position  An iterator pointing to the element to be erased.
This function erases an element, pointed to by the given iterator,
from a set.  Note that this function only erases the element, and
that if the element is itself a pointer, the pointed-to memory is not
touched in any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@set
size_type erase(const key_type& x)
/*-p-
@b:Erases elements according to the provided key.
@p:x  Key of element to be erased.
@r:The number of elements erased.
This function erases all the elements located by the given key from
a set.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-erase@set
void erase(iterator first, iterator last)
/*-p-
@b:Erases a [first,last) range of elements from a set.
@p:first  Iterator pointing to the start of the range to be
@p:last  Iterator pointing to the end of the range to be erased.
erased.
This function erases a sequence of elements from a set.
Note that this function only erases the element, and that if
the element is itself a pointer, the pointed-to memory is not touched
in any way.  Managing the pointer is the user's responsibilty.
/*/

#-clear@set
void clear()
/*-p-
Erases all elements in a set.  Note that this function only erases
the elements, and that if the elements themselves are pointers, the
pointed-to memory is not touched in any way.  Managing the pointer is
the user's responsibilty.
/*/

#-count@set
size_type count(const key_type& x) const
/*-p-
@b:Finds the number of elements.
@p:x  Element to located.
@r:Number of elements with specified key.
This function only makes sense for multisets; for set the result will
either be 0 (not present) or 1 (present).
/*/

#-find@set
iterator find(const key_type& x)
/*-p-
@b:Tries to locate an element in a set.
@p:x  Element to be located.
@r:Iterator pointing to sought-after element, or end() if not
found.
This function takes a key and tries to locate the element with which
the key matches.  If successful the function returns an iterator
pointing to the sought after element.  If unsuccessful it returns the
past-the-end ( c end() ) iterator.
/*/

#-find@set
const_iterator find(const key_type& x) const
/*-p-
/*/

#-lower_bound@set
iterator lower_bound(const key_type& x)
/*-p-
@b:Finds the beginning of a subsequence matching given key.
@p:x  Key to be located.
@r:Iterator pointing to first element equal to or greater
than key, or end().
This function returns the first element of a subsequence of elements
that matches the given key.  If unsuccessful it returns an iterator
pointing to the first element that has a greater value than given key
or end() if no such element exists.
/*/

#-lower_bound@set
const_iterator lower_bound(const key_type& x) const
/*-p-
/*/

#-upper_bound@set
iterator upper_bound(const key_type& x)
/*-p-
@b:Finds the end of a subsequence matching given key.
@p:x  Key to be located.
@r:Iterator pointing to the first element
greater than key, or end().
/*/

#-upper_bound@set
const_iterator upper_bound(const key_type& x) const
/*-p-
/*/

#-equal_range@set
pair equal_range(const key_type& x)
/*-p-
@b:Finds a subsequence matching given key.
@p:x  Key to be located.
@r:Pair of iterators that possibly points to the subsequence
matching given key.
This function is equivalent to
std::make_pair(c.lower_bound(val),
c.upper_bound(val))
(but is faster than making the calls separately).
This function probably only makes sense for multisets.
/*/

#-equal_range@set
pair equal_range(const key_type& x) const
/*-p-
/*/

#-operator ==@set
bool operator ==(const set<_K1, C1, A1>&, const set<_K1, C1, A1>&)
/*-p-
/*/

#-operator <@set
bool operator <(const set<_K1, C1, A1>&, const set<_K1, C1, A1>&)
/*-p-
/*/

##:stack@stack
stack
/*-p-
@b:A standard container giving FILO behavior.
Meets many of the requirements of a
container,
but does not define anything to do with iterators.  Very few of the
other standard container interfaces are defined.
This is not a true container, but an e adaptor.  It holds
another container, and provides a wrapper interface to that
container.  The wrapper is what enforces strict
first-in-last-out stack behavior.
The second template parameter defines the type of the underlying
sequence/container.  It defaults to std::deque, but it can be
any type that supports c back, c push_back, and c pop_front,
such as std::list, std::vector, or an appropriate user-defined
type.
Members not found in "normal" containers are c container_type,
which is a typedef for the second Sequence parameter, and c
push, c pop, and c top, which are standard stack/FILO
operations.
/*/

#-operator ==@stack
bool operator ==(const stack<_Tp1, Seq1>&, const stack<_Tp1, Seq1>&)
/*-p-
/*/

#-operator <@stack
bool operator <(const stack<_Tp1, Seq1>&, const stack<_Tp1, Seq1>&)
/*-p-
/*/

#-stack@stack
explicit stack(const Sequence& c = Sequence())
/*-p-
@b:Default constructor creates no elements.
/*/

#-empty@stack
bool empty() const
/*-p-
Returns true if the stack is empty.
/*/

#-size@stack
size_type size() const
/*-p-
/*/

#-top@stack
reference top()
/*-p-
Returns a read/write reference to the data at the first
element of the stack.
/*/

#-top@stack
const_reference top() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the stack.
/*/

#-push@stack
void push(const value_type& x)
/*-p-
@b:Add data to the top of the stack.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the top of the stack and assigns the given data
to it.  The time complexity of the operation depends on the
underlying sequence.
/*/

#-pop@stack
void pop()
/*-p-
@b:Removes first element.
This is a typical stack operation.  It shrinks the stack
by one.  The time complexity of the operation depends on the
underlying sequence.
Note that no data is returned, and if the first element's
data is needed, it should be retrieved before pop() is
called.
/*/

##:valarray@valarray
valarray
/*-p-
@b:Smart array designed to support numeric processing.
@p:Tp  Type of object in the array.
A valarray is an array that provides constraints intended to allow for
effective optimization of numeric array processing by reducing the
aliasing that can result from pointer representations.  It represents a
one-dimensional array from which different multidimensional subsets can
be accessed and modified.
/*/
#-valarray@valarray
valarray()
/*-p-
/*/

#-valarray@valarray
explicit valarray(size_t)
/*-p-
/*/

#-valarray@valarray
valarray(const Tp&, size_t)
/*-p-
/*/

#-valarray@valarray
valarray(const Tp* restrict__, size_t)
/*-p-
/*/

#-valarray@valarray
valarray(const valarray&)
/*-p-
/*/

#-valarray@valarray
valarray(const slice_array<_Tp>&)
/*-p-
/*/

#-valarray@valarray
valarray(const gslice_array<_Tp>&)
/*-p-
/*/

#-valarray@valarray
valarray(const mask_array<_Tp>&)
/*-p-
/*/

#-valarray@valarray
valarray(const indirect_array<_Tp>&)
/*-p-
/*/

#-valarray@valarray
template valarray(const Expr<_Dom, Tp>& e)
/*-p-
/*/

#-~valarray@valarray
~valarray()
/*-p-
/*/

#-operator =@valarray
valarray operator =(const valarray<_Tp>&)
/*-p-
@b:Assign elements to an array.
@p:v  Valarray to get values from.
Assign elements of array to values in  v.  Results are undefined
if  v does not have the same size as this array.
/*/

#-operator =@valarray
valarray operator =(const Tp&)
/*-p-
@b:Assign elements to a value.
@p:t  Value for elements.
Assign all elements of array to  t.
/*/

#-operator =@valarray
valarray operator =(const slice_array<_Tp>&)
/*-p-
@b:Assign elements to an array subset.
@p:sa  Array slice to get values from.
Assign elements of array to values in  sa.  Results are undefined
if  sa does not have the same size as this array.
/*/

#-operator =@valarray
valarray operator =(const gslice_array<_Tp>&)
/*-p-
@b:Assign elements to an array subset.
@p:ga  Array slice to get values from.
Assign elements of array to values in  ga.  Results are undefined
if  ga does not have the same size as this array.
/*/

#-operator =@valarray
valarray operator =(const mask_array<_Tp>&)
/*-p-
@b:Assign elements to an array subset.
@p:ma  Array slice to get values from.
Assign elements of array to values in  ma.  Results are undefined
if  ma does not have the same size as this array.
/*/

#-operator =@valarray
valarray operator =(const indirect_array<_Tp>&)
/*-p-
@b:Assign elements to an array subset.
@p:ia  Array slice to get values from.
Assign elements of array to values in  ia.  Results are undefined
if  ia does not have the same size as this array.
/*/

#-operator =@valarray
valarray operator =(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator []@valarray
_Tp operator [](size_t)
/*-p-
@p:i  Index of element to return.
@r:Reference to the i'th element.
Return a reference to the i'th array element.  
/*/

#-operator []@valarray
_Tp operator [](size_t) const
/*-p-
/*/

#-operator []@valarray
_Expr operator [](slice) const
/*-p-
@b:Return an array subset.
@p:s  The source slice.
@r:New valarray containing elements in  s.
Returns a new valarray containing the elements of the array
indicated by the slice argument.  The new valarray has the same size
as the input slice.  see slice.
/*/

#-operator []@valarray
slice_array operator [](slice)
/*-p-
@b:Return a reference to an array subset.
@p:s  The source slice.
@r:New valarray containing elements in  s.
Returns a new valarray containing the elements of the array
indicated by the slice argument.  The new valarray has the same size
as the input slice.  see slice.
/*/

#-operator []@valarray
_Expr operator [](const gslice&) const
/*-p-
@b:Return an array subset.
@p:s  The source slice.
@r:Slice_array referencing elements indicated by  s.
Returns a slice_array referencing the elements of the array
indicated by the slice argument.  see gslice.
/*/

#-operator []@valarray
gslice_array operator [](const gslice&)
/*-p-
@b:Return a reference to an array subset.
@p:s  The source gslice.
@r:New valarray containing elements in  s.
Returns a new valarray containing the elements of the array
indicated by the gslice argument.  The new valarray has
the same size as the input gslice.  see gslice.
/*/

#-operator []@valarray
valarray operator [](const valarray<bool>&) const
/*-p-
@b:Return an array subset.
@p:m  The valarray bitmask.
@r:New valarray containing elements indicated by  m.
Returns a new valarray containing the elements of the array
indicated by the argument.  The input is a valarray of bool which
represents a bitmask indicating which elements should be copied into
the new valarray.  Each element of the array is added to the return
valarray if the corresponding element of the argument is true.
/*/

#-operator []@valarray
mask_array operator [](const valarray<bool>&)
/*-p-
@b:Return a reference to an array subset.
@p:m  The valarray bitmask.
@r:New valarray containing elements indicated by  m.
Returns a new mask_array referencing the elements of the array
indicated by the argument.  The input is a valarray of bool which
represents a bitmask indicating which elements are part of the
subset.  Elements of the array are part of the subset if the
corresponding element of the argument is true.
/*/

#-operator []@valarray
_Expr operator [](const valarray<size_t>&) const
/*-p-
@b:Return an array subset.
@p:i  The valarray element index list.
@r:New valarray containing elements in  s.
Returns a new valarray containing the elements of the array
indicated by the argument.  The elements in the argument are
interpreted as the indices of elements of this valarray to copy to
the return valarray.
/*/

#-operator []@valarray
indirect_array operator [](const valarray<size_t>&)
/*-p-
@b:Return a reference to an array subset.
@p:i  The valarray element index list.
@r:Indirect_array referencing elements in  i.
Returns an indirect_array referencing the elements of the array
indicated by the argument.  The elements in the argument are
interpreted as the indices of elements of this valarray to include
in the subset.  The returned indirect_array refers to these
elements.
/*/

#-operator +@valarray
_Rt operator +() const
/*-p-
/*/

#-operator -@valarray
_Rt operator -() const
/*-p-
/*/

#-operator ~@valarray
_Rt operator ~() const
/*-p-
/*/

#-operator !@valarray
_Rt operator !() const
/*-p-
/*/

#-operator *=@valarray
valarray operator *=(const Tp&)
/*-p-
/*/

#-operator /=@valarray
valarray operator /=(const Tp&)
/*-p-
/*/

#-operator %=@valarray
valarray operator %=(const Tp&)
/*-p-
/*/

#-operator +=@valarray
valarray operator +=(const Tp&)
/*-p-
/*/

#-operator -=@valarray
valarray operator -=(const Tp&)
/*-p-
/*/

#-operator ^=@valarray
valarray operator ^=(const Tp&)
/*-p-
/*/

#-operator &=@valarray
valarray operator &=(const Tp&)
/*-p-
/*/

#-operator |=@valarray
valarray operator |=(const Tp&)
/*-p-
/*/

#-operator <<=@valarray
valarray operator <<=(const Tp&)
/*-p-
/*/

#-operator >>=@valarray
valarray operator >>=(const Tp&)
/*-p-
/*/

#-operator *=@valarray
valarray operator *=(const valarray<_Tp>&)
/*-p-
/*/

#-operator /=@valarray
valarray operator /=(const valarray<_Tp>&)
/*-p-
/*/

#-operator %=@valarray
valarray operator %=(const valarray<_Tp>&)
/*-p-
/*/

#-operator +=@valarray
valarray operator +=(const valarray<_Tp>&)
/*-p-
/*/

#-operator -=@valarray
valarray operator -=(const valarray<_Tp>&)
/*-p-
/*/

#-operator ^=@valarray
valarray operator ^=(const valarray<_Tp>&)
/*-p-
/*/

#-operator |=@valarray
valarray operator |=(const valarray<_Tp>&)
/*-p-
/*/

#-operator &=@valarray
valarray operator &=(const valarray<_Tp>&)
/*-p-
/*/

#-operator <<=@valarray
valarray operator <<=(const valarray<_Tp>&)
/*-p-
/*/

#-operator >>=@valarray
valarray operator >>=(const valarray<_Tp>&)
/*-p-
/*/

#-operator *=@valarray
valarray operator *=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator /=@valarray
valarray operator /=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator %=@valarray
valarray operator %=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator +=@valarray
valarray operator +=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator -=@valarray
valarray operator -=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator ^=@valarray
valarray operator ^=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator |=@valarray
valarray operator |=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator &=@valarray
valarray operator &=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator <<=@valarray
valarray operator <<=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-operator >>=@valarray
valarray operator >>=(const Expr<_Dom, Tp>&)
/*-p-
/*/

#-size@valarray
size_t size() const
/*-p-
/*/

#-sum@valarray
_Tp sum() const
/*-p-
@b:Return the sum of all elements in the array.
Accumulates the sum of all elements into a Tp using +=.  The order
of adding the elements is unspecified.
/*/

#-min@valarray
_Tp min() const
/*-p-
/*/

#-max@valarray
_Tp max() const
/*-p-
/*/

#-shift@valarray
valarray shift(int) const
/*-p-
@b:Return a shifted array.
@p:n  Number of element positions to shift.
@r:New valarray with elements in shifted positions.
A new valarray is constructed as a copy of this array with elements
in shifted positions.  For an element with index i, the new position
is i - n.  The new valarray has the same size as the current one.
New elements without a value are set to 0.  Elements whose new
position is outside the bounds of the array are discarded.
Positive arguments shift toward index 0, discarding elements [0, n).
Negative arguments discard elements from the top of the array.
/*/

#-cshift@valarray
valarray cshift(int) const
/*-p-
@b:Return a rotated array.
@p:n  Number of element positions to rotate.
@r:New valarray with elements in shifted positions.
A new valarray is constructed as a copy of this array with elements
in shifted positions.  For an element with index i, the new position
is (i - n)  size().  The new valarray has the same size as the
current one.  Elements that are shifted beyond the array bounds are
shifted into the other end of the array.  No elements are lost.
Positive arguments shift toward index 0, wrapping around the top.
Negative arguments shift towards the top, wrapping around to 0.
/*/

#-func@valarray
apply func(_Tp) const
/*-p-
@b:Apply a function to the array.
@p:func  Function of Tp returning Tp to apply.
@r:New valarray with transformed elements.
Returns a new valarray with elements assigned to the result of
applying func to the corresponding element of this array.  The new
array has the same size as this one.
/*/

#-func@valarray
apply func(const Tp&) const
/*-p-
@b:Apply a function to the array.
@p:func  Function of const Tp& returning Tp to apply.
@r:New valarray with transformed elements.
Returns a new valarray with elements assigned to the result of
applying func to the corresponding element of this array.  The new
array has the same size as this one.
/*/

#-resize@valarray
void resize(size_t size, Tp c = Tp())
/*-p-
@b:Resize array.
@p:size  New array size.
@p:c  New value for all elements.
Resize this array to  size and set all elements to  c.  All
references and iterators are invalidated.
/*/

#-operator []@valarray
operator [](size_t i) const
/*-p-
/*/

#-operator []@valarray
operator [](size_t i)
/*-p-
/*/

#-valarray@valarray
inline valarray()
/*-p-
/*/

#-valarray@valarray
inline valarray(size_t n)
/*-p-
/*/

#-valarray@valarray
inline valarray(const Tp& t, size_t n)
/*-p-
/*/

#-valarray@valarray
inline valarray(const Tp* restrict__ p, size_t n)
/*-p-
/*/

#-valarray@valarray
inline valarray(const valarray<_Tp>& v)
/*-p-
/*/

#-valarray@valarray
inline valarray(const slice_array<_Tp>& sa)
/*-p-
/*/

#-valarray@valarray
inline valarray(const gslice_array<_Tp>& ga)
/*-p-
/*/

#-valarray@valarray
inline valarray(const mask_array<_Tp>& ma)
/*-p-
/*/

#-valarray@valarray
inline valarray(const indirect_array<_Tp>& ia)
/*-p-
/*/

#-valarray@valarray
valarray(const Expr<_Dom, Tp>& e)
/*-p-
/*/

#-~valarray@valarray
inline ~valarray()
/*-p-
/*/

#-operator =@valarray
operator =(const valarray<_Tp>& v)
/*-p-
/*/

#-operator =@valarray
operator =(const Tp& t)
/*-p-
/*/

#-operator =@valarray
operator =(const slice_array<_Tp>& sa)
/*-p-
/*/

#-operator =@valarray
operator =(const gslice_array<_Tp>& ga)
/*-p-
/*/

#-operator =@valarray
operator =(const mask_array<_Tp>& ma)
/*-p-
/*/

#-operator =@valarray
operator =(const indirect_array<_Tp>& ia)
/*-p-
/*/

#-operator =@valarray
operator =(const Expr<_Dom, Tp>& e)
/*-p-
/*/

#-operator []@valarray
operator [](slice s) const
/*-p-
/*/

#-operator []@valarray
operator [](slice s)
/*-p-
/*/

#-operator []@valarray
operator [](const gslice& gs) const
/*-p-
/*/

#-operator []@valarray
operator [](const gslice& gs)
/*-p-
/*/

#-operator []@valarray
operator [](const valarray<bool>& m) const
/*-p-
/*/

#-operator []@valarray
operator [](const valarray<bool>& m)
/*-p-
/*/

#-operator []@valarray
operator [](const valarray<size_t>& i) const
/*-p-
/*/

#-operator []@valarray
operator [](const valarray<size_t>& i)
/*-p-
/*/

#-size@valarray
size() const
/*-p-
/*/

#-sum@valarray
sum() const
/*-p-
/*/

#-shift@valarray
shift(int n) const
/*-p-
/*/

#-cshift@valarray
cshift(int n) const
/*-p-
/*/

#-resize@valarray
resize(size_t n, Tp c)
/*-p-
/*/

#-min@valarray
min() const
/*-p-
/*/

#-max@valarray
max() const
/*-p-
/*/

#-func@valarray
func(_Tp) const
/*-p-
/*/

#-func@valarray
func(const Tp &) const
/*-p-
/*/

##:vector@vector
vector
/*-p-
@b:A standard container which offers fixed time access to
individual elements in any order.
Meets the requirements of a container, a
<a href="tables.html#66">reversible container</a>, and a
<a href="tables.html#67">sequence</a>, including the
<a href="tables.html#68">optional sequence requirements</a> with the
exception of c push_front and c pop_front.
In some terminology a vector can be described as a dynamic
C-style array, it offers fast and efficient access to individual
elements in any order and saves the user from worrying about
memory and size allocation.  Subscripting ( c [] ) access is
also provided as with C-style arrays.
/*/


#-iterator@vector
__normal_iterator iterator
/*-p-
/*/

#-const_iterator@vector
__normal_iterator const_iterator
/*-p-
/*/

#-const_reverse_iterator@vector
reverse_iterator const_reverse_iterator
/*-p-
/*/

#-reverse_iterator@vector
reverse_iterator reverse_iterator
/*-p-
/*/

#-vector@vector
explicit vector(const allocator_type& a = allocator_type())
/*-p-
@b:Default constructor creates no elements.
/*/

#-vector@vector
explicit vector(size_type n, const value_type& value = value_type(), const allocator_type& a = allocator_type())
/*-p-
@b:Create a vector with copies of an exemplar element.
@p:n  The number of elements to initially create.
@p:value  An element to copy.
This constructor fills the vector with  n copies of  value.
/*/

#-vector@vector
vector(const vector& x)
/*-p-
@b:Vector copy constructor.
@p:x  A vector of identical element and allocator types.
The newly-created vector uses a copy of the allocation
object used by  x.  All the elements of  x are copied,
but any extra memory in
/*/

#-vector@vector
template vector(_InputIterator first, InputIterator last, const allocator_type& a = allocator_type())
/*-p-
@b:Builds a vector from a range.
@p:first  An input iterator.
@p:last  An input iterator.
Create a vector consisting of copies of the elements from
[first,last).
If the iterators are forward, bidirectional, or
random-access, then this will call the elements' copy
constructor N times (where N is distance(first,last)) and do
no memory reallocation.  But if only input iterators are
used, then this will do at most 2N calls to the copy
constructor, and logN memory reallocations.
/*/

#-~vector@vector
~vector()
/*-p-
The dtor only erases the elements, and note that if the
elements themselves are pointers, the pointed-to memory is
not touched in any way.  Managing the pointer is the user's
responsibilty.
/*/

#-operator =@vector
vector operator =(const vector& x)
/*-p-
@b:Vector assignment operator.
@p:x  A vector of identical element and allocator types.
All the elements of  x are copied, but any extra memory in
copy constructor, the allocator object is not copied.
/*/

#-assign@vector
void assign(size_type n, const value_type& val)
/*-p-
@b:Assigns a given value to a vector.
@p:n  Number of elements to be assigned.
@p:val  Value to be assigned.
This function fills a vector with  n copies of the given
value.  Note that the assignment completely changes the
vector and that the resulting vector's size is the same as
the number of elements assigned.  Old data may be lost.
/*/

#-assign@vector
void assign(_InputIterator first, InputIterator last)
/*-p-
@b:Assigns a range to a vector.
@p:first  An input iterator.
@p:last   An input iterator.
This function fills a vector with copies of the elements in the
range [first,last).
Note that the assignment completely changes the vector and
that the resulting vector's size is the same as the number
of elements assigned.  Old data may be lost.
/*/

#-begin@vector
iterator begin()
/*-p-
Returns a read/write iterator that points to the first
element in the vector.  Iteration is done in ordinary
element order.
/*/

#-begin@vector
const_iterator begin() const
/*-p-
Returns a read-only (constant) iterator that points to the
first element in the vector.  Iteration is done in ordinary
element order.
/*/

#-end@vector
iterator end()
/*-p-
Returns a read/write iterator that points one past the last
element in the vector.  Iteration is done in ordinary
element order.
/*/

#-end@vector
const_iterator end() const
/*-p-
Returns a read-only (constant) iterator that points one past
the last element in the vector.  Iteration is done in
ordinary element order.
/*/

#-rbegin@vector
reverse_iterator rbegin()
/*-p-
Returns a read/write reverse iterator that points to the
last element in the vector.  Iteration is done in reverse
element order.
/*/

#-rbegin@vector
const_reverse_iterator rbegin() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to the last element in the vector.  Iteration is done in
reverse element order.
/*/

#-rend@vector
reverse_iterator rend()
/*-p-
Returns a read/write reverse iterator that points to one
before the first element in the vector.  Iteration is done
in reverse element order.
/*/

#-rend@vector
const_reverse_iterator rend() const
/*-p-
Returns a read-only (constant) reverse iterator that points
to one before the first element in the vector.  Iteration
is done in reverse element order.
/*/

#-size@vector
size_type size() const
/*-p-
/*/

#-max_size@vector
size_type max_size() const
/*-p-
/*/

#-resize@vector
void resize(size_type new_size, value_type x = value_type())
/*-p-
@b:Resizes the vector to the specified number of elements.
@p:new_size  Number of elements the vector should contain.
@p:x  Data with which new elements should be populated.
This function will resize the vector to the specified
number of elements.  If the number is smaller than the
vector's current size the vector is truncated, otherwise
the vector is extended and new elements are populated with
given data.
/*/

#-capacity@vector
size_type capacity() const
/*-p-
Returns the total number of elements that the vector can
hold before needing to allocate more memory.
/*/

#-empty@vector
bool empty() const
/*-p-
Returns true if the vector is empty.  (Thus begin() would
equal end().)
/*/

#-reserve@vector
void reserve(size_type n)
/*-p-
@b:Attempt to preallocate enough memory for specified number of
@p:n  Number of elements required.
@t:std::length_error  If  n exceeds c max_size().
elements.
This function attempts to reserve enough memory for the
vector to hold the specified number of elements.  If the
number requested is more than max_size(), length_error is
thrown.
The advantage of this function is that if optimal code is a
necessity and the user can determine the number of elements
that will be required, the user can reserve the memory in
dvance, and thus prevent a possible reallocation of memory
and copying of vector data.
/*/

#-operator []@vector
reference operator [](size_type n)
/*-p-
@b:Subscript access to the data contained in the vector.
@p:n The index of the element for which data should be
@r:Read/write reference to data.
accessed.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)
/*/

#-operator []@vector
const_reference operator [](size_type n) const
/*-p-
@b:Subscript access to the data contained in the vector.
@p:n The index of the element for which data should be
@r:Read-only (constant) reference to data.
accessed.
This operator allows for easy, array-style, data access.
Note that data access with this operator is unchecked and
out_of_range lookups are not defined. (For checked lookups
see at().)
/*/

#-at@vector
reference at(size_type n)
/*-p-
@b:Provides access to the data contained in the vector.
@p:n The index of the element for which data should be
@r:Read/write reference to data.
@t:std::out_of_range  If  n is an invalid index.
accessed.
This function provides for safer data access.  The parameter
is first checked that it is in the range of the vector.  The
function throws out_of_range if the check fails.
/*/

#-at@vector
const_reference at(size_type n) const
/*-p-
@b:Provides access to the data contained in the vector.
@p:n The index of the element for which data should be
@r:Read-only (constant) reference to data.
@t:std::out_of_range  If  n is an invalid index.
accessed.
This function provides for safer data access.  The parameter
is first checked that it is in the range of the vector.  The
function throws out_of_range if the check fails.
/*/

#-front@vector
reference front()
/*-p-
Returns a read/write reference to the data at the first
element of the vector.
/*/

#-front@vector
const_reference front() const
/*-p-
Returns a read-only (constant) reference to the data at the first
element of the vector.
/*/

#-back@vector
reference back()
/*-p-
Returns a read/write reference to the data at the last
element of the vector.
/*/

#-back@vector
const_reference back() const
/*-p-
Returns a read-only (constant) reference to the data at the
last element of the vector.
/*/

#-data@vector
pointer data()
/*-p-
Returns a pointer such that [data(), data() + size()) is a valid
range.  For a non-empty vector, data() == &front().
/*/

#-data@vector
const_pointer data() const
/*-p-
/*/

#-push_back@vector
void push_back(const value_type& x)
/*-p-
@b:Add data to the end of the vector.
@p:x  Data to be added.
This is a typical stack operation.  The function creates an
element at the end of the vector and assigns the given data
to it.  Due to the nature of a vector this operation can be
done in constant time if the vector has preallocated space
available.
/*/

#-pop_back@vector
void pop_back()
/*-p-
@b:Removes last element.
This is a typical stack operation. It shrinks the vector by one.
Note that no data is returned, and if the last element's
data is needed, it should be retrieved before pop_back() is
called.
/*/

#-insert@vector
iterator insert(iterator position, const value_type& x)
/*-p-
@b:Inserts given value into vector before specified iterator.
@p:position  An iterator into the vector.
@p:x  Data to be inserted.
@r:An iterator that points to the inserted data.
This function will insert a copy of the given value before
the specified location.  Note that this kind of operation
could be expensive for a vector and if it is frequently
used the user should consider using std::list.
/*/

#-insert@vector
void insert(iterator position, size_type n, const value_type& x)
/*-p-
@b:Inserts a number of copies of given data into the vector.
@p:position  An iterator into the vector.
@p:n  Number of elements to be inserted.
@p:x  Data to be inserted.
This function will insert a specified number of copies of
the given data before the location specified by  position.
Note that this kind of operation could be expensive for a
vector and if it is frequently used the user should
consider using std::list.
/*/

#-insert@vector
void insert(iterator position, InputIterator first, InputIterator last)
/*-p-
@b:Inserts a range into the vector.
@p:position  An iterator into the vector.
@p:first  An input iterator.
@p:last   An input iterator.
This function will insert copies of the data in the range
[first,last) into the vector before the location specified
by  pos.
Note that this kind of operation could be expensive for a
vector and if it is frequently used the user should
consider using std::list.
/*/

#-erase@vector
iterator erase(iterator position)
/*-p-
@b:Remove element at given position.
@p:position  Iterator pointing to element to be erased.
@r:An iterator pointing to the next element (or end()).
This function will erase the element at the given position and thus
shorten the vector by one.
Note This operation could be expensive and if it is
frequently used the user should consider using std::list.
The user is also cautioned that this function only erases
the element, and that if the element is itself a pointer,
the pointed-to memory is not touched in any way.  Managing
the pointer is the user's responsibilty.
/*/

#-erase@vector
iterator erase(iterator first, iterator last)
/*-p-
@b:Remove a range of elements.
@p:first  Iterator pointing to the first element to be erased.
@p:last  Iterator pointing to one past the last element to be
@r:An iterator pointing to the element pointed to by  last
erased.
prior to erasing (or end()).
This function will erase the elements in the range [first,last) and
shorten the vector accordingly.
Note This operation could be expensive and if it is
frequently used the user should consider using std::list.
The user is also cautioned that this function only erases
the elements, and that if the elements themselves are
pointers, the pointed-to memory is not touched in any way.
Managing the pointer is the user's responsibilty.
/*/

#-swap@vector
void swap(vector& x)
/*-p-
@b:Swaps data with another vector.
@p:x  A vector of the same element and allocator types.
This exchanges the elements between two vectors in constant time.
(Three pointers, so it should be quite fast.)
Note that the global std::swap() function is specialized such that
std::swap(v1,v2) will feed to this function.
/*/

#-clear@vector
void clear()
/*-p-
Erases all the elements.  Note that this function only erases the
elements, and that if the elements themselves are pointers, the
pointed-to memory is not touched in any way.  Managing the pointer is
the user's responsibilty.
/*/

